package jsadppac.sarapac;

import java.util.Arrays;
import java.util.EventListener;
import java.util.List;
import java.util.logging.Logger;

import javax.swing.event.EventListenerList;

import sadppac.CenaPacket;
import sadppac.SaraAscFile;
import sadppac.SaraBufferedDemultiplexor;
import sadppac.SaraPacket;
import sadppac.SwimPacket;

/**
 * New implementaion by java of SARA Demultiplexor.
 * This is an event driven demultiplexor.
 * Demultiplexor is started by {@link #startDemultiplex(SaraAscFile, long)}.
 * Before starting demultiplexor, add arrival listeners for Cena and/or Swim
 * by {@link #addCenaPacketListener(CenaPacketArrivalListener)} and/or
 * {@link #addSwimPacketListener(SwimPacketArrivalListener)}.
 * 
 * <pre>
// Define the SARA ascii file (BMU/ASCII format)
SaraAscFile file=new SaraAscFile(filename);

// Instance the demultiplexor
JSaraDemultiplexor demul=new JSaraDemultiplexor();

// Add listener of CenaPacket.
demul.addCenaPacketListener(new CenaPacketArrivalListener(){
	public void cenaPacketArrived(CenaPacketArrivalEvent cenaEvent) {
		System.out.println("CENA packet arrived: "+
						cenaEvent.getPacket().get4sCycleCntr());
	}});

// Add listener of SwimPacket
demul.addSwimPacketListener(new SwimPacketArrivalListener(){
	public void swimPacketArrived(SwimPacketArrivalEvent swimEvent) {
		System.out.println("SWIM packet arrived: "+
						swimEvent.getPacket().get8sCycleCntr());
	}});

// Start demultiplexing with specifying the BMU/ASCII file.
demul.startDemultiplex(file, 0);

 * 
 * 
 * </pre>
 * 
 * @author futaana
 *
 */
public class JSaraDemultiplexor {
	
    private transient EventListenerList cenaListener;
    private transient EventListenerList swimListener;

    /**
     * Instance the demultiplexor
     */
	public JSaraDemultiplexor(){
		this.cenaListener=new EventListenerList();
		this.swimListener=new EventListenerList();
	}
	
	/**
	 * Start the demutiplexor.  If {@link CenaPacket} or {@link SwimPacket} is found,
	 * {@link CenaPacketArrivalEvent} or {@link SwimPacketArrivalEvent} are fired
	 * to all the registered listeners (use {@link #addCenaPacketListener(CenaPacketArrivalListener)}
	 * and {@link #addSwimPacketListener(SwimPacketArrivalListener)} to register).
	 * @param ascFile BMU-formatted file.
	 * @param waitMsec Wait between SaraPacket.
	 */
	public void startDemultiplex(SaraAscFile ascFile, long waitMsec){
		SaraBufferedDemultiplexor demul=new SaraBufferedDemultiplexor();
		if(ascFile.openStream()){
			while(!ascFile.eof()){
				SaraPacket saraPac=ascFile.getSaraPacket();
				if(saraPac==null){
					Logger.global.warning("SARAPAC IS NULL.");
					continue;
				}
				
				demul.appendSaraPacket(saraPac);
				
				CenaPacket cenaPac=demul.getCena();
				if(cenaPac!=null){
					fireCenaPacketArrived(cenaPac);
				}
				
				SwimPacket swimPac=demul.getSwim();
				if(swimPac!=null){
					fireSwimPacketArrived(swimPac);
				}
				
				try {
					Thread.sleep(waitMsec);
				} catch (InterruptedException e) {
					e.printStackTrace();
				}
				
			}
			
		}
	}
	
	/**
	 * Register the {@link CenaPacketArrivalListener}.
	 * @param listener The listener for {@link CenaPacket}.
	 */
	public void addCenaPacketListener(CenaPacketArrivalListener listener){
		this.cenaListener.add(CenaPacketArrivalListener.class, listener);
	}
	/**
	 * Remove the {@link CenaPacketArrivalListener}.
	 * @param listener The listener to be removed.
	 */
	public void removeCenaPacketListener(CenaPacketArrivalListener listener){
		this.cenaListener.remove(CenaPacketArrivalListener.class, listener);
	}
	/**
	 * Inquire whether the listener is regstered or not.
	 * @param listener Listener to be inquired.
	 * @return True if the listener is already registered.
	 */
	public boolean hasCenaListener(EventListener listener){
		List list=Arrays.asList(this.cenaListener.getListenerList());
		return list.contains(listener);
	}
	/**
	 * Called from the {@link #startDemultiplex(SaraAscFile, long)} to fire the {@link CenaPacketArrivalEvent}.
	 * @param cpac {@link CenaPacket} to be send.
	 */
	protected void fireCenaPacketArrived(CenaPacket cpac){
		notifyCenaListeners(new CenaPacketArrivalEvent(this,cpac));
	}
	protected void notifyCenaListeners(CenaPacketArrivalEvent event){
		Object[] listeners = this.cenaListener.getListenerList();
		for(int i=listeners.length -2; i>=0; i-=2){
			if(listeners[i] == CenaPacketArrivalListener.class) {
				((CenaPacketArrivalListener) listeners[i+1]).cenaPacketArrived(event);
			}
		}
	}

	
	public void addSwimPacketListener(SwimPacketArrivalListener listener){
		this.cenaListener.add(SwimPacketArrivalListener.class, listener);
	}
	public void removeSwimPacketListener(SwimPacketArrivalListener listener){
		this.cenaListener.remove(SwimPacketArrivalListener.class, listener);
	}
	public boolean hasSwimListener(EventListener listener){
		List list=Arrays.asList(this.cenaListener.getListenerList());
		return list.contains(listener);
	}
	protected void fireSwimPacketArrived(SwimPacket cpac){
		notifySwimListeners(new SwimPacketArrivalEvent(this,cpac));
	}
	protected void notifySwimListeners(SwimPacketArrivalEvent event){
		Object[] listeners = this.cenaListener.getListenerList();
		for(int i=listeners.length -2; i>=0; i-=2){
			if(listeners[i] == SwimPacketArrivalListener.class) {
				((SwimPacketArrivalListener) listeners[i+1]).swimPacketArrived(event);
			}
		}
	}
	
}

