// Compilation is
// g++ SwimReadSample.cpp -I${HOME}/SARA/include -L${HOME}/SARA/lib -lascfmt -lsarapac -lirfpac -llinlog -lcrypto -lm
// Then, a.out is generated.
// ./a.out  SWIMFILE.swim

#include<iostream>
#include<ctime>

// The following three header files must be included.
#include<IrfPac/irfpac.h>
#include<IrfPac/SaraPac/sarapac.h>
#include<IrfPac/SaraPac/AscFmt/ascfmt.h>

// Three name spaces are used.
using namespace IrfPac;
using namespace IrfPac::SaraPac;
using namespace IrfPac::SaraPac::AscFmt;

int main(int argc, char** argv){

	if(argc==1){
		fprintf(stderr, "USAGE:   %s    filename\n", argv[0]);
		exit(-1);
	}

	// fname here is the name of the SWIM data file in BMU format.
	//
	char* fname=argv[1];

	// Prepare the SwimBmuReader object.  This object handles the SWIM data file.
	//
	SwimBmuReader fp;

	// Read the SWIM data file into the memory.
	//
	fp.readFile(fname);

	// Obtain the whole SWIM data included in the fname.  The data is stored into an instance of SwimPacketArray.
	// SwimPacketArray behaves as 'container' of the SWIM data.
	//
	SwimPacketArray *arr = fp.getArray();

	// If you want to know how much data is available in the SwimPacketArray, you can use getSize() method.
	// You can find documentation of methods implemented in SwimPacketArray class at html/classIrfPac_1_1SaraPac_1_1SwimPacketArray.html
	//
	fprintf(stdout, "Number of data = %ld\n", arr->size());
	
	// Now you want to fetch each SWIM data (every 8 seconds).
	// You can use getPacket(int) method.
	// For example, if you want 100-th data of the data file, you can use
	//
	SwimPacket* pac100 = arr->getPacket(100);

	// pac100 is an instance of SwimPacket.
	// Documentation is available at html/classIrfPac_1_1SaraPac_1_1SwimPacket.html
	// SwimPacket contains all the SWIM data at the specific time.
	// You can get the time of the packet by getUtc() method.
	// The obtained tiem is expressed in the UNIX epoch time system.  (millisec after the 1970-01-01).
	// NOTE: You must not use getUniversalTime() method.  This will not return the exact time.
	// 
	double utc100 = pac100->getUtc();
	fprintf(stdout, "Observation time (at UNIX epoch) is %f\n", utc100);

	time_t time100 = (time_t) (utc100);
	struct tm* tm100 = gmtime(&time100);
	char* asc100 = asctime(tm100);
	fprintf(stdout, "Observation time (human readable) is %s\n", asc100);


	// Then, important parameters are the binning parameter (number of the deflection and energy bins).
	// You may find methods getDeflectionBins() and getEnergyBins() methods in SwimPacket class.
	//
	int nd = pac100->getDeflectionBins();
	int ne = pac100->getEnergyBins();
	fprintf(stdout, "Nd = %d\n", nd);
	fprintf(stdout, "Ne = %d\n", ne);


	// Now it is ready to get 'START' counter.  This is a basic count rate data of SWIM.
	// The START counter data is in the so-called 'accumulation total matrix', which is a part of SwimPacket.
	// You may be able to retrieve the accumulation total matrix class by
	// 
	SwimAccumTotMatx* acc100 = pac100 -> getAccumTotMatx();

	// You may find documentation of SwimAccumTotMatx in html/classIrfPac_1_1SaraPac_1_1SwimAccumTotMatx.html
	

	// Due to the historical reason, the SWIM data in the maximum mode is organized by Nd=8 and Ne=32.
	// However, the real physical counts are organized by Nd=16 and Ne=16.
	// The SwimAccumTotMatx follows the 8x32 definition.
	// However, libsadppac can convert it to 16x16 in a very simple way.
	//
	SwimAccumTotMatx* acc100ff = acc100->reform16x16();

	// You may compare the number of deflection and energy bins.
	// 
	fprintf(stdout, "Original  matrix D=%d E=%d\n", acc100->getDeflectionBins(), acc100->getEnergyBins());
	fprintf(stdout, "Converted matrix D=%d E=%d\n", acc100ff->getDeflectionBins(), acc100ff->getEnergyBins());

	// You can get the count rate of the specific direction and energy by get(int, int, int) method.
	// To get the start count which you are working, you have to specify SwimAccumTotMatx::START_COUNTER as the third argument.
	// For example, if you want to get the D=13 and E=5, you can use the followings.
	//
	long counts_13_5 = acc100ff->get(13,  5, SwimAccumTotMatx::START_COUNTER);
	fprintf(stdout, "Counts at D=13 and E=5 is %ld\n", counts_13_5);



	////////// Hints for further programming
	//  By using for-'loop's, you may get all the counts from deflection, energy and time.
	//
	
	return 0;

}
