#ifndef __BIT_BUFFER_H
#define __BIT_BUFFER_H

#include<iostream>
#include<vector>
#include <inttypes.h>
#include <stdio.h>

namespace IrfPac{
namespace Encode{

/**@brief Class to treat bit stream.

	The bit treatment is not well supported by C++, I have made useful buffer for treating
	bit stream.  This BitBuffer class is the first implementation, only for writing buffer.
	
	The BitBuffer class has two buffers: byte buffer and bit buffer.
	The byte buffer conserves a series of byte stream, and the bit buffer is for a remaining bits.
	You can add bit stream by appendLBits() and appendMBits() methods.
	The size of the bit buffer does not exceed 8:  if the bit buffer size exceed 8, the first
	8 bits are converted to 1 byte storing to the byte buffer.
	You can also add padding to 8 bit by paddingToBoundary() method.
	Note that the bit buffer is ignored in most methods.
	
	When you want the buffered byte stream for further processing, you can use getBuffer() method.
	Also, write() method will dump the stream to the specified file pointer in a binary format.
*/
class BitBuffer {

        public:
		BitBuffer(){
			_subbuffer=0;
			_buffer_pos=0;
		}

		~BitBuffer(){
		}
		
		// Copy constructor
		BitBuffer(const BitBuffer& buf){
			_subbuffer=buf._subbuffer;
			_buffer_pos=buf._buffer_pos;
			_buffer=buf._buffer;            // WARN: IS IT OK?
		}
		
		/** @brief Obtain the size of the byte buffer in byte.
		*/
		size_t getSize(){
			return _buffer.size();
		}
		
		/** @brief Obtain the total size of both the byte and bit buffers in bits.
		*/
		size_t getBitSize(){
			return getSize()*8+_buffer_pos;
		}
		
		/**@brief Obtain the pointer to the buffer.  You need to delete[] the obtained pointer when it becomes useless.

			Obtain the pointer to the buffer.  The buffer is newly created, therefore the buffer should be delete[]ed
			after the use.
			
			The contents of the subbuffer is disregarded.  You have to use paddingToBoundary(uint8_t) method in order to
			fill-up the remaining subbuffer.
		*/
		uint8_t* getBuffer(){
			size_t bufferSize=getSize();
			uint8_t* buffer=new uint8_t[bufferSize];
			for(size_t i=0; i<bufferSize; i++){
				buffer[i] = _buffer[i];
			}
			return buffer;
		}
		
		/** @brief Append bit stream of the specified value with specified number of bits from LSB.
		*/
		void appendLBits(uint8_t value, size_t bits){
			if(bits > 8){
				std::cerr << "Bit size exception at " << __LINE__ << " in " << __FILE__ << std::endl;
				throw "BitSizeException";
			}
			
			value=(value << (8-bits));
			appendMBits(value, bits);
		}

		/** @brief Append bit stream of the specified value with specified number of bits from MSB.
		*/
		void appendMBits(uint8_t value, size_t bits){
			if(bits > 8){
				std::cerr << "Bit size exception at " << __LINE__ << " in " << __FILE__ << std::endl;
				throw "BitSizeException";
			}

			value = ((value >> (8-bits)) << (8-bits));   // The same effect to masking.
			uint16_t v16 = (value << 8);   //  Convert to 16 bit
			v16 = (v16 >> (_buffer_pos));
			
			_subbuffer = (_subbuffer | v16);  // append to subbuffer
			_buffer_pos += bits;

			if(_buffer_pos >=8){
				processBuffer();
			}

		}
		
		/** @brief Returns true if the bit pointer is at the 8 bit boundary.
		*/
		bool atBoundary(){
			if(_buffer_pos%8==0)return true;
			return false;
		}
		
		/** @brief Make a padding with specified value up to the 8 bit boundary.
		
			This operation does affect the getSize() or getBitSize() operation.
			If you want to know the number of bits without padding, you must call
			getBitSize() method before calling paddingToBoundary().
		*/
		void paddingToBoundary(uint8_t zero_or_one){
			if(zero_or_one!=0 && zero_or_one!=1){
				std::cerr << "Bit number exception at " << __LINE__ << " in " << __FILE__ << std::endl;
				throw "BitNumberException";
			}

			if(atBoundary())return;

			uint8_t val;
			
			if(zero_or_one==0)val=0x00;
			if(zero_or_one==1)val=0xff;
			
			appendLBits(val, 8-_buffer_pos);
		}
		
		/** @brief Print contents of the byte buffer and bit buffer.
		*/
		void showAllBuffer(FILE* fp){
			fprintf(fp, "+++++ Buffer Information +++++\n");
			showBuffer(fp);
			showSubBuffer(fp);
		}
		
		/** @brief Print contents of the byte buffer.
		*/
		void showBuffer(FILE* fp){
			fprintf(fp, "+++ ByteBuffer +++\n");
			fprintf(fp, "Size=%d\n", (int)getSize());
			fprintf(fp, "Contents[hex]=\n");
			print(fp);
		}
		
		/** @brief Print contents of the bit buffer.
		*/
		void showSubBuffer(FILE* fp){
			fprintf(fp, "+++ Sub-Buffer +++\n");
			fprintf(fp, "Size=%d\n", (int)_buffer_pos);
			fprintf(fp, "Contents[bin]=\n");
			for(uint16_t mask=0x8000; mask>0; mask>>=1){
				fprintf(fp, "%1d", (((_subbuffer & mask)==0) ? 0:1));
			}
			fprintf(fp, "\n");
			for(size_t i=0; i<_buffer_pos; i++){
				fprintf(fp, " ");
			}
			fprintf(fp, "^\n");
		}
		
		/** @brief Print the contents of buffer.
		
			The byte buffer is printed.  Bit buffer is ignored.
		*/
		void print(FILE* fp){
			for(size_t i=0; i<getSize(); i++){
				fprintf(fp, "%02x", _buffer[i]);
				if(i%2==31)fprintf(fp, "\n");
				else if(i%2==1)fprintf(fp, " ");
			}
			fprintf(fp, "\n");
		}
		
		/** @brief Write the contents of buffer.
		
			The byte buffer is wrote in binary.  Bit buffer is ignored.
			File pointer fp should be opened in binary mode.
		*/
		void write(FILE* fp){
			size_t size=getSize();
			uint8_t* buf=getBuffer();
			fwrite(buf, 1, size, fp);
			delete[] buf;
		}

		
	private:
		uint16_t _subbuffer;			// Bit buffer
		std::vector<uint8_t> _buffer;	// Byte buffer
		size_t _buffer_pos;				// Pointer of bit buffer position.

		/** @brief Internal processing of conversion from subbuffer to byte buffer if the bit buffer size exceed 8.
		*/
		void processBuffer(){
			if(_buffer_pos<8) return;

			// Obtain the first 8 bit to push_back to byte buffer.
			uint16_t newval=(_subbuffer & 0xff00) >> 8;
			uint8_t newval8=(uint8_t)newval;
			_buffer.push_back(newval8);

			// Subbuffer shifts 8 bit.
			_subbuffer <<= 8;
			_buffer_pos-=8;
		}
};


}}

#endif



