#ifndef IRFPACPACKETINTARR_H
#define IRFPACPACKETINTARR_H

#include "IrfPac/generalpacket.h"

namespace IrfPac {

/**	@brief Implementaion of GeneralPacket.

	This class is a template class, so that it is not recommended to be used.
	Use PacketU64Arr, PacketU32Arr, PacketU16Arr, and PacketU8Arr.
	These classes are immutable.

	@author	FUTAANA
*/
template<class T>
class PacketIntArr : public GeneralPacket
{
public:
	PacketIntArr(T *stream, int nIndex);
	/**@brief copy constructor
	*/
	PacketIntArr(const PacketIntArr<T>& obj);

	virtual ~PacketIntArr();
	/**@brief Returns the copy of the contents of byte stream.
		
		Since returned stream is a copy of the original, the returned
		stream should be explicitly deleted by delete[] operation.
	*/
	virtual uint8_t* getStream(void);
	virtual size_t dumpByteStream( FILE*);
	virtual size_t dumpHexStream( FILE* fp);
	virtual size_t getSize();
	virtual size_t getIndexSize();
    	virtual T* getValues(void);
    	virtual T getValue(size_t index);

	virtual void dumpAll(){dumpAll(stdout);};
	virtual void dumpAll(FILE *fp){fprintf(fp, "-- DUMP OF PACKET --\n");dumpHexStream(fp);};

	virtual size_t dumpHexStream(){
		return dumpHexStream(stdout);
	}

	/**@brief Returns a hash of the packet contents.

		The implementation uses MD5.  It needs openssl crypto packge.
	*/
	virtual std::string hash();

private:
	uint8_t* split2byte(T val);
	T *contents;
	size_t size;
	size_t nind;
	size_t bytes;
};

/**	@brief Implementation of packet constituted by 64bit (8byte) Array 
*/
typedef PacketIntArr<uint64_t> PacketU64Arr;

/**	@brief Implementation of packet constituted by 32bit (4byte) Array 
*/
typedef PacketIntArr<uint32_t> PacketU32Arr;

/**	@brief Implementation of packet constituted by 16bit (2byte) Array 
*/
typedef PacketIntArr<uint16_t> PacketU16Arr;

typedef PacketIntArr<uint8_t> _PacketU8Arr;
/**	@brief Implementation of packet constituted by 8bit (1byte) Array 

	This class is immutable, so that only the initialization and change
	is done in the constructor.
*/
class PacketU8Arr : public _PacketU8Arr{
public:
	PacketU8Arr(uint8_t* str, size_t size) : _PacketU8Arr(str, size) {};
	virtual ~PacketU8Arr(){}
	uint8_t* getStream(){
		return getValues();
	}
};

/**@brief Pure virtual formatter class of the PacketU8Arr.

	This is a base class of many formatter classes for PacketU8Arr.
	Refer to, e.g. BmuFormatter class, as a sample.
*/
class PacketU8ArrFormatter {
	public:
		virtual ~PacketU8ArrFormatter(){};
		virtual void format(FILE* fp, PacketU8Arr* packet)=0;
};

/**@brief One of the basic formatters: so-called BMU format.

This is a formatter for packet.  Output is the so-called BMU format.
Sample code is:
<pre>
uint8_t array[100];
for(int i=0; i<100; i++)array[i]=i;
PacketU8Arr* packet=new PacketU8Arr(array, 100);         // Instance PacketU8Arr in some ways.
PacketBmuFormatter *formatter;    // Instance formatter.
formatter.format(stdout, packet); // Format packet and print to stdout.
</pre>
Results looks like:
<pre>
0001 0203 0405 0607 0809 0a0b 0c0d 0e0f 1011 1213 1415 1617 1819 1a1b 1c1d 1e1f
2021 2223 2425 2627 2829 2a2b 2c2d 2e2f 3031 3233 3435 3637 3839 3a3b 3c3d 3e3f
4041 4243 4445 4647 4849 4a4b 4c4d 4e4f 5051 5253 5455 5657 5859 5a5b 5c5d 5e5f
6061 6263 0000 0000 0000 0000 0000 0000 0000 0000 0000 0000 0000 0000 0000 0000

</pre>
Normally the BMU format has header starting "*******", however, it is not
displayed in this formatter.

Sample of header is:
<pre>
********BMU-FMT at TIME XXX XXX XX XX:XX:XX XXXX
</pre>

*/
class PacketBmuFormatter : public PacketU8ArrFormatter {
	public:
		void format(FILE* fp, PacketU8Arr* packet){
//			std::string header=std::string("********BMU-FMT at TIME XXX XXX XX XX:XX:XX XXXX");
//			fprintf(fp, "%s\n", header.c_str());
			size_t size=packet->getSize();
			size_t residual=size%32;
			uint8_t* str=packet->getStream();
			for(size_t i=0; i<size; i++){
				fprintf(fp, "%02x", str[i]);
				if(i%2==1)fprintf(fp, " ");
				if(i%32==31)fprintf(fp, "\n");
			}
			if(residual != 0){
				while( residual !=32 ){
					fprintf(fp, "00");
					if(residual%2==1)fprintf(fp, " ");
					residual++;
				}
				fprintf(fp, "\n");
			}
			fprintf(fp, "\n");
			delete[] str;
		}

};



}

#endif
