/**@file SaraBufferedDemultiplexor.cpp
 * @brief A sample file of demultiplexing.
 */

#include<iostream>
#include <cstdlib>
#include<unistd.h>
#include<IrfPac/irfpac.h>
#include<IrfPac/SaraPac/sarapac.h>
#include<IrfPac/SaraPac/AscFmt/ascfmt.h>
#include <unistd.h>

using namespace std;
using namespace IrfPac;
using namespace IrfPac::SaraPac;
using namespace IrfPac::SaraPac::AscFmt;

/**@page DemultiplexorSample
 * @brief Sample code of demultiplexing.
 * @section HowToDemultiplexing How to demultiplexing
 * Because SaraPacket has complecated structure, it is quite
 * important to know how to analyze the packet.
 * SaraPacket contains parts of CenaPacket and SwimPacket.
 * The procedure to extract CenaPacket and/or SwimPacket from a series of
 * SaraPacket is called demultiplexing.
 * The simplest way of demultiplexing is to use
 * SaraBufferedDemultiplexor class.
 * @subsection RecipesDemultiplex Recipes of demultiplexing.
 * This section describes the recipes of demultiplexing.  You can also refer
 * to the sample code below.
 * @par
 * - Instance the SaraBufferedDemultiplexor class
 * 	@code
 * 		SaraBufferedDemultiplexor *saraDmltplxr = new SaraBufferedDemultiplexor();
 * 	@endcode
 * - Add a pointer to SaraPacket instance in order to demultiplex.
 * 	@code
 * 		SaraPacket *saraPacket = ....;   // Normally it is from a file or stdin.
 * 		saraDmltplxr->appendSaraPacket(saraPacket);
 * 	@endcode
 * - The appendSaraPacket() method will append SaraPacket to the demultiplexor.  
 *   The method also checks the sequence of the packets, contents inside, and so on.
 *   And if demultiplexor find completion of one CenaPacket or one SwimPacket,
 *   these packets are buffered internally.
 * - You can check the number of buffered CenaPacket and SwimPacket by using methods of
 *   getNrCena() and getNrSwim() respectively.
 * 	@code
 * 		std::cout << saraDmltplxr->getNrCena() << std::endl;
 * 	@endcode
 * - To obtain the first (=oldest) packet, you can use getCena() and getSwim() methods.
 *   If no buffered packets are there, NULL is returned.
 *   Note that the returned instance should be deleted explicitly, otherwise memory leak must happen.
 * 	@code
 * 		CenaPacket* cenaPac = saraDmltplxr->getCena();
 * 		cenaPac->dumpAll();   // You can place any commands as you want.
 * 		...
 * 		delete cenaPac;       // You must explicitly release memory.
 * 	@endcode
 * - To obtain only the last (=newest) packet, you can use getLatestCena() and getLatestSwim() methods.
 *   All the other packets are flushed by firing these methods.
 *   If no buffered packets are there, NULL is returned.
 *   You must release memory explicitly as well.
 * 	@code
 * 		std::cout << saraDmltplxr->getNrCena() << std::endl; // Assuming 5 buffered packet.
 * 		CenaPacket* cenaPac = saraDmltplxr->getLatestCena();
 * 		std::cout << saraDmltplxr->getNrCena() << std::endl; // Should be 0 here.
 * 		...
 * 		delete cenaPac;       // You must explicitly release memory.
 * 	@endcode
 * @subsection SampleCodeDemultiplex Sample code
 * Sample file can be found in IrfPac/SaraPac/AscFmt/SaraBufferedDemultiplexor_sample.cpp
 *
 * To compile:
 * 	@code
 * 		g++ -o SaraBufferedDemultiplexor_sample \
 * 			SaraBufferedDemultiplexor_sample.cpp \
 * 			-I${HOME}/SARA/include -L${HOME}/SARA/lib \
 * 			-lascfmt -lsarapac -lirfpac
 * 	@endcode
 * 
 * To run:
 * 	@code
 * 		./SaraBufferedDemultiplexor_sample filename
 * 	@endcode
 * Here filename is the SARA-DPU output dump file.
 *
 * @include IrfPac/SaraPac/AscFmt/SaraBufferedDemultiplexor_sample.cpp
 * 
 */
int main(int argc, char** argv){

	cout << "This is a sample file of using " <<
			"SaraBufferedDemultiplexor class" << endl;
	
	if(argc==1){
		cerr << "USAGE:  " << argv[0] << "  filename" << endl;
		exit(-1);
	}

	// Specify a new input file of SARA-DPU dump format.
	SaraAscFile *asc = new SaraAscFile(argv[1]);
	asc->setTimeout(100000);  // 100 sec wait after the data end.
	
	// Instance the SaraBufferedDemultiplexor
	SaraBufferedDemultiplexor* demul=new SaraBufferedDemultiplexor();
	
	int i=0, iCena=0, iSwim=0;
	
	// Open the stream of input file.
	if(asc->openStream()){
		
		// Check the input stream is not eof.
		while(! asc->eof() ){
			
			// Obtain SaraPacket from the input file
			SaraPacket* saraPac = asc->getSaraPacket();
			
			// Check if SaraPacket can be read
			if(saraPac != NULL){
				cout << "---" << endl;
				fprintf(stdout, "SARA PAC ARRIVED: %d : %04x : %d\n",
					i, saraPac->getPacketId(),
					saraPac->getCounter());
				
				// SaraPacket will be appended to Demultiplexor.
				demul->appendSaraPacket(saraPac);

				fprintf(stdout, "Appended SaraPacket \n");
				
				// You can know how many CenaPacket and SwimPacket are buffered.
				fprintf(stdout, " - CENA Buffered Packet = %d\n", demul->getNrCena());
				fprintf(stdout, " - SWIM Buffered Packet = %d\n", demul->getNrSwim());
			
				// If CenaPacket is buffered
				if(demul->getNrCena() > 0){
					
					// You can get the first (oldest) CenaPacket.
					// You have to delete cena after you use it.
					CenaPacket* cena=demul->getCena();
					
					/////////////////////////////////////////////
					//  HERE YOU CAN PUT CENA PACKET PROCESSING.
					//  cena->dumpAll();
					/////////////////////////////////////////////
					
					cout << "CENA PACKET ARRIVED " << iCena << endl;
					iCena++;
					
					// You must delete cena instance.
					delete cena;
				}
				
				if(demul->getNrSwim() > 0){
					// You can get the first (oldest) SwimPacket.
					// You have to delete swim after you use it.
					SwimPacket* swim=demul->getSwim();
					
					/////////////////////////////////////////////
					//  HERE YOU CAN PUT SWIM PACKET PROCESSING.
					//  swim->dumpAll();
					/////////////////////////////////////////////
			
					cout << "SWIM PACKET ARRIVED " << iSwim << endl;
					iSwim++;
					
					// Delete swim instance
					delete swim;
				}
				
			}
			usleep(10000);  // Wait 10 msec between reading SaraPacket
		}	
	}
	else{  // Opening the input file failed.
		cerr << "Failed to open " << argv[1] << endl;
		exit(-2);
	}

	// Deleting file and demultiplexor.
	delete asc;
	delete demul;


	return 0;
}
