/**@file SaraSimpleDemul.cpp
 * @brief A command line tool to demultiplex BMU formatted SARA data to SWIM/CENA bmu format.

   Usage:   SaraSimpleDemul [options]  SRAxxx.bmu
   Options:
	-s swimname     Specify the output file of SWIM data
	-c cenaname	Specify the output file of CENA data
	-t timefile     Specify the time file
	-j		If it is set, intellectual jump detection is enabled.
 */

#define VERSION 4

#include<iostream>
#include <cstdlib>
#include<libgen.h>
#include<sys/stat.h>
#include<getopt.h>
#include<cmath>

#include<IrfPac/irfpac.h>
#include<IrfPac/SaraPac/sarapac.h>
#include<IrfPac/SaraPac/AscFmt/ascfmt.h>


#include<typeinfo>

using namespace std;
using namespace IrfPac;
using namespace IrfPac::SaraPac;
using namespace IrfPac::SaraPac::AscFmt;

#define ID "$Id: SaraSimpleDemulApp.cpp 1302 2009-03-30 15:25:44Z futaana $"

#define DEBUG

std::string utc2str(double utc){
	char buf[256];

	time_t utc_i = (time_t) (utc+0.5);
	struct tm* tm=gmtime(&utc_i);
	sprintf(buf, "%04d-%02d-%02dT%02d:%02d:%02dZ", tm->tm_year+1900, tm->tm_mon+1, tm->tm_mday, tm->tm_hour, tm->tm_min, tm->tm_sec);

	std::string str(buf);
	return str;
}

int main(int argc, char** argv){

	int result;
	TimeConversion* tconv=NULL;
	TimeConversion* tconv2=NULL;
	char swimout[256]="";
	char cenaout[256]="";
	char conv[256]="None";

	double cena_dt_acceptance=0.150;  // in sec
	double swim_dt_acceptance=0.300;

	bool notimefile=true;

	bool intellect=false;

        while((result = getopt(argc, argv, "jt:s:c:"))!=-1){
                switch(result){
                case 't':
		{
			FILE* tfp=NULL;
			char timefilename[256]="";
			TimeFile tf;
                        fprintf(stderr, "Option -t is set.\n");
                        fprintf(stderr, "Using the time file: %s\n", optarg);
                        sprintf(timefilename, "%s", optarg);
                        tfp=fopen(optarg, "rb");
                        if(tfp==NULL){
                                fprintf(stderr, "Failed to open the time file %s\n", optarg);
                                fprintf(stderr, "Aborting...\n");
                                exit(-3);
                        }
                        tf.readFile(tfp);
			tconv=new TimeConversionTimeFile(&tf);
                        tconv2=new TimeConversionTimeFileExtpol(&tf);
			sprintf(conv, "TIME_FILE_EXTPOL %s ", timefilename);
			notimefile=false;
                        break;
		}
		case 'c':
		{
			fprintf(stderr, "Option -c is set.\n");
			fprintf(stderr, "Using the CENA data: %s\n", optarg);
			sprintf(cenaout, "%s", optarg);
			break;
		}
		case 's':
		{
			fprintf(stderr, "Option -s is set.\n");
			fprintf(stderr, "Using the SWIM data: %s\n", optarg);
			sprintf(swimout, "%s", optarg);
			break;
		}
		case 'j':
		{
			fprintf(stderr, "Option -j is set.\n");
			intellect=true;
			break;
		}

                }
        }

	if(argc==optind){
		fprintf(stderr, "USAGE:   %s  [-t timefile  -c cena_bmu_file -s swim_bmu_file]  bmufile\n", argv[0]);
		exit(-1);
	}

	char inname[256];
	sprintf(inname, "%s", argv[optind]);

	if(strlen(swimout)==0){
		sprintf(swimout, "%s.swim", inname);
	}
	if(strlen(cenaout)==0){
		sprintf(cenaout, "%s.cena", inname);
	}
	fprintf(stderr, "CENA -> %s\n", cenaout);
	fprintf(stderr, "SWIM -> %s\n", swimout);

	SaraAscFile asc(inname);
	SaraBufferedDemultiplexor demul;
	
	PacketBmuFormatter bmuformatter;

	FILE* fpc=fopen(cenaout, "w");
	if(fpc==NULL){
		fprintf(stderr, "FATAL:  Failed to open CENA output file: %s\n", cenaout);
		exit(-22);
	}

	FILE* fps=fopen(swimout, "w");
	if(fps==NULL){
		fprintf(stderr, "FATAL:  Failed to open SWIM output file: %s\n", swimout);
		exit(-24);
	}

	if(intellect){
		sprintf(conv, "%s+JUMP", conv);
	}

	struct stat s;
	stat(inname, &s);
	int fsize=(int)s.st_size;

	fprintf(fpc, "### FORMAT VERSION: %d\n", VERSION);
	fprintf(fpc, "### PROGRAM: %s\n", ID);
	fprintf(fpc, "### ORIGINAL BMU FILE:  %s\n", inname);
	fprintf(fpc, "### ORIGINAL BMU SIZE:  %d\n", fsize);
	fprintf(fpc, "### TIME CONVERSION:  %s\n", conv);

	fprintf(fps, "### FORMAT VERSION: %d\n", VERSION);
	fprintf(fps, "### PROGRAM: %s\n", ID);
	fprintf(fps, "### ORIGINAL BMU FILE:  %s\n", inname);
	fprintf(fps, "### ORIGINAL BMU SIZE:  %d\n", fsize);
	fprintf(fps, "### TIME CONVERSION:  %s\n", conv);

	if(tconv!=NULL){  // If time conversion is available (i.e. using time file), the time file coverage is shown.
	  if( typeid( *tconv ) == typeid(TimeConversionTimeFile)){
	    TimeConversionTimeFile* tctf=dynamic_cast<TimeConversionTimeFile*> (tconv);
	    fprintf(fpc, "### TIMEFILE START %f (%s)\n", tctf->getStartTime(), utc2str(tctf->getStartTime()).c_str());
	    fprintf(fpc, "### TIMEFILE STOP  %f (%s)\n", tctf->getStopTime(), utc2str(tctf->getStopTime()).c_str());

	    fprintf(fps, "### TIMEFILE START %f (%s)\n", tctf->getStartTime(), utc2str(tctf->getStartTime()).c_str());
	    fprintf(fps, "### TIMEFILE STOP  %f (%s)\n", tctf->getStopTime(), utc2str(tctf->getStopTime()).c_str());

	    fprintf(stdout, "### TIMEFILE START %f (%s)\n", tctf->getStartTime(), utc2str(tctf->getStartTime()).c_str());
	    fprintf(stdout, "### TIMEFILE STOP  %f (%s)\n", tctf->getStopTime(), utc2str(tctf->getStopTime()).c_str());

	  }
	}

	fprintf(fpc, "\n");
	fprintf(fps, "\n");


	if (tconv==NULL){  // If time file is not available, use default time conversion.
		tconv=new TimeConversionOffset(0);
		tconv2=new TimeConversionOffset(0);
	}

	double c0utc=-1;
	double c1utc=-1;
	double s0utc=-1;
	double s1utc=-1;

	int npac=0, ncena=0, nswim=0;
	int tcena=0, tswim=0;

	CenaPacketArray cenaArr;
	SwimPacketArray swimArr;


	int cenacntr=-1;
	double cenatime=-1;

	int swimcntr=-1;
	double swimtime=-1;

	if(asc.openStream()){

		while(! asc.eof() ){
			SaraPacket* saraPac = asc.getSaraPacket();
			if(saraPac != NULL){
				npac++;
				demul.appendSaraPacket(saraPac);

				if(demul.getNrCena() > 0){
					ncena++;
					CenaPacket* cena=demul.getCena();
					tcena+=cena->getSize();
					double utc=0;
					int extrapol=0;

					try{
						utc=cena->getUniversalTime(tconv);
					}
					catch(std::invalid_argument &e){
						fprintf(stderr, "TIME FILE EXCEPTION: %s\nUSE EXTRAPOLATION\n", e.what());
						utc=cena->getUniversalTime(tconv2);
						extrapol=1;
					}

					// This block is to calculate the start and stop utc.
					if(c0utc<0){
						c0utc=utc;
					}
					else{
						c1utc=utc;
					}


					if (intellect){
						// This block is to evaluate the jump of UTC.
						if(cenacntr==-1){  // If this is the first packet, previous parameter is dummy.
							cenacntr=cena->get4sCycleCntr()-1;
							cenatime=utc-4;
						}

						int currentcntr=cena->get4sCycleCntr();
						double currenttime=utc;
						bool timejump=false;

#ifdef DEBUG
						fprintf(stderr, "== (CENA) CNT: %x->%x (%d)   TIME: %f->%f (%f)\n", cenacntr, currentcntr, currentcntr-cenacntr, cenatime, currenttime, currenttime-cenatime);
#endif
						if(fabs(currenttime-cenatime) < (4+cena_dt_acceptance) * (currentcntr-cenacntr)  ){ // If ok
							cenatime=currenttime;
							cenacntr=currentcntr;  // Just update.
						}
						else if( currentcntr != cenacntr+1  ){   // Rewind
#ifdef DEBUG
							fprintf(stderr, "PACKET LOSS OR REWIND OF CENA\n");
#endif
							cenatime=currenttime;
							cenacntr=currentcntr;  // Just update.
						}
						else{  // If ng
							fprintf(stderr, "OBT TIME JUMP FOUND:  CENA_CNT=%x   UTC=%f  ", currentcntr, currenttime);
							cenatime=cenatime + 4 * (currentcntr-cenacntr);
							cenacntr=currentcntr;
							fprintf(stderr, "EXP=%f\n", cenatime);
							timejump=true;   // Flag 
							utc=cenatime;    // Instead, expected time is set to UTC.
						}
						if(timejump) cena->setQualityFlag(CenaPacket::QFLAG_TIMESTAMP_NG | CenaPacket::QFLAG_OBT_JUMP_NG);
					}

					cena->setUtc(utc);

					if(extrapol == 1)cena->setQualityFlag(CenaPacket::QFLAG_TIMESTAMP_NG | CenaPacket::QFLAG_EXTRAPOL_TIME_FILE);
					if(notimefile) cena->setQualityFlag(CenaPacket::QFLAG_TIMESTAMP_NG | CenaPacket::QFLAG_TIMEFILE_NOTFOUND);

					cenaArr.append(cena);

					delete cena;
				}

				if(demul.getNrSwim() > 0){
					nswim++;
					SwimPacket* swim=demul.getSwim();
					tswim+=swim->getSize();
					int extrapol=0;
					double utc=0;
					try{
						utc=swim->getUniversalTime(tconv);
					}
					catch(std::invalid_argument &e){
						fprintf(stderr, "TIME FILE EXCPETION: %s\nUSE EXTRAPOLATION\n", e.what());
						utc=swim->getUniversalTime(tconv2);
						extrapol=1;
					}
					
					// This block is to calculate the start and stop utc.
					if(s0utc<0){
						s0utc=utc;
					}
					else{
						s1utc=utc;
					}

					if(intellect){
						// This block is to evaluate the jump of UTC.
						if(swimcntr==-1){  // If this is the first packet, previous parameter is dummy.
						swimcntr=swim->get8sCycleCntr()-1;
							swimtime=utc-8;
						}

						int currentcntr=swim->get8sCycleCntr();
						double currenttime=utc;
						bool timejump=false;

#ifdef DEBUG
						fprintf(stderr, "== (SWIM) CNT: %x->%x (%d)   TIME: %f->%f (%f)\n", swimcntr, currentcntr, currentcntr-swimcntr, swimtime, currenttime, currenttime-swimtime);
#endif
						if(fabs(currenttime-swimtime) < (8+swim_dt_acceptance) * (currentcntr-swimcntr)  ){ // If ok
							swimtime=currenttime;
							swimcntr=currentcntr;  // Just update.
						}
						else if( currentcntr != swimcntr+1  ){   // Rewind
#ifdef DEBUG
							fprintf(stderr, "PACKET LOSS OR REWIND OF SWIM\n");
#endif
							swimtime=currenttime;
							swimcntr=currentcntr;  // Just update.
						}
						else{  // If ng
							fprintf(stderr, "OBT TIME JUMP FOUND:  SWIM_CNT=%x   UTC=%f  ", currentcntr, currenttime);
							swimtime=swimtime + 8 * (currentcntr-swimcntr);
							swimcntr=currentcntr;
							fprintf(stderr, "EXP=%f\n", swimtime);
							timejump=true;   // Flag 
							utc=swimtime;    // Instead, expected time is set to UTC.
						}
						if(timejump)swim->setQualityFlag(SwimPacket::QFLAG_TIMESTAMP_NG | SwimPacket::QFLAG_OBT_JUMP_NG);
					}

					swim->setUtc(utc);
					if(extrapol == 1)swim->setQualityFlag(SwimPacket::QFLAG_TIMESTAMP_NG | SwimPacket::QFLAG_EXTRAPOL_TIME_FILE);
					if(notimefile) swim->setQualityFlag(SwimPacket::QFLAG_TIMESTAMP_NG | SwimPacket::QFLAG_TIMEFILE_NOTFOUND);

					swimArr.append(swim);
				}

				delete saraPac;
			}
		}
	}

	size_t cenaPacSize=cenaArr.size();
	for(size_t ic=0; ic<cenaPacSize; ic++){
	  CenaPacket* pac = cenaArr.getPacket(ic);
	  double utc = pac->getUtc();
	  fprintf(fpc, "******** CENA-DATA at Time  %s %15.3f %d %d\n", utc2str(utc).c_str(),  utc+0.0005, (int) pac->getSize(), pac->getQualityFlag());
	  bmuformatter.format(fpc, pac);
	}

	size_t swimPacSize=swimArr.size();
	for(size_t ic=0; ic<swimPacSize; ic++){
	  SwimPacket* pac = swimArr.getPacket(ic);
	  double utc = pac->getUtc();
	  fprintf(fps, "******** SWIM-DATA at Time  %s %15.3f %d %d\n", utc2str(utc).c_str(),  utc+0.0005, (int) pac->getSize(), pac->getQualityFlag());
	  bmuformatter.format(fps, pac);
	}

	fprintf(fpc, "### Number of SARA packet: %d (rough expect=%7.1f)\n", npac, fsize/2319.);
	fprintf(fps, "### Number of SARA packet: %d (rough expect=%7.1f)\n", npac, fsize/2319.);

	fprintf(fpc, "### Number of CENA packet: %d\n", ncena);
	fprintf(fpc, "### Total size of CENA packet: %d [byte]\n", tcena);

	fprintf(fps, "### Number of SWIM packet: %d\n", nswim);
	fprintf(fps, "### Total size of SWIM packet: %d [byte]\n", tswim);

	fprintf(fpc, "### Start at %f (%s)\n", c0utc, utc2str(c0utc).c_str());
	fprintf(fpc, "### End at   %f (%s)\n", c1utc, utc2str(c1utc).c_str());
	fprintf(fpc, "### Duration %f [sec]\n", c1utc-c0utc);
	fprintf(fpc, "### Data rate [bps] %7.1f / %7.1f\n", 8*tcena/(c1utc-c0utc), 8*npac*896/(c1utc-c0utc));

	fprintf(fps, "### Start at %f (%s)\n", s0utc, utc2str(s0utc).c_str());
	fprintf(fps, "### End at   %f (%s)\n", s1utc, utc2str(s1utc).c_str());
	fprintf(fps, "### Duration %f [sec]\n", c1utc-c0utc);
	fprintf(fps, "### Data rate [bps] %7.1f / %7.1f\n", 8*tswim/(c1utc-c0utc), 8*npac*896/(c1utc-c0utc));

	fclose(fpc);
	fclose(fps);

}
