/**@file SaraSyncCheckApp.cpp
 * @brief A command line tool to investigate the synchronization of time.
 *
 * Usage is
 * <pre>
 *    SaraDemultiplex   [options]   input_file
 * </pre>
 * The output is input_file.cena and input_file.swim
 */

#include<iostream>
#include <cstdlib>
#include<IrfPac/irfpac.h>
#include<IrfPac/SaraPac/sarapac.h>
#include<IrfPac/SaraPac/AscFmt/ascfmt.h>

#include<cmath>

using namespace std;
using namespace IrfPac;
using namespace IrfPac::SaraPac;
using namespace IrfPac::SaraPac::AscFmt;

void showCena(CenaPacket *cena){
	static int cntr0=-1;
	static double obs0=-1;
	int cntr=cena->get4sCycleCntr();
	uint64_t markerObt=cena->getObtTime();
	unsigned long     uMarkerObt=markerObt >> 32;
	unsigned long     lMarkerObt=markerObt & ((unsigned long) 0xffffffff);
	double   dMarkerObt=cena->getMarkerObt();
	uint16_t markerDpu=cena->getDpuTime();
	double   dMarkerDpu=cena->getMarkerDpu();
	uint16_t obsDpu=cena->getEndIntegCycleTime();
	double   dObsDpu=cena->getObsTimeDpu();
	double   dObsObt=cena->getObsTimeObt();
	CenaFastHkPacket* fPac=cena->getFastHkPac();

	if(cntr0==-1){
		cntr0=cntr;
		obs0=dObsObt;
	}

	fprintf(stdout, "CenaPacket         %04x : %04x : %04x : %04lx%08lx = %15.2f : %04x = %10.2f : %04x = %10.2f : %15.2f : %15.2f : %8.2f\n", cntr, fPac->getPsyncCntr(), fPac->getBaseCntr(), uMarkerObt, lMarkerObt, dMarkerObt, markerDpu, dMarkerDpu, obsDpu, dObsDpu, dObsObt, obs0+4000.*(cntr-cntr0), dObsObt-(obs0+4000*(cntr-cntr0)));

	delete fPac;
}
void showSwim(SwimPacket *swim){
	static int cntr0=-1;
	static double obs0=-1;
	int cntr=swim->get8sCycleCntr();
	uint64_t markerObt=swim->getObtTime();
	unsigned long     uMarkerObt=markerObt >> 32;
	unsigned long     lMarkerObt=markerObt & ((unsigned long) 0xffffffff);
	double   dMarkerObt=swim->getMarkerObt();
	uint16_t markerDpu=swim->getDpuTime();
	double   dMarkerDpu=swim->getMarkerDpu();
	uint16_t obsDpu=swim->getEndIntegCycleTime();
	double   dObsDpu=swim->getObsTimeDpu();
	double   dObsObt=swim->getObsTimeObt();

	if(cntr0==-1){
		cntr0=cntr;
		obs0=dObsObt;
	}

	fprintf(stdout, "SwimPacket         %04x :             : %04lx%08lx = %15.2f : %04x = %10.2f : %04x = %10.2f : %15.2f : %15.2f : %8.2f\n", cntr, uMarkerObt, lMarkerObt, dMarkerObt, markerDpu, dMarkerDpu, obsDpu, dObsDpu, dObsObt, obs0+8000.*(cntr-cntr0), dObsObt-(obs0+8000*(cntr-cntr0)));
}

int main(int argc, char** argv){

	if(argc==1){
		cerr << "USAGE:  " << argv[0] << "  filename" << endl;
		exit(-1);
	}

	// Specify a new input file of SARA-DPU dump format.
	SaraAscFile *asc = new SaraAscFile(argv[1]);
	
	// Instance the SaraBufferedDemultiplexor
	SaraBufferedDemultiplexor* demul=new SaraBufferedDemultiplexor();

	fprintf(stdout, "PACKETTYPE  DPU#  TYPE  OBT           @Marker    DataAcqui\n");

	// Open the stream of input file.
	if(asc->openStream()){
		
		// Check the input stream is not eof.
		while(! asc->eof() ){

			
			// Obtain SaraPacket from the input file
			SaraPacket* saraPac = asc->getSaraPacket();

			// Check if SaraPacket can be read
			if(saraPac != NULL){

				// SaraPacket will be appended to Demultiplexor.
				demul->appendSaraPacket(saraPac);

				double cenaTime=HUGE_VAL, swimTime=HUGE_VAL;
				CenaPacket* cena=NULL;
				SwimPacket* swim=NULL;

				// If CenaPacket is buffered
				if(demul->getNrCena() > 0){
					cena=demul->getCena();

					if(cena!=NULL)cenaTime = cena->getDataAcquisTimeInMsec();
				}
				
				if(demul->getNrSwim() > 0){

					swim=demul->getSwim();

					if(swim!=NULL)swimTime = swim->getDataAcquisTimeInMsec();

				}

        if(cena!=NULL && swim!=NULL){
					if(cenaTime <= swimTime){
						showCena(cena);
						showSwim(swim);
					}
					else{
						showSwim(swim);
						showCena(cena);
					}
				}
				else if(cena!=NULL){
					showCena(cena);
				}
				else if(swim!=NULL){
				  showSwim(swim);
				}

				
				if(cena!=NULL)delete cena;
				if(swim!=NULL)delete swim;

				fprintf(stdout, "SaraPacket  %04x  %04x  \n", saraPac->getCounter(), saraPac->getPacketId());

				delete saraPac;
			}

		}	
	}
	else{  // Opening the input file failed.
		cerr << "Failed to open " << argv[1] << endl;
		exit(-2);
	}

	// Deleting file and demultiplexor.
	delete asc;
	delete demul; 

	return 0;
}
