#ifndef __IRFPAC_SARAPAC_ASCFMT_SWIMBMUREADER_H
#define __IRFPAC_SARAPAC_ASCFMT_SWIMBMUREADER_H

#include<iostream>
#include<stdexcept>
#include<strings.h>

#include<IrfPac/irfpac.h>
#include<IrfPac/SaraPac/sarapac.h>

#include<vector>

namespace IrfPac{
namespace SaraPac{
namespace AscFmt{

/**@brief BMU file reader for SWIM.

   BMU file is generated by SaraSimpleDemulApp.  This is a start point of
   high-level analysis.

   The usage is simple:
   <pre>
   SwimBmuReader* bmu = new SwimBmuReader();
   bmu->readFile("/path/to/bmu/swim_xxx.bmu");
   SwimPacketArray *arr = getArray();  // Here you can get the SwimPacketArray
   delete bmu
   </pre>
 
   I'm not sure what happens if you read multiple data, especially the same BMU file.
 */
class SwimBmuReader {
private:
	static const int BUF=256;
	static const int SUPPORTED_VERSION=4;
	std::vector<SwimPacket*> _list;

public:
        ~SwimBmuReader(){
                std::vector<SwimPacket*>::iterator it=_list.begin();
                while(it!=_list.end()){
                        SwimPacket* pac=*it;
                        delete pac;
                        it++;
                }
        }

	/*@brief Read the file from fname converting to the SwimPacket, save it.

	  runtime_error will be thrown when exception happens
	 */
	void readFile(char* fname){
		fprintf(stderr, "### SWIM BMU FILE: %s\n", fname);
		char buffer[BUF];
		
		FILE* fp=fopen(fname, "r");
		if(fp==NULL){
			fprintf(stderr, "FILE NOT FOUND.");
			sprintf(buffer, "File not found: %s", fname);
			throw std::runtime_error(buffer);
		}

		// First line is version.
		fgets(buffer, BUF-1, fp);
#if 0
		std::cout  << buffer << std::endl;
#endif
		char* pos=index(buffer, ':');
		if(pos==NULL){
			fprintf(stderr, "VERSION INFO NOT FOUND.");
			sprintf(buffer, "Version info not found: %s", fname);
			throw std::runtime_error(buffer);
		}

		int ver;
		pos++;
//		std::cout << pos << std::endl;
		sscanf(pos, "%d", &ver);
#if 0
		std::cout << "VER=  " << ver << std::endl;
#endif

		if(ver < SUPPORTED_VERSION){
			fprintf(stderr, "VERSION %d NOT SUPPORTED (>= %d).\n", ver, SUPPORTED_VERSION);
			fprintf(stderr, "You may need to run SaraSimpleDemulApp again.\n");
			sprintf(buffer, "VERSION %d NOT SUPPORTED (>= %d).", ver, SUPPORTED_VERSION);
			throw std::runtime_error(buffer);
		}

		uint8_t* raw=NULL;
		double epoch=0;
		int pacsize=0;  // Actual size
		int size=0;  // Expected size
		int qflag=0; // Quality flag

		while(!feof(fp)){
			char* cc=fgets(buffer, BUF-1, fp);
#if 0
			std::cout << buffer << " " << size << " " << pacsize << std::endl;
#endif

			if(cc==NULL)break;

			if(buffer[0] == '#')continue;

			if(buffer[0] == '*'){  // First line
#if 0
				std::cout << "START OF PACKET" << std::endl;
#endif
				char* posZ=index(buffer, 'Z');
				if(posZ==NULL){
					raw=NULL;
					continue;
				}
				posZ++;
				sscanf(posZ, "%lf %d %d", &epoch, &size, &qflag);
				pacsize=0;

#if 0
				std::cout << epoch << std::endl;
				std::cout << size << std::endl;
#endif

				raw=new uint8_t[size*2+32];  // Of course it is too much but to avoid unexpected feature, larger size is added.
			}
			else if(strlen(buffer)<=1 || size<pacsize){  // Last line
				if(raw==NULL)continue;
#if 0
				std::cout << "END OF PACKET" << std::endl;
#endif

				if(size<pacsize){
					SwimPacket* spac=new SwimPacket(raw, size);
					spac->setUtc(epoch);
					spac->setQualityFlag(qflag);
//					spac->dumpAll(stdout);
					_list.push_back(spac);
				}

				if(raw!=NULL)delete[] raw;
				raw=NULL;
				size=pacsize=0;
				epoch=0;
			}
			else{
				if(raw==NULL)continue;
				int u[32];
				sscanf(buffer,"%2x%2x %2x%2x %2x%2x %2x%2x %2x%2x %2x%2x %2x%2x %2x%2x %2x%2x %2x%2x %2x%2x %2x%2x %2x%2x %2x%2x %2x%2x %2x%2x", u,u+1,u+2,u+3,u+4,u+5,u+6,u+7,u+8,u+9,u+10,u+11,u+12,u+13,u+14,u+15,u+16,u+17,u+18,u+19,u+20,u+21,u+22,u+23,u+24,u+25,u+26,u+27,u+28,u+29,u+30,u+31);
				for(int i=0; i<32; i++){
					raw[i+pacsize]=(uint8_t)u[i];
				}
				pacsize+=32;
			}
			
			
		}

		fprintf(stderr, "### FILE END: SwimPacket read : %d\n" , (int)_list.size());
		fclose(fp);
	}

	/**@brief Returns the read SwimPacket.
	 */
	SwimPacketArray* getArray(){
		SwimPacketArray* arr=new SwimPacketArray();
		std::vector<SwimPacket*>::iterator it=_list.begin();
		while(it!=_list.end()){
			SwimPacket* pac=*it;
			arr->append(pac);
			it++;
		}
		
		
		return arr;
	}

};


}}}

#endif
