#ifndef _IRFPAC_SARAPAC_ASCFMT_TIMEFILE_H
#define _IRFPAC_SARAPAC_ASCFMT_TIMEFILE_H

#include<list>
#include<map>
#include<stdexcept>

#include<IrfPac/irfpac.h>
#include<IrfPac/SaraPac/sarapac.h>

namespace IrfPac{
namespace SaraPac{
namespace AscFmt{

/**@brief Time file reader.

	This class corresponds to the time file.
	The time file is provided by project.
	The time file describes the conversion from OBT to UTC.
	OBT is in the unit of milliseconds.
	getObt2Utc() will return the map from OBT to UTC.

	OBT has a roll-over with 21 days (=21*86400e3).
	Detection of OBT roll-over is currently done by comparing the first and last OBT values.
	When roll-over is detected, the rollover flag is set, and users can know by a method of isRollover().
	In this case, the OBT value (key) of the map returned by getObt2Utc() is incremented by 21 days from the original contents.

	Therefore, users have to check isRollover(), and if it is true, you have to add 21 days to the input OBT after the roll-over.
	Well, it is difficult to say in English, so please look at the implementaion.
	The implementation of TimeConversionTimeFile class will be a good example.
*/
class TimeFile {
private:
	std::list<TimePacket*> _tpacs;
	bool isRollOver;
public:
	TimeFile(){
          isRollOver=false;
	};
	~TimeFile(){
		std::list<TimePacket*>::iterator it=_tpacs.begin();
		while(it != _tpacs.end() ){
			delete (*it);
			it++;
		}
	};
	void readFile(FILE* fp){
//		fprintf(stdout, "\n---\n");
		if(_tpacs.size()>0){
			throw std::runtime_error("TimeFile already loaded.  No new data read.");
		}
		while( 1 ){
			uint8_t pac[TimePacket::PACKET_SIZE+1];
			size_t ret=fread(pac, 1, TimePacket::PACKET_SIZE, fp);
//			std::cout << ret << std::endl;
			if(ret != (size_t)TimePacket::PACKET_SIZE)break;
			TimePacket* tp = new TimePacket(pac);
//			fprintf(stdout, "** %d ** %f\n", (int)_tpacs.size(), (double)(tp->getGmt()));
			_tpacs.push_back(tp);
		}
		if(getFirstObt() > getLastObt()){  // This is a case of roll-over.
			fprintf(stderr, "(%s,%d):Rollover detected\n", __FILE__, __LINE__);
			isRollOver=true;
		}
	}
	/**@brief Obtain the OBT of the first packet .
	*/
        double getFirstObt(){
                std::list<TimePacket*>::iterator it=_tpacs.begin();
                return (*it)->getObtTimeDouble();
        }
	/**@brief Obtain the OBT of the last packet .
	*/
        double getLastObt(){
                std::list<TimePacket*>::reverse_iterator it=_tpacs.rbegin();
                return (*it)->getObtTimeDouble();
        }
	/**@brief Return true if roll over is found.

	Rollover is decided by getFirstObt()>getLastObt();
	*/
	bool isRollover(){
		return isRollOver;
	}
	/**@brief Return the conversion look up table.

	   You have to delete the returned pointer explicitly.
	 */
	std::map<double, double>* getObt2Utc(){
#ifdef DEBUG
		fprintf(stdout, "ROLLOVER? %d\n", isRollover());
		fprintf(stdout, "OBT             EPO             UTC            OBT'\n");
#endif
		std::map<double, double> *obt2utc=new std::map<double,double>();
		std::list<TimePacket*>::iterator it=_tpacs.begin();
		while(it != _tpacs.end() ){
			TimePacket* tp=(*it);
			double obt=tp->getObtTimeDouble();
#ifdef DEBUG
			fprintf(stdout, "%15.3f ", tp->getObtTimeDouble());
			fprintf(stdout, "%15.3f ", tp->getGmt());
			fprintf(stdout, "%s ", epoch2str(tp->getGmt()).c_str());
#endif

			if(isRollover() && obt<getFirstObt()){
				obt+=21*86400e3;
			}

#ifdef DEBUG
			fprintf(stdout, "%15.3f ", obt);
			fprintf(stdout, "\n");
#endif

			it++;
			obt2utc->insert(std::map<double,double>::value_type(obt, tp->getGmt()));
		}
		return obt2utc;
	}

	/**@brief Show summary of the contents.
	*/
	void showSummary(FILE* fp){
		std::list<TimePacket*>::iterator it=_tpacs.begin();
		fprintf(fp, "        OBT             TIME            GMT\n");
		fprintf(fp, "START:  %15.3f   %15.3f   %s\n", (*it)->getObtTimeDouble(), (*it)->getGmt(), epoch2str((*it)->getGmt()).c_str());
		std::list<TimePacket*>::reverse_iterator itr=_tpacs.rbegin();
		fprintf(fp, "END  :  %15.3f   %15.3f   %s\n", (*itr)->getObtTimeDouble(), (*itr)->getGmt(), epoch2str((*itr)->getGmt()).c_str());
	}

	/** @brief Show contents of the time file contents.
	*/
	void showContents(FILE* fp){
		std::list<TimePacket*>::iterator it=_tpacs.begin();
		while(it != _tpacs.end() ){
			TimePacket* tp=(*it);
			fprintf(fp, "%15.3f   ", tp->getObtTimeDouble());
			fprintf(fp, "%15.3f   ", tp->getGmt());
			fprintf(fp, "%s   \n" , epoch2str(tp->getGmt()).c_str());
			it++;
		}
	}
};


/**@brief TimeConversion implementation from TimeFile with extrapolation support.

   From data on time file, the conversion is apploximated by linear function as
   <pre>
   UTC = coeff * OBT + offset
   </pre>

   The 21-day roll-over is supported.

*/
class TimeConversionTimeFileExtpol : public TimeConversion {

private:
  double coeff;
  double offset;
  bool rollover;
  double rollover_threshold;

public:
  TimeConversionTimeFileExtpol(TimeFile* tfile){
    std::map<double,double>* conv = tfile->getObt2Utc();
    rollover=tfile->isRollover();
    rollover_threshold=(tfile->getFirstObt()+tfile->getLastObt())/2;

    double x2=0, x=0, y=0, xy=0, n=0;

    std::map<double,double>::iterator it = conv->begin();
    while(it!=conv->end()){
      double xi = (*it).first;
      double yi = (*it).second;

      x+=xi;
      y+=yi;
      xy+=(xi*yi);
      x2+=(xi*xi);
      n++;

#ifdef DEBUG
      fprintf(stdout, "%e %e : %e %e %e %e %e\n", xi, yi, n, x, y, xy, x2);
#endif

      it++;
    }

    coeff = (n*xy - x*y) / (n*x2 - x*x);
    offset = (-xy*x + x2*y) / (n*x2-x*x);

    fprintf(stdout, "Instanced coefficient: %e %e\n", coeff, offset);

    if(coeff < 0 ){
      throw std::runtime_error("Coefficient negative.  Invalid time file?");
    }

    delete conv;
  }
  /**@brief Convert from OBT to UTC.

  Return the corresponding UTC (epoch) from OBT.
  */
  double obt2utc(double obt){
    if(rollover && obt<rollover_threshold) obt+=21*86400e3;
    return coeff*obt+offset;
  }


};

/**@brief TimeConversion implementation from TimeFile.

  The 21-day roll over is now supported.
 */
class TimeConversionTimeFile : public TimeConversion {
private:
	std::map<double,double>* _obt2utc;
	bool rollover;  // True if roll over is detected.
	double rollover_threshold;   // Threshold to determine before or after the rollover.
	double t0, t1;

public:
	TimeConversionTimeFile(TimeFile* tfile){
		_obt2utc=tfile->getObt2Utc();
		rollover=tfile->isRollover();
		rollover_threshold=(tfile->getFirstObt()+tfile->getLastObt())/2;
		t0=obt2utc(tfile->getFirstObt());
		t1=obt2utc(tfile->getLastObt());
	}
	~TimeConversionTimeFile(){
		delete _obt2utc;
	}
	double getStartTime(){
	  return t0;
	}
	double getStopTime(){
	  return t1;
	}
	double obt2utc(double obt_in){
#ifdef DEBUG
		std::cout << "SIZE OF MAP: " << _obt2utc->size() << std::endl;
#endif

		double obt=obt_in;
		if(rollover){
			if(obt<rollover_threshold){
				obt+=21*86400e3;
			}
		}

		if(_obt2utc->count(obt) == 1){  // If specified obt is the same as the times file time.
			return (*_obt2utc)[obt];
		}
		else {
			std::map<double,double>::iterator ub = _obt2utc->upper_bound(obt);
			std::map<double,double>::iterator lb = _obt2utc->lower_bound(obt);

			if(lb == _obt2utc->begin() || ub == _obt2utc->end()){
				std::string mesg="**** Specified OBT is out of range in the specified TimeFile.";
				fprintf(stderr, "**** Specified OBT (%15.4f) is out of range in the specified TimeFile.\n", obt);
				throw std::invalid_argument(mesg);
			}
			else{
				lb--;
				double x0 = lb->first;
				double x1 = ub->first;
				double y0 = lb->second;
				double y1 = ub->second;
				double yobt= y0+(obt-x0)*(y1-y0)/(x1-x0);
#if DEBUG
				fprintf(stdout, "OBT:UTC = %15.3f %15.3f : %15.3f\n", obt, obt_in, yobt);
#endif
				return yobt;
			}
			
		}
		std::string mesg="**** Specified OBT is out of range in the specified TimeFile... UNEXPECTED REACH";
		throw std::invalid_argument(mesg);
	}
};


}}}
#endif
