/*
 *  CenaAccumSclMatx.h
 *  libsadppac
 *
 *  Created by Futaana Yoshifumi on 07/06/26.
 *  Copyright 2007 __MyCompanyName__. All rights reserved.
 *
 */

#include <IrfPac/irfpac.h>

namespace IrfPac {
namespace SaraPac {

/** @brief CENA accumulation scale matrix

	CENA accumulation scale matrix are counters for 3 channels
	corresponding to energy and phase values.
	The data consists of n_e * n_p * 4 array of uint8_t type.
	Each uint8_t type data (so-called raw value) will be decompressed
	by liblinlog library into 32 bit (uint32_t type).
	
	Normally this accumulation scale matrix is not instanced directory.
	The CenaPacket class will create the instance by using CenaPacket#getAccumSclMatx() method.
	
	@see liblinlog libarary.
*/
class CenaAccumSclMatx : public PacketU8Arr {
public:
//	CenaAccumSclMatx(uint8_t* stream, size_t size) : PacketU8Arr (stream, size) {};
	/** @brief Instance CENA accumulation scale matrix.

	*/
	CenaAccumSclMatx(uint8_t* stream, int n_e, int n_p) : PacketU8Arr(stream, n_e*n_p*4) {
		_ne=n_e;
		_np=n_p;
	}
	
	~CenaAccumSclMatx() {}

	/**	@brief Returns the raw value of accumulation scaling matrix for specified energy, phase,
			and counter.
			
		The value is transfered by 8 bit (0-255), which is
		32-bit log-compressed value.  This method returns raw value (0-255).

		(For developer) The data stream will be re-structured to be [E, P, C], where
		E is the number of energy, P is the number of phase, and C is the number of counter
		(4 in the data stream while the real number is 3: the last index is just a padding).
		
		
	*/
	uint8_t getRaw(int i_energy, int i_phase, int i_counter);

	/** @brief Returns true if the accumulation scaling matrix has data.
	 */
	bool hasData(){
		return !(_ne==0 && _np==0);
	}
	
	/** @brief Returns the value of accumulation scaling matrix for specified energy, phase, and counter.
	
		The value treansfered is log-compressed from 32 bit to 8 bit.
		@see also getRawValue(i_energy, i_phase, i_counter);
	*/
	uint32_t get(int i_energy, int i_phase, int i_counter);

	uint32_t collapseEnergyRaw(int i_phase, int i_counter){
		if(!hasData()){
#ifdef DEBUG
			fprintf(stderr, "No data for collapse (e)\n");
#endif
			return 0;
		}
		uint32_t cnt=0;
		for(int ie=0; ie<_ne; ie++){
			cnt+=getRaw(ie, i_phase, i_counter);
		}
		return cnt;
	}
	uint32_t collapsePhaseRaw(int i_energy, int i_counter){
		if(!hasData()){
#ifdef DEBUG
			fprintf(stderr, "No data for collapse (p)\n");
#endif
			return 0;
		}
		uint32_t cnt=0;
		for(int ip=0; ip<_np; ip++){
			cnt+=getRaw(i_energy, ip, i_counter);
		}
		return cnt;
	}
	uint32_t collapseAllRaw(int counter){
		if(!hasData()){
#ifdef DEBUG
			fprintf(stderr, "No data for collapse (a)\n");
#endif
			return 0;
		}
		uint32_t cnt=0;
		for(int ie=0; ie<_ne; ie++){
			for(int ip=0; ip<_np; ip++){
				cnt+=getRaw(ie, ip, counter);
			}
		}
		return cnt;
	}

	static const int START_COUNTER=0;
	static const int STOP_COUNTER=1;
	static const int COIN_COUNTER=2;
	static const int PADDING=3;

	static const char* getCounterName(int index){
		if(index==0){
			return "START COUNTER";
		}
		else if(index==1){
			return "STOP COUNTER";
		}
		else if(index==2){
			return "COIN COUNTER";
		}
		else if(index==3){
			return "PADDING";
		}
		throw "Out of index for CENA scaling counter.";
	}

	void dumpAll(FILE* fp){
		fprintf(fp, "Nene = %d\n", _ne);
		fprintf(fp, "Nphs = %d\n", _np);

		for(int ic=0; ic<3; ic++){
			fprintf(fp, "yyy %s\n", CenaAccumSclMatx::getCounterName(ic));
			fprintf(fp, "EE   PP| ");
			for(int id=0; id<_np; id++){
				fprintf(fp, "%4d ", id);
			}
			fprintf(fp, "| TOTAL\n");
			fprintf(fp, "----------------------------------------------------------\n");
			fprintf(fp, "SUM    | ");
			for(int id=0; id<_np; id++){
				fprintf(fp, "%4d ", collapseEnergyRaw(id, ic));
			}
			fprintf(fp, "| %4d\n", collapseAllRaw(ic));
			fprintf(fp, "----------------------------------------------------------\n");
			for(int ie=0; ie<_ne; ie++){
				fprintf(fp, "%4d   | ", ie);
				for(int id=0; id<_np; id++){
					fprintf(fp, "%4d ", getRaw(ie,id,ic));
				}
				fprintf(fp, "| %4d\n", collapsePhaseRaw(ie, ic));
			}
		}


#if 0
		fprintf(fp, "EN PH | START | STOP  | COIN  | PADDN \n");
		fprintf(fp, "------+-------+-------+-------+-------\n");
		for(int ie=0; ie<_ne; ie++){
			for(int id=0; id<_np; id++){
				fprintf(fp, "%2d %2d | %5d | %5d | %5d | %5d \n", 
					ie, id, getRaw(ie,id,START_COUNTER), getRaw(ie,id,STOP_COUNTER),
					getRaw(ie,id,COIN_COUNTER), getRaw(ie, id, PADDING));
			}
		}
#endif

		fprintf(fp, "..... HEX DUMP\n");
		dumpHexStream(fp);
	
	}

private:
		int _ne;
		int _np;



};


}
}

