#ifndef IRFPAC_SARAPACCENAFASTHKPACKET_H
#define IRFPAC_SARAPACCENAFASTHKPACKET_H

#include <IrfPac/irfpac.h>
#include <IrfPac/SaraPac/SaraPacket.h>

namespace IrfPac {

namespace SaraPac {

	/**	@brief Implementation of CENA HK packet in CENA science packet.

		The definition of the packet contents is described
		in Table 19 in CHA-SARA-DS-0006.

	*/
        class CenaFastHkPacket : public PacketU8Arr
        {
        public:
	        CenaFastHkPacket(uint8_t* stream, size_t size);
	        CenaFastHkPacket(uint8_t* stream);

	        ~CenaFastHkPacket();

		/** @brief Dump a detailed information into stdout.
		*/
		void dumpAll(void){dumpAll(stdout);};
		/** @brief Dump a detailed information into specified file.
		*/
		void dumpAll(FILE* fp);

		/** @brief return Flag P */
		bool getFlagHvMain(){return getBitFlag(0);};
		/** @brief return Flag S */
		bool getFlagHvSafety(){return getBitFlag(1);};
		/** @brief return Flag PE */
		bool getFlagHvOnEna(){return getBitFlag(4);};
		/** @brief return Flag SE */
		bool getFlagHvLvlEna(){return getBitFlag(5);};
		/** @brief return Flag ER */
		bool getFlagError(){return getBitFlag(7);};

		uint16_t getHvMain();//!< DV2-3, H1Level
		uint16_t getHvStartMcp(); //!< DV4-5, H2Level
		uint16_t getHvStopMcp();  //!< DV6-7, H3Level
		uint16_t getHvTof();      //!< DV8-9, H4Level
		uint16_t getHvDef();      //!< DV10-11, H5Level
		uint16_t getTempIFE();    //!< DV12-13, TempIFE
		uint16_t getTempHVPS();   //!< DV14-15,TempHVPS

		uint8_t getSv1Table();   //!< DV16[3:0], SV1_TABLE, SV_WAVE1
		uint8_t getSv2Table();   //!< DV16[7:4], SV2_TABLE, SV_WAVE2A
		uint8_t getSv3Table();   //!< DV17[3:0], SV3_TABLE, SV_WAVE2B
		uint8_t getSv4Table();   //!< DV17[7:4], SV4_TABLE, SV_LENS

		/** @brief An accessor of SV table of specified SV names.

		@param SvName either SV_WAVE1, SV_WAVE2A, SV_WAVE2B, or SV_LENS.
		@sa getSv1Table, getSv2Table, getSv3Table, getSv4Table
		*/
		uint8_t getSvTable(int SvName);
 

		uint16_t getPsyncCntr();  //!< DV34-35, Number of sync pulses received from DPU since IFE power on
		uint16_t getBaseCntr();   //!< DV36-37, Number of sync pulse intervals clapsed since IFE power on

		uint8_t getObsModeRaw();	//!< DV18
		uint8_t getObsModeSync();  //!< DV18, B7.  0 is synchronous mode, 1 is asynchronous mode.
		uint8_t getObsModeSensor(); //!< DV18 B1 and B0.  00 is coincident, 01 is counter, 10 is enginerring and 11 is not defined.
		///////// Below (DV16-) is not yet implemented.
		///////// Implementation would also be needed to swig interface file

		uint16_t getTotalStartCounts();  //!< DV102-103
		uint16_t getTotalStopCounts();  //!< DV104-105
		uint16_t getTotalCoinCounts();  //!< DV106-107

		/** @brief get SV Monitor of specified Name and level.

		@param SvName either SV_WAVE1, SV_WAVE2A, SV_WAVE2B, or SV_LENS.
		@param level [1,8]
		*/
		uint16_t getSvMonitor(int SvName, int level);

		static const int SV_WAVE1=0;
		static const int SV_WAVE2A=1;
		static const int SV_WAVE2B=2;
		static const int SV_LENS=3;

#ifdef DEBUG
		static uint16_t test12b(uint8_t u, uint8_t l){
		  return get12bValue(u,l);
		}
		static bool testbit(uint8_t val, size_t pos){
		  return getBitAt(val,pos);
		}

#endif

	private:
		static uint16_t get12bValue(uint8_t upper, uint8_t lower){
		  uint16_t val=0;
		  val=val|(((uint16_t)upper & 0x000f) << 8);
		  val=val|(((uint16_t)lower & 0x00ff));
		  return val;
		}

		static uint16_t get16bValue(uint8_t upper, uint8_t lower){
		  uint16_t val=0;
		  val=val|(((uint16_t)upper & 0x00ff) << 8);
		  val=val|(((uint16_t)lower & 0x00ff));
		  return val;
		}

		/**@brief Returns 12 bit value constitute of idx0-th and idx1-th bytes.
		 */
		uint16_t get12bitValue(int idx0, int idx1){
		  uint8_t u = IrfPac::SaraPac::getByteAt(*this, (size_t)idx0);
		  uint8_t l = IrfPac::SaraPac::getByteAt(*this, (size_t)idx1);
		  return get12bValue(u,l);
		};

		uint16_t get16bitValue(int idx0, int idx1){
		  uint8_t u = IrfPac::SaraPac::getByteAt(*this, (size_t)idx0);
		  uint8_t l = IrfPac::SaraPac::getByteAt(*this, (size_t)idx1);
		  return get16bValue(u,l);
		}

		static bool getBitAt(uint8_t value, size_t position){
		  return (value & (1 << position)) >> position;
		}// Only used from getBitFlag


		/**@brief Returns the bit flag of the DV1 */
		bool getBitFlag(int position){
		  uint8_t v=IrfPac::SaraPac::getByteAt(*this, (size_t)0);
		  return getBitAt(v, (size_t)position);
		}

        };

}

}

#endif
