/*
 *  CenaMatrix.h
 *  libsadppac
 *
 *  Created by Futaana Yoshifumi on 07/06/24.
 *  Copyright 2007 __MyCompanyName__. All rights reserved.
 *
 */

#ifndef IRFPAC_SARAPAC_CENAMATRIX
#define IRFPAC_SARAPAC_CENAMATRIX

#include<string>

#include"IrfPac/irfpac.h"
#include"IrfPac/SaraPac/CenaScienceDataPacket.h"

namespace IrfPac {
namespace SaraPac {

	/**	@brief CENA accumulation matrix class.
	
		This class is a part of CenaPacket, which contains the raw stream of
		CENA accumulation matrix.  This is either compressed or uncompressed.
		If the first 4 byte start with AA55AA55, this matrix is compressed.
		
	*/
	class CenaMatrix : public PacketU8Arr {
		public:
			CenaMatrix(uint8_t* packet, int size) ;
			~CenaMatrix() ;

			/**@brief Decompress the packet.
			
				Decompress the packet and create new CenaMatrix instance, and return the pointer to it.
				
				@exception NOT_COMPRESSED if the packet is not compressed.
				@exception ILLEGAL_PACKET if the packet is illegally formatted.
				@exception DECOMPRESSION_FAILED if the decompression failed.
			*/
			CenaMatrix* decompress(int verbose);
			
			/**@brief Debug verbosity when reporting errors during decompression
			*/
			static const int VERBOSE_DEBUG=3;
			/**@brief Normal verbosity when reporting errors during decompression
			*/
			static const int VERBOSE_NORMAL=2;
			/**@brief Minimum verbosity when reporting errors during decompression.
			*/
			static const int VERBOSE_MIN=1;
			/**@brief No verbosity when reporting errors during decompression.
			*/
			static const int VERBOSE_NONE=0;
			
			/**@brief Decompress the packet with selectable versosity == VERBOSE_NORMAL.
			*/
			CenaMatrix* decompress(){	
				return decompress(VERBOSE_NORMAL);
			}	
			
			/**@brief Exception thrown when you try to decompress a packet which is not compressed.
			*/
			static const int NOT_COMPRESSED=120;
			/**@brief Exception thrown when the decompression fails.
			*/
			static const int DECOMPRESSION_FAILED=255;
			/**@brief Exception if the packet contents is illegal.
			*/
			static const int ILLEGAL_PACKET=75;
			
			/**@brief Exception if tm mode is unknown.
			*/
			static const int UNKNOWN_TMMODE=180;
			
			/**@brief A header and descripter of compressed packet.
			*/
			static const uint32_t COMPRESSED_HEADER=0xAA55AA55;
			/**@brief A header of uncompressed packet.
			*/
			static const uint32_t NON_COMPRESSED_HEADER=0xAADDAADD;
			

			/**@brief Interpret CenaMatrix, and convert to datatype depending on TM mode.

			The method interpret CenaMatrix up to TM mode.
			The returned type is a derived class of CenaScienceDataPacket, up to TM mode as well.
			For example, you want to interpret the packet to be mass accumulation mode,
			the syntax is like
			<pre>
				CenaPacket* cenaPacket = .... ;   // If you have already CenaPacket,
				CenaMatrix* cenaMatrix = cenaPacket -> getCenaMatrix();   // Obtain CenaMatrix,
				CenaSciMassAcc* scidata = cenaMatrix -> interpret ( cenaPacket->getTMmode(), 
						cenaPacket->getNrCh(), cenaPacket->getNrEne(),
						cenaPacket->getNrPhase(), cenaPacket->getNrMass()   );
			</pre>
			*/
			CenaScienceDataPacket* interpret(uint8_t tm_mode, uint8_t nc, uint8_t ne, uint8_t np, uint8_t nm);
			CenaScienceDataPacket* interpret(){
			  return interpret(_tm_mode, _nc, _ne, _np, _nm);
			}

			void setParameters(uint8_t tm_mode, uint8_t nc, uint8_t ne, uint8_t np, uint8_t nm){
			  _tm_mode=tm_mode;
			  _nc=nc;
			  _ne=ne;
			  _np=np;
			  _nm=nm;
			}

			/**@brief Returns the science matrix size in byte (the same as the size of array).

			   Returns the size of the matrix in byte.
			   The value is the same as the size of array.
			   The size is up to the TM mode.
			   If unknown TM mode is specified, -1 is returned.
			   @todo Check non-process mode if needed.
			 */
			int getExpectedMatrixSize(uint8_t tm_mode, uint8_t nc, uint8_t ne, uint8_t np, uint8_t nm);

			int getExpectedMatrixSize(){
			  return getExpectedMatrixSize(_tm_mode, _nc, _ne, _np, _nm);
			}
			
			void dumpAll(FILE* fp){

				fprintf(fp, "..... HEX DUMP total %d byte (Human readable part will be implemented soon) .....\n", (int)this->getSize());
				dumpHexStream(fp);
	
			};

			/** @brief Mass accumulation mode.
			*/
			static const uint8_t TMMODE_MASS_ACC=0x02;    // 010
			/** @brief TOF accumulation mode.
			*/
			static const uint8_t TMMODE_TOF_ACC=0x04;     // 100
			/** @brief Count accumulation mode.
			*/
			static const uint8_t TMMODE_COUNT_ACC=0x03;   // 011
			/** @brief Non-process mode.
			*/
			static const uint8_t TMMODE_NON_PROCESS=0x01; // 001
			
			/** @brief TM mode unknown
			*/
			static const uint8_t TMMODE_UNKNWON=0xff;   

		private:
			char* preformat(int verbose);  //!< Only called from decompress() 
			uint8_t* decomp_main(char* inputStream, size_t* decompressedSize, int verbose);  //!< Only called from decompress() 

			uint8_t _tm_mode;  //  =TMMODE_UNKNOWN;
			uint8_t _nc;
			uint8_t _ne;
			uint8_t _np;
			uint8_t _nm;
			

	};

}
}

#endif

