#ifndef IRFPAC_SARAPACCENAPACKET_H
#define IRFPAC_SARAPACCENAPACKET_H

#include <string>
#include <stdexcept>
#include <vector>

#include <IrfPac/irfpac.h>

#include "IrfPac/SaraPac/TimeConversion.h"

#include "IrfPac/SaraPac/SaraPacket.h"
#include "IrfPac/SaraPac/CenaSciCntrAcc.h"
#include "IrfPac/SaraPac/CenaSciMassAcc.h"
#include "IrfPac/SaraPac/CenaSciTofAcc.h"
#include "IrfPac/SaraPac/CenaScienceDataPacket.h"
#include "IrfPac/SaraPac/CenaSubcHkPac.h"
#include "IrfPac/SaraPac/CenaFastHkPacket.h"
#include "IrfPac/SaraPac/CenaAccumSclMatx.h"
#include "IrfPac/SaraPac/CenaMatrix.h"

namespace IrfPac {

namespace SaraPac {


/**  @page CenaMode
     @brief Description of modes of CENA.

     @section Sensor_mode_and_telemetry_mode Sensor mode and telemetry mode

     Once one says CENA modes, there are two meanings:
     the sensor mode and the telemetry mode.

     The sensor mode is selectable from the list below.
     <ul>
     <li> Coincidence mode  CenaPacket::SENSOR_COINCIDENCE.
     <li> Counter mode  CenaPacket::SENSOR_COUNTER.
     <li> Engeenering mode... Unsupported yet by libsadppac.
     </ul>
     The sensor mode of the specified packet is obtained by CenaPacket::getSensorMode() or CenaPacket::getSensorModeStr().
     
     The telemetry mode is selectable from the list below.
     <ul>
     <li> Mass accumulation mode  CenaPacket::TMMODE_MASS_ACC
     <li> Tof accumulation mode  CenaPacket::TMMODE_TOF_ACC
     <li> Counter accumulation mode  CenaPacket::TMMODE_COUNT_ACC
     <li> Non process mode  CenaPacket::TMMODE_NON_PROCESS
     </ul>
     The telemetry mode of the specified packet is obtained by CenaPacket::getTMmode() or CenaPacket::getTMmodeStr().

     The combinations between the sensor mode and the TM mode is limited as the following table.
     <table>
     <tr><td>Sens\\TM</td> <td>Coin</td>      <td>Cntr</td>      <td>Engn</td>      </tr>
     <tr><td>Non Proc</td> <td><b>OK</b></td> <td><b>OK</b></td> <td><b>OK</b></td> </tr>
     <tr><td>Mass Acc</td> <td><b>OK</b></td> <td><i>NG</i></td> <td><i>NG</i></td> </tr>
     <tr><td>Cntr Acc</td> <td><i>NG</i></td> <td><b>OK</b></td> <td><i>NG</i></td> </tr>
     <tr><td>Tof Acc </td> <td><b>OK</b></td> <td><i>NG</i></td> <td><i>NG</i></td> </tr>
     </table>

     @section Notes Notes
     @subsection Scaling_matrix Scaling matrix
     The CENA has data regarding to the scaling matrix (CenaAccumSclMatx).
     This data is only available for the coincidence TM-mode.

 */

/**	@brief CENA packet.

	The packet is an implementation of CENA packet.
	Actually because the DPU packet is so complicated and the documentation is not so well structured,
	it is difficult to understand the packet contents.
	CENA packet here is a data stream constituting from stuff, HK and science data.
	The CENA packet should start from 0xAAAA5555 (const static uint32_t CENA_SCI_ID).

	Stuff is various information, including time and mode.  Method for accessing for
	these information is currently implemented at CenaPacket class.

	HK is a health check data consistute of 119 byte subpacket.
	The packet is implemented as CenaFastHkPacket class.
	
	Science data is a matrix, whose structure depends on modes.
	There are two types of matrix data.  One is accumulation scaling matrix data, which only exist
	coincidence mode (Details of mode are written in {@link CenaMode} page).   The implementation is in the CenaAccumSclMatx class.
	The other class is accumulation matrix data.
	The implementation is at the CenaMatrix class.
	The CenaMatrix class instances are sometimes compressed, and they are
	converted to a subclass of CenaScienceDataPacket (up to mode) to get data.
	<pre>
	
    |----- CenaPacket -------------------------------------------------------------------------------|
    |--- Stuff ---|--- CenaFastHkPacket ---|--- CenaAccumSclMatx ---|--- CenaMatrix -----------------|
                                                                              | (convert to)
                                                                     |--- CenaScienceDataPacket ---|
	</pre>
	
	@author  Futaana
*/
class CenaPacket : public PacketU8Arr
{
private:
	int _qflag;
	double _savedUtc;  // This is initialized by 0.  If it is set,

public:
	static const int QFLAG_NOTSET=0;
	/**@brief This flag is prepared for the time stamp NG whatever the reason is.
	*/
	static const int QFLAG_TIMESTAMP_NG=0x01;
	/**@brief This flag is prepared for the OBT jump NG.  Normally set with QFLAG_TIMESTAMP_NG.
	*/
	static const int QFLAG_OBT_JUMP_NG=0x02;
	/**@brief This flag is used to specify that the UTC conversion is done by extrapolation of time file contents.  Normally set with QFLAG_TIMESTAMP_NG. If a jump could be identified to be due to BMU-DPU time szncronization, then QFLAG_EXTRAPOL_DPU_SYNC is used instead
	*/
	static const int QFLAG_EXTRAPOL_TIME_FILE=0x04;
	static const int QFLAG_TIMEFILE_NOTFOUND=0x08;

	/**@brief This flag is used to specify that the UTC conversion is done by extrapolation of time file contents due time sync between BMU and DPU. Normally set with QFLAG_TIMESTAMP_NG.
	*/
	static const int QFLAG_EXTRAPOL_DPU_SYNC=0x10;
	/**@brief If QFLAG_EXTRAPOL_DPU_SYNC is set for more than 16 (TBC) packets this flag is added to indicate a probably dubious synthetic time interpolation. Normally set with QFLAG_TIMESTAMP_NG and QFLAG_EXTRAPOL_DPU_SYNC.
	*/
	static const int QFLAG_EXTRAPOL_DPU_SYNC_DUBIOUS=0x20;

	CenaPacket(uint8_t* stream, size_t size);

	~CenaPacket();

	/** @brief Dump a detailed information into stdout.
	*/
	void dumpAll(void){dumpAll(stdout);};
	/** @brief Dump a detailed information into specified file.
	*/
	void dumpAll(FILE* fp);
	
	/**
	   @deprecated Use identical method of printParamsTbl2(fp, NULL)
	 */
	DEPRECATED
	void printParamsTbl(FILE* fp);
	
	/** @brief Same as printParamsTbl3(fp, TimeConversion, false)
	*/
	void printParamsTbl2(FILE* fp, TimeConversion* tconv);
	
	/** @brief If useUtc == true then use interpolated time from the demultiplexed packet header instead of directly converted time from inside the packet. If false, the time is obtained from the OBT timestamp in the packet.
	*/
	void printParamsTbl3(FILE* fp, TimeConversion* tconv, bool useUtc);

	static void printParamsTblHeader(FILE* fp);

	/** @brief Identifier of the CENA science packet.
	*/
	const static uint32_t CENA_SCI_ID=0xAAAA5555;

	/** @brief Returns field of the packet for the identifier of CENA science packet.

		Return the field of identifier of the CENA science packet.
		First and second words in the memory map.
		In the DPU document, it is specified as 0xAAAA5555.
		This value is defined as CENA_SCI_ID field.
	*/
	uint32_t getHeadMagicNumber(){
		return getU32(1,2);
	};

	/** @brief Returns 48 bit BMU OBT @ marker

	    This time is the OBT time synchronized to DPU time.

	    The OBT is in so-called 1750 format.
	    The function returns the OBT as in the stream.
	    The interpretation of the value is specified in the project.
	*/
	uint64_t getObtTime(){
		return getU64(3,4,5);
	}

	/** @brief Returns BMU OBT @ marker in Msec.

	    The OBT time at marker (synchronization) is converted to the unit of msec.
	    @deprecated Use getMarkerObt()
	*/
	double getObtTimeInMsec(){
	  uint64_t obt=getObtTime();
	  return obt2msec(obt);
	}

	/** @brief Returns the marker time in OBT time format.
	*/
	double getMarkerObt(){
	  uint64_t obt=getObtTime();
	  return obt2msec(obt);
	}
	/** @brief Returns the marker time in DPU format.
	*/
	double getMarkerDpu(){
		int t = getDpuTime();
		int dt = getTcountDpu();
		double acTime = t*31.25 + dt * 2.048;
		return acTime;
	}
	/** @brief Returns the observation time in DPU format.
	*/
	double getObsTimeDpu(){
		int t = getEndIntegCycleTime();
		int dt = getTcountAcq();
		double acTime = t*31.25 + dt * 2.048;
		return acTime;
	}
	/** @brief Returns the observation time in OBT format calculated by getMarkerObt(), getMarkerDpu(), getObsTimeDpu()
	*/
	double getObsTimeObt(){
          double offset= getObsTimeDpu() - getMarkerDpu();
          if(offset<0)offset+=31.25*0x10000;
	  return getMarkerObt() + offset;
	}

	/** @brief Returns DPU time at marker (word=6)

	    DPU time at marker (when synchronized).  The unit is 31.25 msec.
	*/
	uint16_t getDpuTime(){
		return getU16(6);
	}

	/** @brief Returns Tcount value from timer (DPU??).

	    The fraction of the DPU time at marker in 2.048 msec unit.
	*/
	uint16_t getTcountDpu(){
		return getU16(7);
	}

	/** @brief Returns DPU time @ Marker in msec.

	    Actual formulation is getDpuTime()*31.25 + getTcountDpu()*2.048.
	    @deprecated Use getMarkerDpu().
	 */
	double getDpuTimeMarkerInMsec(){
		int t = getDpuTime();
		int dt = getTcountDpu();
		double acTime = t*31.25 + dt * 2.048;
		return acTime;
	}

	/** @brief (OBSOLETE) Returns Data acquisition time (word-8)

	@deprecated This method is renamed to getEndIntegCycleTime().
	*/
	uint16_t getDataAcquisTime(){
		return getU16(8);
	}

	/** @brief Returns the DPU time of the end of data acquisition (word-8)

	    Return the DPU time when the data comes.  31.25 msec unit.
	*/
	uint16_t getEndIntegCycleTime(){
		return getU16(8);
	}


	/** @brief Returns Tcount value from timer for the end of the data aquisition time.

	    The fraction part of the data acuisition time.  2.048 msec unit.
	*/
	uint16_t getTcountAcq(){
		return getU16(9);
	}

	/** @brief Return the DPU time at the end of integration cycles.

		Return the DPU time in msec of the end of integration cycles.
		The calculation is done as follows.  This may be wrong, or true.... I'm not sure.
		<pre>
		int t = getEndIntegCycleTime();
		int dt = getTcountAcq();
		double acTime = t*31.25 + dt * 2.048;
		return acTime;
		</pre>

	@deprecated Use getObsTimeDpu();
	*/
	double getDataAcquisTimeInMsec(){
		int t = getEndIntegCycleTime();
		int dt = getTcountAcq();
		double acTime = t*31.25 + dt * 2.048;
		return acTime;
	}
       
	/**@brief Return the observation time in msec with OBT time origin.  

	    Actual formulation is
	    <pre>
	    getObtTimeInMsec() + ( getDataAcquisTimeInMsec() - getDpuTimeMarkerInMsec()  )
	    </pre>

	@deprecated Use getObsTimeObt()
	 */
	double getEpochTimeInMsec(){
          double offset= getDataAcquisTimeInMsec() - getDpuTimeMarkerInMsec();
          if(offset<0)offset+=31.25*0x10000;
	  return getObtTimeInMsec() + offset;
	}

	/**@brief Return the universal time in seconds [Origin is the UNIX epoch time as time_t].

	   Return the universal time in seconds.  Note that this value is calculated from the packet contents after TimeConversion.
	 */
	double getUniversalTime(TimeConversion* tconv){
		if(tconv==NULL){
			return getObsTimeObt()/1000.;
		}
		return tconv->obt2utc(getObsTimeObt());
	}

	/** @brief Returns the Universal Time in msec.  (Exactly saying, this method returns the time after the epoch).
	    
	    Actual formulation is
	    <pre>
	    getObtTimeInMsec() + ( getDataAcquisTimeInMsec() - getDpuTimeMarkerInMsec()  )
	    </pre>

	    @deprecated Use getUniversalTime() or getEpochInMsec().  You can replace it getUniversalTime(NULL)*1000.
	 */
	DEPRECATED
	double getUniversalTimeInMsec(){
          double offset= getDataAcquisTimeInMsec() - getDpuTimeMarkerInMsec();
          if(offset<0)offset+=31.25*0x10000;
	  return getObtTimeInMsec() + offset;
	}

	/** @brief Returns 4S cycle counter (word-10)

		Returns the cyclic counter of CenaPacket.
	*/
	uint16_t get4sCycleCntr(){
		return getU16(10);
	}

	/** @brief An alias for get4sCycleCntr()
	*/
	uint16_t getPacketCounter(){
		return get4sCycleCntr();
	}

	/** @brief Binning parameter (word_11)

		Returns the binning parameters.
		- Bit 15 is the channel, where 0 is 1 ch and 1 is 7 chs.
		- Bit 14-12 is the energy, n(E)-1.
		- Bit 11-7 is the Phase, n(P)-1.
		- Bit 6-0 is the Mass, n(M)-1.
	*/
	uint16_t getBinningPrm(){
		return getU16(11);
	}

	/** @brief Number of channel (word-11/bit-15)

		Number of channel (1 or 7) is returned.
		If 11th word 15th bit is 0, 1 is returned.
		If 11th word 15th bit is 1, 7 is returned.
	*/
	uint8_t getNrCh(){
		uint16_t br=getBinningPrm();
		uint16_t mask=0x8000;
		if((br&mask)==0){
			return 1;
		} else {
			return 7;
		}
	}

	/** @brief Number of energy (word-11/bit-14,13,12)

		Number of energy is returned.
		B14-B12 : n(E)-1
	*/
	uint8_t getNrEne(){
		uint16_t br=getBinningPrm();
		uint16_t mask=0x7000;
		uint16_t ne=br&mask;
		ne = ne >> 12;
		return ne+1;
	}	

	/** @brief Number of phase (word-11/bit-11 to 7)

		Number of phase is returned.
		B11-B07 : n(P)-1
	*/
	uint8_t getNrPhase(){
		uint16_t br=getBinningPrm();
		uint16_t mask=0x0f80;
		uint16_t np=br&mask;
		np = np >> 7;
		return np+1;
	}

	/** @brief Number of mass (word-11/bit-6 to 0)

		Number of mass is returned.
		B6-B0 : n(M)-1
	*/
	uint8_t getNrMass(){
		uint16_t br=getBinningPrm();
		uint16_t mask=0x007f;
		uint16_t np=br&mask;
		np = np;
		return np+1;
	}

	/** @brief number of cycles for integration (word-12)
	*/
	uint16_t getNrCycles(){
		return getU16(12);
	}

	/** @brief modes (word-13)

		Returns the modes.  Logical bit, 1 as ON, 0 as OFF.
		- B15: SWIM ON/OFF
		- B14: CENA ON/OFF
		- B13: CENA compression
		- B12-B10: TM mode
			- 010: Non Acc Mode
			- 100: Tof Acc Mode
			- 011: Count Acc Mode
			- 001: Non Process Mode
		- B9-B8: Mode of Sensor (Not documented in the DPU doc)

		While B7-B0 is considered as first byte of HK, but this values are
		included in the returned value.  You need to disregards low 8 bits.
	*/
	uint16_t getModes(){
		return getU16(13);
	}
	
	bool isSwimOn(){
		uint16_t mode=getModes();
		uint16_t mask=0x8000;	// 15 bit
		return (mode & mask) != 0;
	}
	
	bool isCenaOn(){
		uint16_t mode=getModes();
		uint16_t mask=0x4000;	// 14 bit
		return (mode & mask) != 0;
	}

	/**@brief True if compression is on for the DPU side.  (Independent of real CenaMatrix is compressed or not)
	
		When the rice-compression was enabled on the DPU side, true is returned.
		Note that decompress() method will NOT change this status.
		@sa isCompressed()
	*/
	bool isCenaCompOn(){
		uint16_t mode=getModes();
		uint16_t mask=0x2000;	// 13 bit
		return (mode & mask) != 0;
	}

	/**@brief Get telemetry mode.

	   @sa @link CenaMode
	 */
	uint8_t getTMmode(){
		uint16_t mode=getModes();
		uint16_t mask=0x1c00;     // 12, 11, 10 bits
		uint16_t tm=mode&mask;
		tm = tm >> 10;
		return (uint8_t) tm;
	}

	std::string getTMmodeStr(){
                uint8_t mode=getTMmode();	  
  		std::string str="UNK";
		if(mode == TMMODE_MASS_ACC)str="MAS";
		if(mode == TMMODE_TOF_ACC)str="TOF";
		if(mode == TMMODE_COUNT_ACC)str="CNT";
		if(mode == TMMODE_NON_PROCESS)str="NPR";
		return str;
	}

	/**@brief Get sensor mode.

	   @sa @link CenaMode
	 */
	uint8_t getSensorMode(){
		uint16_t mode=getModes();
		uint16_t mask=0x0300;     // 9,8 bits
		uint16_t sn=mode&mask;
		sn = sn >> 8;
		return (uint8_t) sn;
	}

	std::string getSensorModeStr(){
                uint8_t smode=getSensorMode();
                std::string str="UNKW";
		if(smode==SENSOR_COINCIDENCE)str="COIN";
		if(smode==SENSOR_COUNTER)str="CONT";
                return str;
	}

	/**	@brief Get subcommuted HK packet.
	
		Get subcom HK packet, which constitute of 119 byte.
		Starting from the last byte of the word-13, to the word-72.
		This is consistent with Byte-25 to byte 143.
		
		Please delete the obtained packet after use.
		@deprecated Use getFastHkPac()
	*/
	DEPRECATED
	CenaSubcHkPacket* getSubcomHkPac();

	/**	@brief Get fast HK packet for CENA.
	
		Get fast HK packet, which constitute of 119 byte, normally every 4 sec.
		Starting from the last byte of the word-13, to the word-72.
		This is consistent with Byte-25 to byte 143.
		
		Please delete the obtained packet after use.
	*/
	CenaFastHkPacket* getFastHkPac(){
		uint8_t* str=getStream();		
		CenaFastHkPacket *pac=new CenaFastHkPacket(str+25);
		delete[] str;
		return pac;
	}


	/**	@brief Get accumulation scale matrix.
	
		The accumulation scale matrix is returned.  The length of this packet
		varies up to numbers of energy bin and phase bin.  This is only for coincidence mode.
		In other mode, 0-byte matrix is returned.		

		Please delete the obtained packet after use.
	*/
	CenaAccumSclMatx* getAccumSclMatx(){
		uint8_t mode=getTMmode();
		CenaAccumSclMatx *matx=NULL;
		
		if(mode == TMMODE_MASS_ACC || mode == TMMODE_TOF_ACC){
			int n_e=getNrEne();
			int n_p=getNrPhase();

			uint8_t* str=getStream();
			matx=new CenaAccumSclMatx(str+144, n_e, n_p);
			delete[] str;
		}
		else{
	       

		  fprintf(stdout, "NO ACCUMSCLMATX FOR THIS MODE\n");
	       
			matx=new CenaAccumSclMatx(NULL, 0, 0);   // I don't know whether it works.  Will see.
		}
		
		return matx;
	}
	
	CenaMatrix* getCenaMatrix(){
//		uint8_t mode=getTMmode();
		CenaMatrix* matx=NULL;
		
		size_t offset=offsetByteToCenaMatrix();
		
		uint16_t fullsize=this->getSize();
		uint8_t* str=getStream();
		matx=new CenaMatrix(str+offset, fullsize-offset);		

		matx->setParameters(getTMmode(), getNrCh(), getNrEne(), getNrPhase(), getNrMass());
		
		delete[] str;		
		return matx;
	
	}
	
	/** @brief Returns CENA science data of counter accumulate mode.
	
		Returns the CENA science data matrix of counter accumulate mode.
		If the packet is not counter accumulate mode, NULL is returned.
		If the packet is compressed, "Science Matrix is still Compressed" exception thrown.
	*/
	CenaSciCntrAcc* getCenaSciCntrAcc();
	
	/** @brief Returns CENA science data of TOF accumulate mode.
	
		Returns the CENA science data matrix of TOF accumulate mode.
		If the packet is not TOF accumulate mode, NULL is returned.
		If the packet is compressed, "Science Matrix is still Compressed" exception thrown.
	*/
	CenaSciTofAcc* getCenaSciTofAcc();
	
	/** @brief Returns CENA science data of MASS accumulate mode.
	
		Returns the CENA science data matrix of MASS accumulate mode.
		If the packet is not MASS accumulate mode, NULL is returned.
		If the packet is compressed, "Science Matrix is still Compressed" exception thrown.
	*/
	CenaSciMassAcc* getCenaSciMassAcc();

	/** @brief Returns CENA science data of MASS accumulation mode, but in a generic energy table.
 	*
 	* It returns the CENA mass accumulation mode data, but in a generic energy table.
 	* The generic energy table is implemented in CenaSciMassAccE16 class, which can be
 	* (at least aiming to be) treated as the same as CenaSciMassAcc mode.
 	*
 	* runtime_error is thrown if known error happens. 
 	* <ul>
 	* <li> nE!=8
 	* <li> SV table is not uniquely defined.
 	* <li> If the packet is compressed (use decompress() before).
 	* </ul>
 	*
 	* The generic energy table is composed of 16 energies.
 	* The CENA measures 8 energies of the 16, up to the SV table.
 	* The benefit of this class is that the data can be treated independent of SV table.
 	*/
	CenaSciMassAccE16* getCenaSciMassAccE16();
	
	/** @brief Returns true if the CenaMatrix is compressed.  (independent of DPU flag of CenaCompOn)
	
		Check whether the CenaMatrix is compressed.  If the CenaMatrix stream is compressed, true is
		returned.  If user explicitly decompressed by decompress() method, the returned value is changed.
	*/
	bool isMatrixCompressed(){
		CenaMatrix* cMat=getCenaMatrix();
		uint8_t* cMatB=cMat->getStream();
		delete cMat;
		
		bool isC=(cMatB[0]==0xaa && cMatB[1]==0x55 && cMatB[2]==0xaa && cMatB[3]==0x55);
						
		delete[] cMatB;

		return isC;
	}

	/**@brief Debug verbosity when reporting errors during decompression
	*/
	static const int VERBOSE_DEBUG=3;
	/**@brief Normal verbosity when reporting errors during decompression
	*/
	static const int VERBOSE_NORMAL=2;
	/**@brief Minimum verbosity when reporting errors during decompression.
	*/
	static const int VERBOSE_MIN=1;
	/**@brief No verbosity when reporting errors during decompression.
	*/
	static const int VERBOSE_NONE=0;

	
	/**@brief Returns newly created, rice-decompressed CenaPacket.
	
	Returns the newly created decompressed packet.
	If the packet is already decompressed, the copy of this packet is returned.
	If decompression failed, NULL is returned.

	The parameter defines the level of verbosity on stderr. 

	*/
	CenaPacket* decompress(int verbose);
	
	
	/**@brief Returns newly created, rice-decompressed CenaPacket with verbosity == VERBOSE_NORMAL
	*/
	CenaPacket* decompress(){	
		return decompress(VERBOSE_NORMAL);
	}	

	/** @brief Mass accumulation TM-mode.

	    @sa @link CenaMode
	*/
	static const uint8_t TMMODE_MASS_ACC=CenaMatrix::TMMODE_MASS_ACC;    // 010

	/** @brief TOF accumulation TM-mode.

	    @sa @link CenaMode
	*/
	static const uint8_t TMMODE_TOF_ACC=CenaMatrix::TMMODE_TOF_ACC;     // 100

	/** @brief Count accumulation TM-mode.

	    @sa @link CenaMode
	*/
	static const uint8_t TMMODE_COUNT_ACC=CenaMatrix::TMMODE_COUNT_ACC;   // 011

	/** @brief Non-process TM-mode.

	    @sa @link CenaMode
	*/
	static const uint8_t TMMODE_NON_PROCESS=CenaMatrix::TMMODE_NON_PROCESS; // 001

	
	/** @brief Coincidence sensor-mode.

	    @sa @link CenaMode
	*/
	static const uint8_t SENSOR_COINCIDENCE=0x00;

	/** @brief Counter sensor-mode.

	    @sa @link CenaMode
	*/
	static const uint8_t SENSOR_COUNTER    =0x01;
	
	/**@brief Check whether the packet format is valid for CenaPacket.
	
		Check items are:
		- First 4 byte is 0xaaaa5555
		- CenaMatrix starts with 0xaa55aa55 or 0xaaddaadd
		- OBT time is less than 1 year! (Actually it should be less than 21 days, I guess -2009-01-24-)

	@todo OBT time check is proper???
	*/
	bool isValid(){
		uint8_t* mat=getStream();
		bool test1=(mat[0]==0xaa && mat[1]==0xaa && mat[2]==0x55 && mat[3]==0x55);
		if(!test1)return false;
		delete[] mat;
		
		CenaMatrix* sMat=getCenaMatrix();
		uint8_t* cMatB=sMat->getStream();
		delete sMat;

		bool test21=(cMatB[0]==0xaa && cMatB[1]==0x55 && cMatB[2]==0xaa && cMatB[3]==0x55);
		bool test22=(cMatB[0]==0xaa && cMatB[1]==0xdd && cMatB[2]==0xaa && cMatB[3]==0xdd);
		if(!test21 && !test22)return false;
		
		delete[] cMatB;

		// OBT time check.  Sometimes very strange number.
		double obt=getObsTimeObt();
		double maxobt=365*86400.*1000.;

		if(obt>maxobt)return false;
		
		return true;
	}

	/** @brief Checks if there is RAW data available.
	
	        The implemented test is theoretically not 100% collision free: It tests for the presence of a 
			aa01 910x header at the location where the scaling matrix (for mass maccumulation mode) or the 
			rice compressed accumulation matrix (count accumulation mode) starts. No collision in count accumulation
			mode as in this mode the next data block has a unique header, but small probability of collision in mass
			accumulation mode because the scaling matrix has no header and might therefore contain aa01910x as data
			values. 
	*/
	bool hasRawModeData(){
		bool isRaw=(getU16(73) == 0xaa01 && ((getU16(74) == 0x9100) || (getU16(74) == 0x9101) || (getU16(74) == 0x9102)));
		
		return isRaw;
	}

	void setUtc(double utc){
		_savedUtc=utc;
	}
	/**@brief Return UTC in UNIX epoch, or return 0.
	 */
	double getUtc(){
		if(_savedUtc==0)throw std::runtime_error("UTC not calculated.  Use getUniversalTime(TimeConversion*)");
		return _savedUtc;
	}
	int getQualityFlag(){
		return _qflag;
	}
	void setQualityFlag(int qflag){
		_qflag|=qflag;
	}

	CenaPacket* newInstance(){
		uint8_t* stream=getStream();
		size_t size=getSize();
		CenaPacket* p=new CenaPacket(stream, size);
		p->_savedUtc=_savedUtc;
		p->_qflag=_qflag;
		delete[] stream;
		return p;
	}


	private:
	uint16_t getU16 ( int wordId0 ){return IrfPac::SaraPac::getWordAt(*this, wordId0);};
	uint32_t getU32 ( int wordId1 ){return IrfPac::SaraPac::getWordAt(*this, wordId1);};
	uint32_t getU32 ( int wordId0, int wordId1 ){return (getU32(wordId0) << 16)+getU32(wordId1);};
	uint64_t getU64 ( int wordId3 ){return IrfPac::SaraPac::getWordAt(*this, wordId3);};
	uint64_t getU64 ( int wordId2, int wordId3  ){return (getU64(wordId2) << 16)+getU64(wordId3);};
	uint64_t getU64 ( int wordId1, int wordId2 , int wordId3  ){
		return (getU64(wordId1) << 32)+(getU64(wordId2) << 16)+getU64(wordId3);};
	uint64_t getU64 ( int wordId0, int wordId1 , int wordId2 , int wordId3  ){
		return (getU64(wordId0) << 48) + (getU64(wordId1) << 32)+
			(getU64(wordId2) << 16)+getU64(wordId3);};

	/**@brief Returns the offset in byte to the CenaMatrix part.
	
			Sum of Stuff and HK part (144 byte) and Accumulation scaling matrix is returned.
	*/
	size_t offsetByteToCenaMatrix(){
			// Calculate offset size. for STUFF, HK and SCL_MATX
			size_t offset=144;      // 144 is the size of the STUFF and HK.
			uint8_t mode=getTMmode();
			if(mode == TMMODE_MASS_ACC || mode == TMMODE_TOF_ACC){
			   offset += getNrEne()*getNrPhase()*4;
			}  // Add size of scaling matrix
			return offset;
	}


};

/** @brief Array of CENA packet.

    Normally you do not need to create this class.
    This is returned by CenaFile::getArray().
*/
class CenaPacketArray {
private:
	std::vector<CenaPacket*> _arr;
public:
	~CenaPacketArray(){
#ifdef DEBUG
		fprintf(stderr, "## CenaPacketArray destrcutor called\n");
#endif
		destroy();
	}
	void append(CenaPacket* pac){
		_arr.push_back(pac->newInstance());
	}
	/**@brief Get the newly generated CenaPacket* of i-th packet.

		You have to delete the returned pointer explicitly.
	 */
	CenaPacket* getPacket(int i){
		return (_arr.at(i))->newInstance();
	}
	/**@brief Get the i-th packet.  Use #getPacket(int) instead.

	@deprecated Use #getPacket(int) because this method can break the contents inside the array.
	*/
	DEPRECATED
	CenaPacket* get(int i){
#if 0
		std::cout << "INDEX=" << i << std::endl;
		std::cout << "SIZE=" << _arr.size() << std::endl;
#endif
		fprintf(stderr, "%s, %d : YOU MUST NOT USE CenaPacketArray.get().\n", __FILE__, __LINE__);
		fprintf(stderr, "%s, %d : USE CenaPacketArray#getPacket() INSTEAD.\n", __FILE__, __LINE__);
		return _arr.at(i);
	}

	/**@brief Get the UTC timestamp of the i-th packet without copying the packet.
		
		This is useful when searching for a packet with a specific UTC without the need to copy each packet.
	 */
	double getPacketUtc(int i){
		return (_arr.at(i))->getUtc();
	}
	
	/**@brief Get the quality of the i-th packet without copying the packet.
		
		This is useful when searching for a packet with a specific quality without the need to copy each packet.
	 */
	int getPacketQualityFlag(int i){
		return (_arr.at(i))->getQualityFlag();
	}

	/**@brief Get the number of elements.
	 */
	size_t size(){
		return _arr.size();
	}
	/**@brief Clear the array after destroy all the packet contained.  You don't need to call it directly (ie called from Destructor).
	 */
	void destroy(){
		std::vector<CenaPacket*>::iterator it=_arr.begin();
		while(it!=_arr.end()){
			CenaPacket* pac=(*it);
			delete pac;
			it++;
		}
		_arr.clear();
	}
	
};

}

}

#endif
