/*
 *  CenaSciCntrAcc.h
 *  libsadppac
 *
 *  Created by Futaana Yoshifumi on 07/06/26.
 *  Copyright 2007 __MyCompanyName__. All rights reserved.
 *
 */

#ifndef _IRFPAC_SARAPAC_CENASCICntrAcc
#define _IRFPAC_SARAPAC_CENASCICntrAcc

#include<sstream>

#include "IrfPac/SaraPac/CenaScienceDataPacket.h"

namespace IrfPac{
namespace SaraPac{

/**	@brief CENA science packet (non-rice-compressed) of counter accumulation mode.

	Count accumulation mode is consititute of [E][P][55] matrix, where
	E is energy, P is phase (it is just a time in CENA's case), and
	counters.
	
	Counter names are as follows.
	<ul>
	<li>IDX0_START_RING_SECTOR   [4][7]
	<li>IDX0_START_RING          [4]
	<li>IDX0_START_SECTOR        [7]
	<li>IDX0_STOP_PLATE          [8]
	<li>IDX0_STOP_MESH           [1]
	<li>IDX0_COIN_SECTOR         [7]
	</ul>
	Note that libsadppac index is usually counted from 0 (zero),
	even though some document starts from 1 (one).
	This is a historical reason, but since many IRF people likes to start from zero, the index name IDX0_ is recommended.

	However, for convention and compatibility, counter names with index start one areprepared as follows.
	This is not recommended to be used.
	<ul>
	<li>START_RING_SECTOR   [4][7]
	<li>START_RING          [4]
	<li>START_SECTOR        [7]
	<li>STOP_PLATE          [8]
	<li>STOP_MESH           [1]
	<li>COIN_SECTOR         [7]
	</ul>
	
	@sa Section 10.9 on CHA-SARA-DS-0006
*/
class CenaSciCntrAcc : public CenaScienceDataPacket {

public:
	/** @brief Constructor.  Normally one do not need to call the constructor directly.
	
	The constructor sepcifying the number of energy, phase and counter.  Number of counter is 55.
	Normally no one does need to call the constructor.  CenaPacket::getCenaSciCntrAcc() will create this class.
	*/
	CenaSciCntrAcc(uint8_t *pac, int nEnergy, int nPhase, int nC);
	~CenaSciCntrAcc(){};

	/**@brief get 8-bit raw value of the specified energy, phase and counter name.

	   @param iE Energy step
	   @param iP Phase
	   @param iC Counter name.  Using macro IDX0_XXX is recommended.
	 */
	uint8_t getRaw(int iE, int iP, int iC);

	/**@brief get linlog decompressed 16-bit value of the specified energy, phase and counter name.

	   @param iE Energy step
	   @param iP Phase
	   @param iC Counter name.  Using macro IDX0_XXX is recommended.
	 */
	uint16_t get(int iE, int iP, int iC);
	
	void dumpAll(){dumpAll(stdout);};
	void dumpAll(FILE* fp);
	
	int getNrEne(){return _ne;};
	int getNrPhase(){return _np;};
	int getNrCounter(){return _nc;};

	/** @brief Returns the index of the START_RING_SECTOR.
	    @deprecated Use IDX0_START_RING_SECTOR as a replacement.  (Catch statement should be modified.)
	
	The ring is 1-4, sector is 1-7.  Returned value is (ring-1)+(sector-1)*4.
	If index violation happens, (r is 1 to 4, s is 1 to 7), "char const*" type exception is thrown.
	*/
	DEPRECATED
	static int START_RING_SECTOR(int r, int s){
		_range_ring(r);
		_range_sector(s);
		return (r-1)+(s-1)*4;		
	}

	/** @brief Return the index of START_RING_SECTOR.
	    @throw std::runtime_error is the specified ring or sector is out of range.

	    Returns the index of START_RING_SECTOR.  Ring is 0-3 and sector is 0-6.
	    Corresponding index is <pre>r + s*4</pre>.
	    If either of them are out of range, std::runtime_error is thrown.
	*/
	static int IDX0_START_RING_SECTOR(int r, int s){
		_idx0_range_ring(r);
		_idx0_range_sector(s);
		return r+s*4;
	}

	/** @brief Return the index of the START_RING counter id.
	    @deprecated Use IDX0_START_RING as a replacement.  (Catch statement should be modified.)

		RING is 1,2,3, or 4: corresponding to 28,29,30, or 31, respectively.
	*/
	DEPRECATED
	static int START_RING(int r){
		_range_ring(r);
		return 27+r;
	}

	/** @brief Return the index of the start ring counter.
	    @throw std::runtime_error if r is out of range.

	    Return the index of the start ring counter.  Ring is 0,1,2 or 3, corresponding to 28,29,30 or 31, respectively.  If the specified ring is out of range, std::runtime_error is thrown.
	 */
	static int IDX0_START_RING(int r){
		_idx0_range_ring(r);
		return 28+r;
	}

	/** @brief Return the index of the START_SECTOR counter id.
	    @deprecated Use IDX0_START_SECTOR as a replacement.  (Catch statement should be modified.)

		Sector is 1 to 7: corresponding to 32 to 38.
	*/
	DEPRECATED
	static int START_SECTOR(int s){
		_range_sector(s);
		return 31+s;
	}
	/** @brief Return the index of the START_SECTOR counter id.
	    @throw std::runtime_error if r is out of range.

	    Sector is 0 to 6, correspoinding to 32 to 38.
	*/
	static int IDX0_START_SECTOR(int s){
		_idx0_range_sector(s);
		return 32+s;
	}

	/** @brief Return the index of the STOP_PLATE counter id
	    @deprecated Use IDX0_STOP_PLATE as a replacement.  (Catch statement should be modified.)

		PLATE is 1 to 8: corresponding to 39 to 46, respectively.
	*/
	DEPRECATED
	static int STOP_PLATE(int p){
		_range_plate(p);
		return 38+p;
	}

	/** @brief Return the index of the STOP_PLATE counter id
	    @throw std::runtime_error if p is out of range.

	    PLATE is 0 to 7: corresponding to 39 to 46, respectively.
	*/
	static int IDX0_STOP_PLATE(int p){
		_idx0_range_plate(p);
		return 39+p;
	}

	/** @brief Return the index of the STOP_MESH.  Hard-coded as 47.
	    @deprecated Use IDX0_STOP_MESH as a replacement.
	*/
	DEPRECATED
	static int STOP_MESH(){
		return 47;
	}
	/** @brief Return the index of the STOP_MESH.  Hard-coded as 47.
	*/
	static int IDX0_STOP_MESH(){
		return 47;
	}

	/** @brief Return the index of the COIN_SECTOR counter id

		SECTOR is 1 to 7: corresponding to 48 to 54, respectively.
	*/
	DEPRECATED
	static int COIN_SECTOR(int s){
		_range_sector(s);
		return 47+s;
	}

	/** @brief Return the index of the COIN_SECTOR counter id
	    @throw std::runtime_error if s is out of range.

		SECTOR is 0 to 6: corresponding to 48 to 54, respectively.
	*/
	static int IDX0_COIN_SECTOR(int s){
		_idx0_range_sector(s);
		return 48+s;
	}


private:
	int _ne, _np, _nc;

	static void _range_ring(int r){  // Deprecated.  Use _idx0_range_ring
		if(r<=0 || r>=5){
			std::cerr << "RING RANGE ERROR: " << r << std::endl;
			throw "RING RANGE ERROR";
		}
	}
	static void _idx0_range_ring(int r){
		if(r<0 || r>=4){
			std::ostringstream oss;
			oss << "RING RANGE ERROR: " << r;
			std::string emesg=oss.str();
			std::cerr << emesg << std::endl;
			throw std::runtime_error(emesg);
		}
	}

	static void _range_sector(int s){  // Deprecated.  Use _idx0_range_sector
		if(s<=0 || s>=8){
			std::cerr << "SECTOR RANGE ERROR: " << s << std::endl;
			throw "SECTOR RANGE ERROR";
		}
	}
	static void _idx0_range_sector(int s){
		if(s<0 || s>=7){
			std::ostringstream oss;
			oss << "SECTOR RANGE ERROR: " << s;
			std::string emesg=oss.str();
			std::cerr << emesg << std::endl;
			throw std::runtime_error(emesg);
		}
	}

	static void _range_plate(int p){  // Deprecated.  Use _idx0_range_plate
		if(p<=0 || p>=9){
			std::cerr << "PLATE RANGE ERROR: " << p << std::endl;
			throw "PLATE RANGE ERROR";
		}
	}
	static void _idx0_range_plate(int p){
		if(p<0 || p>=8){
			std::ostringstream oss;
			oss << "PLATE RANGE ERROR: " << p;
			std::string emesg=oss.str();
			std::cerr << emesg << std::endl;
			throw std::runtime_error(emesg);
		}
	}
	
};

/**@brief Virtual class of CenaSciCntrAcc formatter.
*/

class CenaSciCntrAccFormatter {
	public:
		virtual ~CenaSciCntrAccFormatter(){};
		virtual void dump(FILE* fp, CenaSciCntrAcc* pac)=0;
};
/**@brief An implementation of CenaSciCntrAccFormatter.

	One-line formatter for energy-time diagram for start sectors.
	"nr_of_data nr_of_ene  ene0sect1  ene1sect1 .. eneNsect1 ene0sect2 ... eneNsect7"
	Maximum nr_of_data is (N=8)*7=56.  If less than maximum, -1 is filled.
*/
class CenaSciCntrAccFormatterStartNC : public CenaSciCntrAccFormatter {
	public:
		~CenaSciCntrAccFormatterStartNC(){}
		void dump(FILE* fp, CenaSciCntrAcc* pac){
			int nen=pac->getNrEne();
			int nph=pac->getNrPhase();
			fprintf(fp, "%d %d ", nen*7, nen);
			for(int isect=1; isect<=7; isect++){
				int counter=CenaSciCntrAcc::START_SECTOR(isect);
				for(int ien=0; ien<nen; ien++){
					int v=0;
					for(int iph=0; iph<nph; iph++){
						v+=pac->get(ien, iph, counter);
					}
					fprintf(fp, " %d", v);
				}
				for(int ien=nen; ien<8; ien++){
					fprintf(fp, " %d", -1);
				}
			}
		}
};
/**@brief An implementation of CenaSciCntrAccFormatter.

	One-line formatter for energy-time diagram for coincidence sectors.
        "nr_of_data nr_of_ene  ene0sect1  ene1sect1 .. eneNsect1 ene0sect2 ... eneNsect7"
        Maximum nr_of_data is (N=8)*7=56.  If less than maximum, -1 is filled.
*/
class CenaSciCntrAccFormatterCoin : public CenaSciCntrAccFormatter {
        public:
                ~CenaSciCntrAccFormatterCoin(){}
                void dump(FILE* fp, CenaSciCntrAcc* pac){
                        int nen=pac->getNrEne();
                        int nph=pac->getNrPhase();
                        fprintf(fp, "%d %d ", nen*7, nen);
                        for(int isect=1; isect<=7; isect++){
                                int counter=CenaSciCntrAcc::COIN_SECTOR(isect);
                                for(int ien=0; ien<nen; ien++){
                                        int v=0;
                                        for(int iph=0; iph<nph; iph++){
                                                v+=pac->get(ien, iph, counter);
                                        }
                                        fprintf(fp, " %d", v);
                                }
                                for(int ien=nen; ien<8; ien++){
                                        fprintf(fp, " %d", -1);
                                }
                        }
                }
};
/**@brief An implementation of CenaSciCntrAccFormatter.

	One-line formatter for energy-time diagram for stop plate.
        "nr_of_data nr_of_ene  ene0plate1  ene1plate1 .. eneNplate1 ene0plate2 ... eneNplate8"
        Maximum nr_of_data is (N=8)*8=64.  If less than maximum, -1 is filled.
*/
class CenaSciCntrAccFormatterStopPlate : public CenaSciCntrAccFormatter {
        public:
                ~CenaSciCntrAccFormatterStopPlate(){}
                void dump(FILE* fp, CenaSciCntrAcc* pac){
                        int nen=pac->getNrEne();
                        int nph=pac->getNrPhase();
                        fprintf(fp, "%d %d ", nen*8, nen);
                        for(int isect=1; isect<=8; isect++){
                                int counter=CenaSciCntrAcc::STOP_PLATE(isect);
                                for(int ien=0; ien<nen; ien++){
                                        int v=0;
                                        for(int iph=0; iph<nph; iph++){
                                                v+=pac->get(ien, iph, counter);
                                        }
                                        fprintf(fp, " %d", v);
                                }
                                for(int ien=nen; ien<8; ien++){
                                        fprintf(fp, " %d", -1);
                                }
                        }
                }
};
/**@brief An implementation of CenaSciCntrAccFormatter.

	One-line formatter for energy-time diagram for stop mesh.
        "nr_of_data nr_of_ene  ene0 ene1 ... eneN"
        Maximum nr_of_data is N=8.  If less than maximum, -1 is filled.
*/
class CenaSciCntrAccFormatterStopMesh : public CenaSciCntrAccFormatter {
        public:
                ~CenaSciCntrAccFormatterStopMesh(){}
                void dump(FILE* fp, CenaSciCntrAcc* pac){
                        int nen=pac->getNrEne();
                        int nph=pac->getNrPhase();
                        fprintf(fp, "%d %d ", nen*8, nen);
                        int counter=CenaSciCntrAcc::STOP_MESH();
                        for(int ien=0; ien<nen; ien++){
                                int v=0;
                                for(int iph=0; iph<nph; iph++){
                                        v+=pac->get(ien, iph, counter);
                                }
                                fprintf(fp, " %d", v);
                        }
                        for(int ien=nen; ien<8; ien++){
                                fprintf(fp, " %d", -1);
                        }
                }
};



//end of namespace
}
}



#endif

