/*
 *  CenaSciMassAcc.h
 *  libsadppac
 *
 *  Created by Futaana Yoshifumi on 07/06/26.
 *  Copyright 2007 __MyCompanyName__. All rights reserved.
 *
 */

#ifndef _IRFPAC_SARAPAC_CENASCIMASSACC
#define _IRFPAC_SARAPAC_CENASCIMASSACC

#include <vector>
#include "IrfPac/SaraPac/CenaScienceDataPacket.h"

namespace IrfPac{
namespace SaraPac{

/**	@brief CENA science packet (non-rice-compressed) of mass accumulation mode.

	Count matrix of CENA mass mode.  The coincident count is stored to the corresponding
	bins for energy, phase, channel and mass.  
	Maximum numbers of bins are 8 for energy, 32 for phase, 7 for channel and 128 for mass.
	Note that when the number of channel bins are 7, a padding is added to the internal
	expression.  This is a DPU specification, and the library follows to the definition.
	Therefore the class support two ways of getting the number of bins: getNrCh() and getNrChIntr().
	
	@sa CenaMode
*/
class CenaSciMassAcc : public CenaScienceDataPacket {

public:
	CenaSciMassAcc(uint8_t *pac, int nEnergy, int nPhase, int nC, int nMass);
	~CenaSciMassAcc(){};

	virtual uint8_t getRaw(int iE, int iP, int iC, int iM);
	/**@brief Get the count in the packet.
	*/
	virtual uint16_t get(int iE, int iP, int iC, int iM);
	/**@brief Get the observed coincident count as an array.

	The count is decompressed.  The array size constitute of iE*iP*iCintr*iM, where iCintr is the internal expression of channel bin numbers, i.e., (iC==7)?8:ic.
	The index can be obtained by indexOf(int,int,int,int) method.

	For users of python extension, the method is extremely useful, and faster than using iteration of get() method.
	Work routine is
	<pre>
	from Sadppac import *
	c=CenaBmuReader()
	c.readFile('SRA01D18CH101252003709051005532453.cena')
	arr=c.getArray()
	pac=arr.getPacket(30)
	pac2=pac.decompress()
	mass=pac2.getCenaSciMassAcc()
	print mass.getNrCh()
	print mass.getNrEne()
	print mass.getNrMass()
	print mass.getNrPhase()
	arr=mass.getCountAsArray()
	</pre>
	For users of matplotlib, more useful way is
	<pre>
	matx=array(arr)
	matx.shape=[mass.getNrEne(), mass.getNrPhase(), mass.getNrChIntr(), mass.getNrMass()]
	</pre>
	@return A pointer to the new object of valarray<double> in which the count data is stored.
	*/
	virtual std::vector<double> getCountAsArray();

	/**@brief Return the index to the internal expression of the memory layout.
	*/
	virtual int indexOf(int iE, int iP, int iC, int iM);

	virtual int getNrEne(){return _ne;};
	virtual int getNrPhase(){return _np;};
	/**@brief Returns the number of channel bins.  (1,2,4,7)
	 */
	virtual int getNrCh(){return _nc;};
	/**@brief Returns the number of internal expression of channel bins.  (1,2,4,8)
	 */
	virtual int getNrChIntr(){return _nc_padd;}
	virtual int getNrMass(){return _nm;};
	
	virtual void dumpAll(){dumpAll(stdout);};
	virtual void dumpAll(FILE* fp){
		int size=getSize();
		fprintf(fp, "*=*=*= MASS ACC MODE (raw dump) Size=%d [E=%d P=%d C=%d M=%d] =*=*=*\n", size, _ne, _np, _nc, _nm);
		uint8_t* str=getStream();
		for(int i=0; i<size; i++){
			if(i%16==0)fprintf(fp, "%08x: ", i);
			fprintf(fp, "%02x ", str[i]);
			if(i%16==15)fprintf(fp, "\n");
		}
		fprintf(fp, "\n*=*=*= END MASS ACC MODE =*=*=*\n");
		delete[] str;
	};


protected:
	int _ne, _np, _nc, _nm, _nc_padd;

};

/** @brief Another view of the science matrix for mass mode.
 *
 * Number of different CENA energy setting is 16.  In each SV table, 8 of 16 is selected
 * In this view, energy step looks like 16, and 8 of 16 has meaningful data, and the other is garbage (255 for getRaw).
 * The benefit of using this class is that the data can be treated independent of SV table.
 *
 * The non-observed data is filled by 255 (raw) and log8linear16(255) = 63488.
 *
 * This class is prepared to handle the CENA mass mode data independent of CENA SV table setting.
 * This class instance is generated by CenaPacket::getCenaSciMassAccE16() method.
 *
 * According to the CENA calibration table, the SV table (sv) dependence of the E8 index (iE) and E16 index is as follows.
 * <pre>
 * iE		SV=0	SV=1	SV=2		SV=3
 * 0		n/a		8		11		15
 * 1		n/a		7		10		14
 * ...
 * 7		n/a		1		4		8
 * </pre>
 * 
 * 
 * Normal data access with CenaPacket::getCenaSciMassAccE16() maps the E16 index to E8. Where no data is present because 
 * the E16 index does not have a corresponding E8 index, the non-observed data case is used.
 * The case of SV=0 is special such that independent of E16 index, the result is the non-observed data case. 
 * This is better than to throw an exception as it allows to use other properties of the data packet nevertheless.
 */
class CenaSciMassAccE16 : public CenaSciMassAcc {
private:
	int _sv; // SV table setting
public:
	CenaSciMassAccE16(CenaSciMassAcc* pac, int sv) : CenaSciMassAcc(pac->getStream(), pac->getNrEne(), pac->getNrPhase(), pac->getNrCh(), pac->getNrMass()){
		if(pac->getNrEne() != 8){
			std::cerr << "Number of energy = " << pac->getNrEne() << std::endl;
			throw std::runtime_error("The number of energy should be 8, but it was not!!");
		}
		if(sv<0 || sv>=4){
			std::cerr << "SV table setting out of range: " << sv << std::endl;
			throw std::runtime_error("SV setting is outside of useful range (0, 1, 2 or 3)");
		}
		_sv = sv;
//		std::cout << "CENA SCI MASS ACC E16 Instanced!!" << std::endl;
	}

	virtual ~CenaSciMassAccE16(){}

	int getNrEne(){
		return 16;
	}

	void dumpAll(){dumpAll(stdout);}
	void dumpAll(FILE* fp){
		fprintf(fp, "CLASS:  CenaSciMassAccE16. Size=%d [E=%d P=%d C=%d M=%d]\n", (int)getSize(), (int)getNrEne(), (int)getNrPhase(), (int)getNrCh(), (int)getNrMass());
		fprintf(fp, "Ph Ch Ma | E0    E1    E2    E3    E4    E5    E6    E7    E8    E9   E10   E11   E12   E13   E14   E15\n");
		fprintf(fp, "-------------------------------------------------------------------------------------------------------\n");
		for(int ip=0; ip<getNrPhase(); ip++){
			for(int ic=0; ic<getNrCh(); ic++){
				for(int im=0; im<getNrMass(); im++){
					fprintf(fp, "%02d %02d %02d | ", ip, ic, im);
					for(int ie=0; ie<getNrEne(); ie++){
						fprintf(fp, "%5d ", getRaw(ie, ip, ic, im));
					}
					fprintf(fp, "\n");
				}
			}
		}
	}

	int getSvTable(){
		return _sv;
	}

	std::vector<double> getCountAsArray();
	int indexOf(int iE16, int iP, int iC, int iM);
	uint8_t getRaw(int iE16, int iP, int iC, int iM){
		int idx = indexOf(iE16, iP, iC, iM);
		if(idx < 0){
			return 255;
		}
		else{
			return getValue(idx);
		}
	}


//	Both many not be needed...
//	uint8_t getRaw(int iE16, int iP, int iC, int iM);
//	uint16_t get(int iE16, int iP, int iC, int iM);
	


};

/**@brief Virtual class of CenaSciMassAcc formatter.

   CenaSciMassAcc class has 4 dimensional data in one packet : iE x iP x iC x iM .
 */
class CenaSciMassAccFormatter {
public:
	virtual ~CenaSciMassAccFormatter(){};
	virtual void dump(FILE* fp, CenaSciMassAcc* pac)=0;
};


/**@brief A dump of energy spectrum.

   Collapse over all phase, channel, and mass.
   Maximum real energy step is 8, but here filled by -1 until 32 steps.
   If the Matrix is reformated to be SV independent with 16 energy steps then 
   it wil be printed accordingly with unpopulated energy steps = -1.
	   
 */
class CenaSciMassAccFormatterEspec : public CenaSciMassAccFormatter{
public:
	~CenaSciMassAccFormatterEspec(){}
	void dump(FILE* fp, CenaSciMassAcc* pac){
		int nen=pac->getNrEne();
		int nph=pac->getNrPhase();
		int nch=pac->getNrCh();
		int nms=pac->getNrMass();
		fprintf(fp, "%d %d %d %d ", nen, nph, nch, nms);
		for(int ien=0; ien<nen; ien++){
			int v=0;
			for(int iph=0; iph<nph; iph++){
				for(int ic=0; ic<nch; ic++){
					for(int im=0; im<nms; im++){
						v+=pac->get(ien, iph, ic, im);
					}
				}
			}
			if (v >= 63488.) {
				fprintf(fp, "-1 ");
			} else {
				fprintf(fp, "%d ", v);
			}	
		}
		for(int ien=nen; ien<32; ien++){
			fprintf(fp, "-1 ");
		}
	}
};


/**@brief A dump of mass spectrum.

   Collapse over all enegy, phase, and channel.
   If I remember correctly, maximum mass step is 128, but here filled by -1 until 256 steps.
 */
class CenaSciMassAccFormatterMspec : public CenaSciMassAccFormatter{
public:
	~CenaSciMassAccFormatterMspec(){}
	void dump(FILE* fp, CenaSciMassAcc* pac){
		int nen=pac->getNrEne();
		int nph=pac->getNrPhase();
		int nch=pac->getNrCh();
		int nms=pac->getNrMass();
		fprintf(fp, "%d %d %d %d ", nen, nph, nch, nms);
		for(int ims=0; ims<nms; ims++){
			int v=0;
			for(int ien=0; ien<nen; ien++){
				for(int iph=0; iph<nph; iph++){
					for(int ic=0; ic<nch; ic++){
						v+=pac->get(ien, iph, ic, ims);
					}
				}
			}
			fprintf(fp, "%d ", v);
		}
		for(int ims=nms; ims<256; ims++){
			fprintf(fp, "-1 ");
		}
	}
};

/**@brief A dump of mass spectrum for each energy step.

   Collapse over phase, and channel.
   Each line constitute of 128 mass steps * 8 energy steps.
   First mass step index cycle, then energy step: i.e.,
   mass0ene0 mass1ene0 mass2ene0 ... mass127ene0 mass0ene1 mass1ene1 ...
   If no data availble due to binning parameter, -1 is filled.
 */
class CenaSciMassAccFormatterMspecEne : public CenaSciMassAccFormatter{
public:
	~CenaSciMassAccFormatterMspecEne(){}
	void dump(FILE* fp, CenaSciMassAcc* pac){
		int nen=pac->getNrEne();
		int nph=pac->getNrPhase();
		int nch=pac->getNrCh();
		int nms=pac->getNrMass();
		fprintf(fp, "%d %d %d %d ", nen, nph, nch, nms);
		for(int ien=0; ien<nen; ien++){
			for(int ims=0; ims<nms; ims++){
				int v=0;
				for(int iph=0; iph<nph; iph++){
					for(int ic=0; ic<nch; ic++){
						v+=pac->get(ien, iph, ic, ims);
					}
				}
				fprintf(fp, "%d ", v);
			}
			for(int ims=nms; ims<128; ims++){
				fprintf(fp, "-1 ");
			}
		}
		for(int ien=nen; ien<8; ien++){
			for(int ims=0; ims<128; ims++){
				fprintf(fp, "-1 ");
			}
		}
	}
};


/**@brief A dump of energy spectrum for each channel step, only mass bins 0..63 are used.

   Collapse over phase, and mass 0..63. mass 64..128 is not used.
   Each line constitute of 8 energy steps * 7 direction steps.
   First energy step index cycle, then channel step: i.e.,
   ene0chan0 ene1chan0 ene2chan0 ... ene7chan0 ene0chan1 ene1chan1 ...
   If no data availble due to binning parameter, -1 is filled.
 */
class CenaSciMassAccFormatterLightEspecC : public CenaSciMassAccFormatter{
public:
    ~CenaSciMassAccFormatterLightEspecC(){}
    void dump(FILE* fp, CenaSciMassAcc* pac){
        int nen=pac->getNrEne();
        int nph=pac->getNrPhase();
        int nch=pac->getNrCh();
        int nms=pac->getNrMass(); // use only 0..nms/2-1
        fprintf(fp, "%d %d %d %d ", nen, nph, nch, nms);
        for(int ic=0; ic<nch; ic++){
            for(int ien=0; ien<nen; ien++){
                int v=0;
                for(int iph=0; iph<nph; iph++){
                    for(int ims=0; ims<(nms/2); ims++){
                        v+=pac->get(ien, iph, ic, ims);
                    }
                }
                fprintf(fp, "%d ", v);
            }
            for(int ien=nen; ien<8; ien++){
                fprintf(fp, "-1 ");
            }
        }
        for(int ic=nch; ic<7; ic++){
            for(int ien=0; ien<8; ien++){
                fprintf(fp, "-1 ");
            }
        }
    }
};

/**@brief A dump of energy spectrum for each channel step, only mass bins 64..127 are used.

   Collapse over phase, and mass 64..127. mass bins 0..63 are not used.
   Each line constitute of 8 energy steps * 7 direction steps.
   First energy step index cycle, then channel step: i.e.,
   ene0chan0 ene1chan0 ene2chan0 ... ene7chan0 ene0chan1 ene1chan1 ...
   If no data availble due to binning parameter, -1 is filled.
 */
class CenaSciMassAccFormatterHeavyEspecC : public CenaSciMassAccFormatter{
public:
    ~CenaSciMassAccFormatterHeavyEspecC(){}
    void dump(FILE* fp, CenaSciMassAcc* pac){
        int nen=pac->getNrEne();
        int nph=pac->getNrPhase();
        int nch=pac->getNrCh();
        int nms=pac->getNrMass(); // use only 0..nms/2-1
        fprintf(fp, "%d %d %d %d ", nen, nph, nch, nms);
        for(int ic=0; ic<nch; ic++){
            for(int ien=0; ien<nen; ien++){
                int v=0;
                for(int iph=0; iph<nph; iph++){
                    for(int ims=nms/2; ims<nms; ims++){
                        v+=pac->get(ien, iph, ic, ims);
                    }
                }
                fprintf(fp, "%d ", v);
            }
            for(int ien=nen; ien<8; ien++){
                fprintf(fp, "-1 ");
            }
        }
        for(int ic=nch; ic<7; ic++){
            for(int ien=0; ien<8; ien++){
                fprintf(fp, "-1 ");
            }
        }
    }
};


/**@brief A dump of directional spectrum.

   Collapse over all enegy, phase, and mass.
   Maximum energy step is 7.
 */
class CenaSciMassAccFormatterCspec : public CenaSciMassAccFormatter{
public:
	~CenaSciMassAccFormatterCspec(){}
	void dump(FILE* fp, CenaSciMassAcc* pac){
		int nen=pac->getNrEne();
		int nph=pac->getNrPhase();
		int nch=pac->getNrCh();
		int nms=pac->getNrMass();
		fprintf(fp, "%d %d %d %d ", nen, nph, nch, nms);
		for(int ich=0; ich<nch; ich++){
			int v=0;
			for(int ien=0; ien<nen; ien++){
				for(int iph=0; iph<nph; iph++){
					for(int ims=0; ims<nms; ims++){
						v+=pac->get(ien, iph, ich, ims);
					}
				}
			}
			fprintf(fp, "%d ", v);
		}
		for(int ich=nch; ich<7; ich++){
			fprintf(fp, "-1 ");
		}
	}
};



//end of namespace
}
}



#endif

