/*
 *  SaraBufferedDemultiplexor.h
 *  libsadppac
 *
 *  Created by Futaana Yoshifumi on 07/10/23.
 *  Copyright 2007 __MyCompanyName__. All rights reserved.
 *
 */

#ifndef __IRFPAC_SARAPAC_SARABUFFEREDDEMULTIPLEXOR_H
#define __IRFPAC_SARAPAC_SARABUFFEREDDEMULTIPLEXOR_H

#include<list>
#include<IrfPac/irfpac.h>
#include<IrfPac/SaraPac/CenaPacket.h>
#include<IrfPac/SaraPac/SwimPacket.h>
#include<IrfPac/SaraPac/SaraPacket.h>
#include<IrfPac/SaraPac/SaraDemultiplexor.h>

namespace IrfPac{
namespace SaraPac{

/** @brief SaraPacket demultiplexor.  Usage can be found in @link DemultiplexorSample @endlink page.

	This class is for demultiplexing of SaraPacket to CenaPcket and SwimPacket.
	A sample code is described in @link DemultiplexorSample @endlink

	Basic ideas of this class are
	<ul>
	<li>Input SaraPacket* instances.
	<li>Demultiplex the input SaraPacket* to CenaPacket* and SwimPacket*.
	<li>Once CenaPacket and/or SwimPacket is completed, the CenaPacket* and SwimPacket* are buffered internally.
	<li>Access to the buffered CenaPacket* and SwimPacket* instances.
	</ul>

	To whom being interested in the details:
	There are 4 bufferes internally.  Incomplete CenaPacket and SwimPacket, and Complete CenaPacket and SwimPacket.
	Once SaraPacket comes, the packet is analyzed into a part of CenaPacket and SwimPacket up to the SaraPacket header.
	The analyzed CenaPacket and SwimPacket is stored to 'incomplete' CenaPacket and SwimPacket buffer respectively.
	Once CenaPacket (or SwimPacket) is completed, and it is copied to the 'complete' CenaPacket (or SwimPacket) buffer.
	There are no public method to manually modify the 'incomplete' CenaBuffer and SwimBuffer (Only appending SaraPacket is possible).

 */
class SaraBufferedDemultiplexor {

	public:
		SaraBufferedDemultiplexor();
		~SaraBufferedDemultiplexor();

		/**@brief Append SaraPacket instance for demultiplexor.
		 */
		void appendSaraPacket(SaraPacket* saraPac);
		/**@brief Returns the number of buffered CenaPacket.
		 */
		int getNrCena();
		/**@brief Returns the number of buffered SwimPacket.
		 */
		int getNrSwim();
		/**@brief Returns the first CenaPacket in the CENA buffer(like FIFO).

		   Returns the first (=oldest) CenaPacket.
		   If CenaPacket is not in the CENA buffer, NULL is returned.
		 */
		CenaPacket* getCena();
		/**@brief Returns the first SwimPacket in the SWIM buffer (like FIFO).

		   Returns the first (=oldest) SwimPacket.
		   If SwimPacket is not in the SWIM buffer, NULL is returned.
		 */
		SwimPacket* getSwim();
		
		/**@brief Returns the latest CenaPacket, and flush the CenaPacket* buffer.

			Returns the latest CenaPacket in the CENA buffer.  All the packets in the CENA buffer are then destroyed.
			If no CenaPacket instance is in the buffer, NULL is returned.
		 */
		CenaPacket* getLatestCena();
		/**@brief Returns the latest SwimPacket, and flush the SwimPacket* buffer.

			Returns the latest SwimPacket in the SWIM buffer.  All the packets in the SWIM buffer are then destroyed.
			If no SwimPacket instance is in the buffer, NULL is returned.
		 */
		SwimPacket* getLatestSwim();
		
		/**@brief Flush all the buffered CenaPacket.
		 */
		void flushCena();
		/**@brief Flush all the buffered SwimPacket.
		 */
		void flushSwim();
		
		void showBufferStatus(){
			fprintf(stdout, "---CENA---\n");
			fprintf(stdout, "   Elems: %d\n", (int) cenaBuffer.size());
			fprintf(stdout, "---SWIM---\n");
			fprintf(stdout, "   Elems: %d\n", (int) swimBuffer.size());
		}

	private:
		SaraDemultiplexor *demul;
		std::list<CenaPacket*> cenaBuffer;
		std::list<SwimPacket*> swimBuffer;

};


}}// namespace


#endif

