#ifndef __IRAPAC_SARAPAC_SWIMACCUMTOTMATX_H
#define __IRAPAC_SARAPAC_SWIMACCUMTOTMATX_H


/*
 *  SwimAccumTotMatx.h
 *  libsadppac
 *
 *  Created by Futaana Yoshifumi on 07/11/21.
 *  Copyright 2007 __MyCompanyName__. All rights reserved.
 *
 */
#include<IrfPac/irfpac.h>
#include<linlog++.h>

namespace IrfPac{
namespace SaraPac{

class SwimAccumTotMatxDump;   // Forward declaration

/** @brief Accumulation matrix data of total count values.

	The size is described as [ExDx5] bytes.
*/
class SwimAccumTotMatx : public PacketU8Arr {

public:
	/** @brief A constructor of SWIM accumulation matrix 
	
		@todo Unknown size of the buffer....They say n_e * n_d * 5, but.....(Padding??)
	*/
	SwimAccumTotMatx(uint8_t* stream, int n_e, int n_d);
	virtual ~SwimAccumTotMatx();

	/**@brief Create a 16x16 reformed total matrix for SWIM.
	   @return New packet that has E=16 D=16 matrix.  The returned object is in fact an instance of SwimAccumTotMatx16x16, but you can use it tranparently as SwimAccumTotMatx.
	   @throw std::runtime_error when the method is applied to the packet which has not E=32 and D=8.
	   @sa SwimAccumTotMatx16x16
	 */
	SwimAccumTotMatx* reform16x16();

	/**@brief Dump the contents using default formatter (SwimAccumTotMatxAllDump).
	 */
	virtual void dumpAll(FILE* fp);
	virtual void dump(FILE* fp, SwimAccumTotMatxDump* dump);

	static const char* getCounterName(int index){
		if(index==0){
			return "START COUNTER";
		}
		else if(index==1){
			return "STOP COUNTER";
		}
		else if(index==2){
			return "OVER COUNTER";
		}
		else if(index==3){
			return "UNDER COUNTER";
		}
		else if(index==4){
			return "TOF COUNTER";
		}
		throw "Out of index for SWIM total counter.";
	}

	static const int START_COUNTER=0;
	static const int STOP_COUNTER=1;
	static const int OVER_COUNTER=2;
	static const int UNDER_COUNTER=3;
	static const int TOF_COUNTER=4;

	/** @brief Returns the raw value (8bit) corresponding to the specified energy and deflection_bin.
	 *
	 */
	virtual uint8_t getRaw(int i_e, int i_d, int counter){
		int index = i_e * _nd * 5 + i_d * 5 + counter;	
		return getValue(index);
	}

	/** @brief Returns the value (32 bit) corresponding to the specified energy and deflection bin.

	    The value is lin-log compressed (32bit->8bit).
	    This function returnes the 32 bit decoded values.
	 */
	virtual uint32_t get(int i_e, int i_d, int counter){
	  uint8_t val8=getRaw(i_e, i_d, counter);
	  return (uint32_t) log8tolinear32((unsigned char) val8);
	}

	/** @brief Returns the sum of raw counts over energy for specified deflection bin and counter

	    Note that this is a sum of the raw values (without lin-log decompression).  No physical meaning.
	 */
	virtual uint32_t collapseEnergyRaw(int i_d, int counter){
		uint32_t cnt=0;
		for(int ie=0; ie<_ne; ie++){
			cnt+=getRaw(ie, i_d, counter);
		}
		return cnt;
	}


	/** @brief Returns the sum of decompressed counts over energy for specified deflection bin and counter 
	 */
	virtual double collapseEnergy(int i_d, int counter){
		double cnt=0;
		for(int ie=0; ie<_ne; ie++){
			cnt+=get(ie, i_d, counter);
		}
		return cnt;
	}

	/** @brief Returns the sum of raw counts over deflection bins for specified energy bin and counter

	    Note that this is a sum of the raw values (without lin-log decompression).  No physical meaning.
	 */
	virtual uint32_t collapseDeflbinRaw(int i_e, int counter){
		uint32_t cnt=0;
		for(int id=0; id<_nd; id++){
			cnt+=getRaw(i_e, id, counter);
		}
		return cnt;
	}


	/** @brief Returns the sum of decompressed counts over deflection bins for specified energy bin and counter
	 */
	virtual double collapseDeflbin(int i_e, int counter){
		double cnt=0;
		for(int id=0; id<_nd; id++){
			cnt+=get(i_e, id, counter);
		}
		return cnt;
	}

	/** @brief Returns the sum of raw counts over deflection bins and energy for specified counter

	    Note that this is a sum of the raw values (without lin-log decompression).  No physical meaning.
	 */
	virtual uint32_t collapseAllRaw(int counter){
		uint32_t cnt=0;
		for(int ie=0; ie<_ne; ie++){
			for(int id=0; id<_nd; id++){
				cnt+=getRaw(ie, id, counter);
			}
		}
		return cnt;
	}

	/** @brief Returns the sum of lin-log decompressed counts over deflection bins and energy bins for specified counter
	 */
	virtual double collapseAll(int counter){
		double cnt=0;
		for(int ie=0; ie<_ne; ie++){
			for(int id=0; id<_nd; id++){
				cnt+=get(ie, id, counter);
			}
		}
		return cnt;
	}

	virtual int getEnergyBins(){return _ne;}
	virtual int getDeflectionBins(){return _nd;}

private:
	int _ne;  //<! Number of energy steps
	int _nd;  //<! Number of deflector bins
	SwimAccumTotMatxDump* _dump; //<! Default dump class

};

/** @brief Pure virtual class for formatting the contents of the SwimAccumTotMatx.
    
    Dumping the contents is always the issue for this kind of library.
    As an trial implementation, I made a pure virtual class of SwimAccumTotMatxDump.
    This class is intened to be extended to a concrete class.
    The instance of one concrete class will be added to SwimAccumTotMatx class,
    and the SwimAccumTotMatx::dumpAll(FILE*) method will dump the contents.
 */
class SwimAccumTotMatxDump {
public:
	virtual ~SwimAccumTotMatxDump(){}
	virtual void dump(FILE* fp, SwimAccumTotMatx* matx)=0;
};
/** @brief Implementation of HEX dumping class.
 */
class SwimAccumTotMatxHexDump : public SwimAccumTotMatxDump {
public:
	void dump(FILE* fp, SwimAccumTotMatx* matx){
		matx->dumpHexStream(fp);
	}
};

/** @brief Implementation of All Dump.
 */
class SwimAccumTotMatxAllDump : public SwimAccumTotMatxDump {
public:
	void dump(FILE* fp, SwimAccumTotMatx* matx);
};

/** @brief Implementation of energy spectrum dump for start counter.

 */
class SwimAccumTotMatxEspecDump : public SwimAccumTotMatxDump {
private:
	int cnt_id;
public:
	SwimAccumTotMatxEspecDump(int counter_id){
		cnt_id=counter_id;
	}
	void dump(FILE* fp2, SwimAccumTotMatx* tot){
//		fprintf(fp2, "0x%04x %14.2f ",  swim->get8sCycleCntr(), swim->getUniversalTimeInMsec());
		int ne=tot->getEnergyBins();
		int emax=32;    // Maximum energy bins are 32.
		fprintf(fp2, "%2d ", ne);
		for(int i=0; i<ne; i++){
			fprintf(fp2, "%6ld ", (long)tot->collapseDeflbin(i, cnt_id));
		}
		for(int i=ne; i<emax; i++){
			fprintf(fp2, "    -1 ");
		}
    
		fprintf(fp2, "\n");

	}

};

//  ?? TODO debug: not working.
class SwimAccumTotMatxDeflDump : public SwimAccumTotMatxDump {
private:
	int cnt_id;
public:
	SwimAccumTotMatxDeflDump(int counter_id){
		cnt_id=counter_id;
	}
	void dump(FILE* fp2, SwimAccumTotMatx* tot){
		int nd=tot->getDeflectionBins();
		int dmax=32;    // Maximum energy bins are 32.
		fprintf(fp2, "%2d ", nd);
		for(int i=0; i<nd; i++){
			fprintf(fp2, "%6ld ", (long)tot->collapseEnergy(i, cnt_id));
		}
		for(int i=nd; i<dmax; i++){
			fprintf(fp2, "    -1 ");
		}
    
		fprintf(fp2, "\n");

	}
};

/**@brief Formatter of SWIM accumulation total matrix to make E-t diagram in each directions.
  

   The output is 2+8*32=256 list.  The order is
   <pre>
   nd  ne  count_d=0_e=0  count_d=0_e=1  count_d=0_e=2  ... count_d=0_e=31  count_d=1_e=0 ... count_d=1_e=31 ... count_d=7_e=0 ... count_d=7_e=31
   </pre>
   If no values are avaiable, -1 is written.
 
 */
class SwimAccumTotMatxEspecDump8Dir : public SwimAccumTotMatxDump {
private:
	int _cnt_id;
public:
	SwimAccumTotMatxEspecDump8Dir(int counter_id){
		_cnt_id=counter_id;
	}
	void dump(FILE* fp, SwimAccumTotMatx* matx){
		int nd=matx->getDeflectionBins();  // Maximum=8
		int dmax=8;
		int ne = matx->getEnergyBins();

		fprintf(fp, "%2d %2d ", nd, ne);

		for(int id=0; id<nd; id++){      // Direction with data
			for(int ie=0; ie<ne; ie++){
				fprintf(fp, "%6ld ", (long) matx->get(ie,id,_cnt_id));
			}
			for(int ie=ne; ie<32; ie++){
				fprintf(fp, "     -1 ");
			}
		}
		for(int id=nd; id<dmax; id++){   // Direction without data
			for(int ie=0; ie<32; ie++){
				fprintf(fp, "     -1 ");
			}
		}
		fprintf(fp, "\n");
	}
};

/**@brief An implementation of 16x16-reformed SWIM total matrix.
   @sa SwimSciAcc16x16 for details of 16x16 reform.
 */
class SwimAccumTotMatx16x16 : public SwimAccumTotMatx {
public:
	SwimAccumTotMatx16x16(uint8_t* pac);
	~SwimAccumTotMatx16x16(){}
	/**@brief Get raw counts.
	   @sa get(int,int,int)

	   Accessor to the raw count. aE and aD should be 0 to 15.
	 */
	uint8_t getRaw(int aE, int aD, int iC);

};

}}  // namespace


#endif
