/*
 *  SwimMatrix.h
 *  libsadppac
 *
 *  Created by Futaana Yoshifumi on 07/11/21.
 *  Copyright 2007 __MyCompanyName__. All rights reserved.
 *
 */

#ifndef __IRFPAC_SARAPAC_SWIMMATRIX_H
#define __IRFPAC_SARAPAC_SWIMMATRIX_H

#include "IrfPac/irfpac.h"
#include "IrfPac/SaraPac/SwimSciAcc.h"
#include "IrfPac/SaraPac/SwimScienceDataPacket.h"


namespace IrfPac{
namespace SaraPac{

/**@brief Matrix part of SWIM packet.

	@sa SwimPacket
*/
class SwimMatrix : public PacketU8Arr {

public:

	SwimMatrix(uint8_t* packet, int size);
	~SwimMatrix(){};
	
	/**@brief Exception thrown when you try to decompress a packet which is not compressed.
	*/
	static const int NOT_COMPRESSED=120;
	/**@brief Exception thrown when the decompression fails.
	*/
	static const int DECOMPRESSION_FAILED=255;
	/**@brief Exception if the packet contents is illegal.
	*/
	static const int ILLEGAL_PACKET=75;


	/**@brief A header and descripter of compressed packet.
	*/
	static const uint32_t COMPRESSED_HEADER=0xAA55AA55;
	/**@brief A header of uncompressed packet.
	*/
	static const uint32_t NON_COMPRESSED_HEADER=0xAADDAADD;
			
	void dumpAll(FILE* fp){
		fprintf(fp, "..... HEX DUMP total %d byte (Human readable part will be implemented soon) .....\n", (int)this->getSize());
		dumpHexStream(fp);
	};

	

	/**@brief Debug verbosity when reporting errors during decompression
	*/
	static const int VERBOSE_DEBUG=3;
	/**@brief Normal verbosity when reporting errors during decompression
	*/
	static const int VERBOSE_NORMAL=2;
	/**@brief Minimum verbosity when reporting errors during decompression.
	*/
	static const int VERBOSE_MIN=1;
	/**@brief No verbosity when reporting errors during decompression.
	*/
	static const int VERBOSE_NONE=0;
	
	/**@brief Decompress the packet with selectable versosity.

		As decompress() but with selectable verbosity level:
		
		VERBOSE_NORMAL : Standard verbosity on stderr
		VERBOSE_MIN    : Minimum verbosity on stderr (no c.txt dump)
		VERBOSE_NONE   : (almost) no output on stderr in case of error
		
	*/
	SwimMatrix* decompress(int verbose);

	/**@brief Decompress the packet.
		Decompress the packet and create new SwimMatrix instance.
		After that, the pointer to the instance is returned.
		
		@exception NOT_COMPRESSED if the packet is not compressed.
		@exception ILLEGAL_PACKET if the packet is illegally formatted.
		@exception DECOMPRESSION_FAILED if the decompression failed.
		
		this is the same as decompress(VERBOSE_NORMAL)
	*/
	SwimMatrix* decompress(){	
		return decompress(VERBOSE_NORMAL);
	}	
	
	/**@brief Interpret the contents of matrix, and returns the science data packet.
		
		The returned type is SwimScienceDataPacket*, but, this is a base class for
		mode dependent (see CENA case).  Fortunately, SWIM is only one mode,
		only available packet is SwimSciAcc class.
		
		@sa SwimScienceDataPacket
		@sa SwimSciAcc
	*/
	SwimScienceDataPacket* interpret(uint8_t ne, uint8_t nd, uint8_t nm){
		// Obtain the raw stream for the SwimScienceDataPacket
		uint8_t* stream=getStream();
		
		// If it is compressed...
		uint32_t header = (((((((uint32_t)stream[0])<<8)|stream[1])<<8)|stream[2])<<8)|stream[3];
#ifdef DEBUG
				fprintf(stdout, "Header: %lx\n", (long)header);
#endif
		if(header == COMPRESSED_HEADER){
			fprintf(stderr, "COMPRESSED PACKET IS NOT YET SUPPORTED.  PLEASE DO compress() FIRST.\n");
			throw ILLEGAL_PACKET;
		}
		
		// For SWIM, only one mode is available.
		
#ifdef DEBUG
		uint32_t size=getSize();
		
		fprintf(stdout, "SWIM_MATRIX: (Ne,Nd,Nm)=(%d,%d,%d)\n", ne, nd, nm);
		fprintf(stdout, "SWIM_MATRIX: Expected SIZE (%d) = \n", ne*nd*nm);
		fprintf(stdout, "SWIM_MATRIX: Real SIZE (%d) = \n", size);
#endif
		
		
		SwimSciAcc *mats = new SwimSciAcc(stream+4, ne, nd, nm);
		delete[] stream;

//	mats->dumpAll(stdout);

		return mats;
		
	};

private:

	char* preformat();   //!< Only called from decompress()
	uint8_t* decomp_main(char* inputStream, size_t* decompressedSize);   //!< Only called from decompress()


};

}};

#endif
