/*
 *  SwimPacket.h
 *  libsadppac
 *
 *  Created by Futaana Yoshifumi on 07/09/12.
 *  Copyright 2007 __MyCompanyName__. All rights reserved.
 *
 */

#ifndef IRFPAC_SARAPAC_SWIMPACKET_H
#define IRFPAC_SARAPAC_SWIMPACKET_H

#include<vector>
#include<stdexcept>

#include <IrfPac/irfpac.h>
#include "IrfPac/SaraPac/TimeConversion.h"
#include "IrfPac/SaraPac/SaraPacket.h"
#include "IrfPac/SaraPac/SwimSubcHkPacket.h"
#include "IrfPac/SaraPac/SwimFastHkPacket.h"
#include "IrfPac/SaraPac/SwimAccumTotMatx.h"
#include "IrfPac/SaraPac/SwimMatrix.h"
#include "IrfPac/SaraPac/SwimRawModeData.h"

namespace IrfPac{
namespace SaraPac{

/** @brief SWIM packet

	This is the implementation of SWIM packet.
	SWIM packet is a data stream consisting of stuff, HK and science data.
	
	Stuff part contains various information, including time and mode.
	Methods to access these fields are implemented in this SwimPacket class.
	
	HK is a house keeping packet which constitute of 14 byte subpacket.
	The packet is implemented in SwimFastHkPacket class.
	
	Science part is not yet implemented.
	
	The data structure is:
	<pre>
	0             28                       42                       variable
	|----- SwimPacket ----------------------------------------------------------------------------|
	|--- Stuff ---|--- SwimFastHkPacket ---|--- SwimAccumTotMatx ---|--- SwimMatrix --------------| 
	                                                                       | (convert to)
	                                                                |--- SwimScienceDataPacket ---|
	                                                                
	                        (alternative)  |--- SwimRawData --------|
	</pre>

	@section work_routine Work routine

	SwimPacket analysis routine is as follows.
	<pre>
	SwimPacket *swim=demul->appendSwim(saraPac);   // demul is SaraDemultiplexor instance.  saraPac is SaraPacket instance.
	if(swim!=NULL){  // swim packet is arrived
		double univTime=swim->getUniversalTimeInMsec();  // Obtain universal time of observations.
		SwimFastHkPacket* fastHk = swim->getFastHkPac();  // Obtain embedded house keeping.
		SwimAccumTotMatx* totalMatx = swim->getAccumTotMatx();  // Obtain the total matrix.
		SwimSciAcc* sciMatx;

		SwimPacket *swim_decmp = swim->decompress();  // Decompress packet.  (For non-compressed packet, a copy is returned)
		sciMatx=swim_decmp->getSwimSciAcc();  // You can now get the science matrix.
		
		//////
		Here you can use whatever you want.
		//////

		delete swim_decmp;
		delete sciMatx;
		delete fastHk;
		delete swim;
	}
	</pre>

	@section reform16x16 Reforming matrix of nE=32 and nD=8
	Special treatment should be taken when the binning parameters of SWIM are E=32, D=8.
	Because of last minute sensor configuration change, the binning parameter of (E,D)=(32,8) should be
	considered as (E,D)=(16,16).

	The SwimSciAcc class and SwimAccumTotMatx class will be affected.  To conduct science analysis,
	users should know the fact.
	Because the policy of the libsadppac is simple and flexible, the conversion is not done automatically.

	Take an example of SwimAccumTotMatx. The instance can be obtained by
	<pre>
	SwimAccumTotMatx* totalMatx = swim->getAccumTotMatx();  // Obtain the total matrix.
	</pre>
	as in the previous example.
	If the binning parameter of the packet is (E,D)=(32,8),
	<pre>
	cout << totalMatx->getEnergyBins() << endl;
	cout << totalMatx->getDeflectionBins() << endl;
	</pre>
	will display 32 and 8.  Therefore the matrix cannot be used for science analysis.

	So one should reform the matrix to (E,D)=(16,16).
	libsadppac prepares the functionality: SwimAccumTotMatx::reform16x16().
	<pre>
	SwimAccumTotMatx* refMatx = totalMatx->reform16x16();
	</pre>
	The returned instance, refMatx, is the reformed matrix to (16,16).  Therefore,
	<pre>
	cout << refMatx->getEnergyBins() << endl;
	cout << refMatx->getDeflectionBins() << endl;
	</pre>
	will return 16 and 16.

	@subsection reform_tech_det Technical details
	The refMatx in the sample code is in fact the instance of SwimAccumTotMatx16x16.
	SwimAccumTotMatx16x16 is a derived class of SwimAccumTotMatx.
	Therefore one can use it transparently indentical to SwimAccumTotMatx.

	Conversion from E,D=32,8 to E,D=16,16 is as follows.  Assuming that iE and iD is the index
	for the original matrix (0<=iE<32; 0<=iD<8) and aE and aD is the index of the reformed matrix
	(0<=aE,aD<16).
	<pre>
	Original (iE, iD)  ->  Reform (aE, aD)
	</pre>
	The relationship is
	<pre>
	aE = (int)(iE/2) = (iE >> 1)
	aD = (iD*2 + iE mod 2) = (iD << 1) | (iE & 1)
	</pre>
	and
	<pre>
	iD = aD >> 1
	iE = (aE << 1) | (aD & 1)
	</pre>

*/
class SwimPacket : public PacketU8Arr
{
private:
	double _savedUtc;
	int _qflag;

public:

	static const int QFLAG_NOTSET=0;
	/**@brief This flag is prepared for the time stamp NG whatever the reason is.
	*/
	static const int QFLAG_TIMESTAMP_NG=0x01;
	/**@brief This flag is prepared for the OBT jump NG.  Normally QFLAG_TIMESTAMP_NG will be set as well.
	*/
	static const int QFLAG_OBT_JUMP_NG=0x02;
	/**@brief This flag is used to specify that the UTC conversion is done by extrapolation of time file contents.  Normally set with QFLAG_TIMESTAMP_NG. If a jump could be identified to be due to BMU-DPU time szncronization, then QFLAG_EXTRAPOL_DPU_SYNC is used instead
	*/
	static const int QFLAG_EXTRAPOL_TIME_FILE=0x04;
	/**@brief This flag is set when the time file is not found during demultiplex.  Normally QFLAG_TIMESTAMP_NG is also set.
 	*/
	static const int QFLAG_TIMEFILE_NOTFOUND=0x08;

	/**@brief This flag is used to specify that the UTC conversion is done by extrapolation of time file contents due time sync between BMU and DPU. Normally set with QFLAG_TIMESTAMP_NG.
	*/
	static const int QFLAG_EXTRAPOL_DPU_SYNC=0x10;
	/**@brief If QFLAG_EXTRAPOL_DPU_SYNC is set for more than 16 (TBC) packets this flag is added to indicate a probably dubious synthetic time interpolation. Normally set with QFLAG_TIMESTAMP_NG and QFLAG_EXTRAPOL_DPU_SYNC.
	*/
	static const int QFLAG_EXTRAPOL_DPU_SYNC_DUBIOUS=0x20;
	
	
	
	SwimPacket(uint8_t* stream, size_t size);
	~SwimPacket();

	void dumpAll(void){dumpAll(stdout);};
	void dumpAll(FILE* fp);

	void printParamsTbl(FILE* fp);
	void printParamsTbl2(FILE* fp, TimeConversion* tconv);
	void printParamsTbl3(FILE* fp, TimeConversion* tconv,bool useUtc);
	static void printParamsTblHeader(FILE* fp);

	/** @brief Identifier of the SWIM science packet.
	*/
	const static uint32_t SWIM_SCI_ID=0xDDDDEEEE;

	/** @brief Returns field of the packet for the identifier of SWIM science packet.

		Return the field of identifier of the SWIM science packet.
		First and second words in the memory map.
		In the DPU document, it is specified as 0xDDDDEEEE.
		This value is defined as SWIM_SCI_ID field.
	*/
	uint32_t getHeadMagicNumber(){
		return getU32(1,2);
	}

	/** @brief Returns 48 bit BMU OBT @ marker
	*/
	uint64_t getObtTime(){
		return getU64(3,4,5);
	}

  /** @brief Get the OBT time at marker in msec.
	@deprecated Use getMarkerObt()
  */
DEPRECATED
  double getObtTimeInMsec(){
    uint64_t obt=getObtTime();
    return obt2msec(obt);
  }

  /** @brief Get marker time in OBT format.
  */
  double getMarkerObt(){
    uint64_t obt=getObtTime();
    return obt2msec(obt);
  }
	
	/** @brief Returns DPU time at marker (word=6)
	*/
	uint16_t getDpuTime(){
		return getU16(6);
	}

	/** @brief Returns Tcount value from timer (DPU??).

		Actullay I do not understand what it means.  This method will be changed in future.
	*/
	uint16_t getTcountDpu(){
		return getU16(7);
	}

	/**@deprecated Use getMarkerDpu()
	*/
DEPRECATED
	double getDpuTimeMarkerInMsec(){
		int t = getDpuTime();
		int dt = getTcountDpu();
		double acTime = t*31.25 + dt * 2.048;
		return acTime;
	}
	/**@brief Returns the marker time in DPU format.
	*/
	double getMarkerDpu(){
		int t = getDpuTime();
		int dt = getTcountDpu();
		double acTime = t*31.25 + dt * 2.048;
		return acTime;
	}

	uint16_t getEndIntegCycleTime(){
		return getU16(8);
	}

	/** @brief Returns Tcount value from timer (DPU??).

		Actullay I do not understand what it means.  This method will be changed in future.
	*/
	uint16_t getTcountAcq(){
		return getU16(9);
	}

        /** @brief Return the DPU time at the end of integration cycles.

                Return the DPU time in msec of the end of integration cycles.
                The calculation is done as follows.  This may be wrong, or true.... I'm not sure.
                <pre>
	                int t = getEndIntegCycleTime();
	                int dt = getTcountAcq();
	                double acTime = t*31.25 + dt * 2.048;
	                return acTime;
                </pre>
	@deprecated Use getObsTimeDpu()
        */
DEPRECATED
        double getDataAcquisTimeInMsec(){
                int t = getEndIntegCycleTime();
                int dt = getTcountAcq();
		double acTime = t*31.25 + dt * 2.048;
		return acTime;
	}

        /** @brief Return the DPU time at the end of integration cycles.

                Return the DPU time in msec of the end of integration cycles.
	*/
	double getObsTimeDpu(){
                int t = getEndIntegCycleTime();
                int dt = getTcountAcq();
		double acTime = t*31.25 + dt * 2.048;
		return acTime;
	}

        /** @brief Returns the Universal Time in msec.
            
            Actual formulation is
            <pre>
            getObtTimeInMsec() + ( getDataAcquisTimeInMsec() - getDpuTimeMarkerInMsec()  )
            </pre>

	    You can also find the time related issue in @link UniversalTimeSample @endlink .
	@deprecated Use getObsTimeObt();

         */
DEPRECATED
	double getEpochTimeInMsec(){
          double offset= getDataAcquisTimeInMsec() - getDpuTimeMarkerInMsec();
          if(offset<0)offset+=31.25*0x10000;
          return getObtTimeInMsec() + offset;
	}

        /** @brief Returns the observation time in OBT.
	*/
	double getObsTimeObt(){
          double offset= getObsTimeDpu() - getMarkerDpu();
          if(offset<0)offset+=31.25*0x10000;
          return getMarkerObt() + offset;
	}

	/**@brief Return the universal time in seconds [Origin is the UNIX epoch time as time_t].

	    You can also find the time related issue in @link UniversalTimeSample @endlink .

	 */
	double getUniversalTime(TimeConversion* tconv){
		if(tconv==NULL){
			return getObsTimeObt()/1000;
		}
		return tconv->obt2utc(getObsTimeObt());
	}

        /** @brief Returns the Universal Time in msec.
            
            Actual formulation is
            <pre>
            getObtTimeInMsec() + ( getDataAcquisTimeInMsec() - getDpuTimeMarkerI
nMsec()  )
            </pre>
	    @deprecated Use getEpochTimeInMsec() or getUniversalTime()
         */
	DEPRECATED
        double getUniversalTimeInMsec(){
          double offset= getDataAcquisTimeInMsec() - getDpuTimeMarkerInMsec();
          if(offset<0)offset+=31.25*0x10000;
          return getObtTimeInMsec() + offset;
        }





	/** @brief Returns 8S cycle counter (word-10)
	*/
	uint16_t get8sCycleCntr(){
		return getU16(10);
	}

	/** @brief An alias for get8sCycleCntr.
	*/
	uint16_t getPacketCounter(){
		return get8sCycleCntr();
	}
	
	/** @brief Binning parameter (word_11)

		Returns the binning parameters.
		- Bit 15-8 Post Acc Index
		- Bit 7-0 Mass bins
	*/
	uint16_t getBinningPrm1(){
		return getU16(11);
	}
	
	uint8_t getPostAccIdx(){
		return (uint8_t)((0xff00 & getBinningPrm1())>>8);
	}

	uint8_t getMassBins(){
		return (uint8_t)(0x00ff & getBinningPrm1());
	}
	
	
	/** @brief Binning parameter (word_12)

		Returns the binning parameters.
		- Bit 15-8 Deflection bins
		- Bit 7-0 Energy bins
	*/
	uint16_t getBinningPrm2(){
		return getU16(12);
	}
	
	uint8_t getDeflectionBins(){
		return (uint8_t)((0xff00 & getBinningPrm2())>>8);
	}

	uint8_t getEnergyBins(){
		return (uint8_t)(0x00ff & getBinningPrm2());
	}
	
	/** @brief number of cycles for integration (word-13)
	*/
	uint16_t getNrCycles(){
		return getU16(13);
	}

	/** @brief modes (word-14)
		MSB is Swim status byte.  (See 9.3.1 on DS-0005 (SWIM SICD)).
		<pre>
		B15-8: Swim status byte 
			8: not used
			9: mode: 0=idle, 1=sampling
			10: selected memory bank
			11: HVPS oscillator status: 0=disabled, 1=enabled
			12: TDC latch up status: 0=TDC working, 1=TDC shut down (latch up protection enabled)
 			13: TDC latch up protection setting: 0=disabled, 1=enabled
			14: TDC power status: 0=off, 1=on
 			15: not used
		</pre>
		Returns the modes.  Logical bit, 1 as ON, 0 as OFF.
		<pre>
		B7 SWIM on/off
		B6 CENA on/off
		B5 Compression EN/DS
		B0 Health check event indication
		</pre>
	*/
	uint16_t getModes(){
		return getU16(14);
	}

	bool isTdcPowerOn(){
		uint16_t mode=getModes();
		uint16_t mask=0x4000;	// 14 bit
		return (mode & mask) != 0;		
	}

	bool isLatchUpProtectOn(){
		uint16_t mode=getModes();
		uint16_t mask=0x2000;	// 13 bit
		return (mode & mask) != 0;		
	}

	bool isLatchUp(){
		uint16_t mode=getModes();
		uint16_t mask=0x1000;	// 12 bit
		return (mode & mask) != 0;		
	}

	bool isHvpsOscOn(){
		uint16_t mode=getModes();
		uint16_t mask=0x0800;	// 11 bit
		return (mode & mask) != 0;		
	}

	bool isSelectedMemoryBank(){
		uint16_t mode=getModes();
		uint16_t mask=0x0400;	// 10 bit
		return (mode & mask) != 0;		
	}

	bool isModeSampling(){
		uint16_t mode=getModes();
		uint16_t mask=0x0200;	// 9 bit
		return (mode & mask) != 0;		
	}

	bool isSwimOn(){
		uint16_t mode=getModes();
		uint16_t mask=0x0080;	// 7 bit
		return (mode & mask) != 0;
	}
	
	bool isCenaOn(){
		uint16_t mode=getModes();
		uint16_t mask=0x0040;	// 6 bit
		return (mode & mask) != 0;
	}

	bool isSwimCompOn(){
		uint16_t mode=getModes();
		uint16_t mask=0x0020;	// 5 bit
		return (mode & mask) != 0;
	}

	bool isHealthCheckEvent(){
		uint16_t mode=getModes();
		uint16_t mask=0x0001;	// 0 bit
		return (mode & mask) != 0;
	}

	DEPRECATED
	SwimSubcHkPacket* getSubcomHkPac();

	SwimFastHkPacket* getFastHkPac(){
		uint8_t* str=getStream();
		SwimFastHkPacket *pac=new SwimFastHkPacket(str+28);
		delete[] str;
		return pac;
	}

	/** @brief Checks if there is RAW data available.
	
	        The implemented test is theoretically not 100% collision free: It tests for the presence of a 
			bbbb 3333 header at the location where the accumulation total matrix starts. As these 
			values are valid for counter values a small probability of collision exists.
	*/
	bool hasRawModeData(){
		bool isRaw=(getU16(22) == 0xbbbb && getU16(23) == 0x3333);
		return isRaw;
	}

	/** @brief Returns the pointer to the Swim RAW data array containing 8 sec worth of raw data.
	
			Calls hasRawModeData() internally to ensure RAW data is present in the packet. 
			If no RAW data is present then NULL is returned.
			
	*/
	SwimRawModeData* getRawModeData(){
		SwimRawModeData* rawd = NULL;
		if (this->hasRawModeData()) {
			uint8_t* str=getStream();
			rawd=new SwimRawModeData(str+44); // as in getAccumTotMatx() but skip 0xbbbb
			delete[] str;
		}
		return rawd;
	}


	/** @brief Returns the pointer to the Swim accumulation matrix of total count.
	
	@todo	Confirm the boundary.  Due to inconsistency between document and test vector,
			I do not yet confirm that the SwimAccumTotMatx is correct. 
			Currently, this problem is avoided by added 2byte for the calculation of
			offset to the science matrix.  This should be replaced by more sophisticated way.[2007-11-21]
			
			Prior calling this function the presence of RAW data must be excluded. RAW data presence 
			can be checked with hasRawModeData().
			
	*/
	SwimAccumTotMatx* getAccumTotMatx(){
		int n_e = getEnergyBins();
		int n_d = getDeflectionBins();
		uint8_t* str=getStream();
		SwimAccumTotMatx* matx=new SwimAccumTotMatx(str+42, n_e, n_d);
		delete[] str;
		return matx;
	}

	SwimMatrix* getSwimMatrix(){
		size_t offset= offsetByteToSwimMatrix();
		
		uint16_t fullsize=this->getSize();
		uint8_t* str=getStream();
		
		SwimMatrix* matx=new SwimMatrix(str+offset, fullsize-offset);
		
		delete[] str;
		return matx;
	}

	SwimSciAcc* getSwimSciAcc(){
		if(isMatrixCompressed()){
			fprintf(stderr, "Science Matrix is still Compressed.  Use decompress() first.  (%s:%d)\n", __FILE__, __LINE__);
			throw "Science Matrix is still compressed.";
		}
		SwimMatrix *matx=getSwimMatrix();
		SwimSciAcc* sci=(SwimSciAcc*)(matx->interpret(getEnergyBins(), getDeflectionBins(), getMassBins()));
		delete matx;
		return sci;
	}

	/** @brief Returns true if the SwimMatrix is compressed.  (independent of DPU flag of SwimCompOn)
	
		Check whether the SwimMatrix is compressed.  If the SwimMatrix stream is compressed, true is
		returned.  If user explicitly decompressed by decompress() method, the returned value is changed.
	*/
	bool isMatrixCompressed(){
		SwimMatrix* sMat=getSwimMatrix();
		uint8_t* cMatB=sMat->getStream();
		delete sMat;
		
		bool isC=(cMatB[0]==0xaa && cMatB[1]==0x55 && cMatB[2]==0xaa && cMatB[3]==0x55);

		delete[] cMatB;
		return isC;
	}
	
	/**@brief Check whether the packet format is valid for SwimPacket.
	
		Check items are:
		- First 4 byte is 0xddddeeee
		- SwimMatrix starts with 0xaa55aa55 or 0xaaddaadd
		- OBT time is less than 1 year! (Actually it should be less than 21 days, I guess -2009-01-24-)

	@todo OBT time check is proper???
	*/
	bool isValid(){
		uint8_t* mat=getStream();
		bool test1=(mat[0]==0xdd && mat[1]==0xdd && mat[2]==0xee && mat[3]==0xee);
		if(!test1)return false;
		delete[] mat;
		
		SwimMatrix* sMat=getSwimMatrix();
		uint8_t* cMatB=sMat->getStream();
		delete sMat;

		bool test21=(cMatB[0]==0xaa && cMatB[1]==0x55 && cMatB[2]==0xaa && cMatB[3]==0x55);
		bool test22=(cMatB[0]==0xaa && cMatB[1]==0xdd && cMatB[2]==0xaa && cMatB[3]==0xdd);
		if(!test21 && !test22)return false;
		
		delete[] cMatB;

		// OBT time check.  Sometimes very strange number.
		double obt=getObsTimeObt();
		double maxobt=365*86400.*1000.;

		if(obt>maxobt)return false;
		
		return true;
	}
	/**@brief Debug verbosity when reporting errors during decompression
	*/
	static const int VERBOSE_DEBUG=3;
	/**@brief Normal verbosity when reporting errors during decompression
	*/
	static const int VERBOSE_NORMAL=2;
	/**@brief Minimum verbosity when reporting errors during decompression.
	*/
	static const int VERBOSE_MIN=1;
	/**@brief No verbosity when reporting errors during decompression.
	*/
	static const int VERBOSE_NONE=0;
	
	/**@brief Returns newly created, rice-decompressed SwimPacket.
	
	Returns the decompressed packet which is newly created.  The HK flag isSwimCompOn() of newly generated packet is not changed.
	If decompression is already done, the copy of the packet is returned.
	If decompression failed, NULL is returned.
	
	The parameter defines the level of verbosity on stderr. 
	*/
	SwimPacket* decompress(int verbose);

	
	/**@brief Returns newly created, rice-decompressed SwimPacket with verbosity == VERBOSE_NORMAL
	*/

	SwimPacket* decompress(){	
		return decompress(VERBOSE_NORMAL);
	}	
	
	void setUtc(double utc){
		_savedUtc=utc;
	}
	double getUtc(){
		if(_savedUtc==0)throw std::runtime_error("UTC not calculated.  Use getUniversalTime(TimeConversion*)");
		return _savedUtc;
	}

	int getQualityFlag(){
		return _qflag;
	}

	void setQualityFlag(int qflag){
		_qflag|=qflag;
	}

	SwimPacket* newInstance(){
		uint8_t* stream=getStream();
		size_t size=getSize();
		SwimPacket* p=new SwimPacket(stream, size);
		p->_savedUtc=_savedUtc;
		p->_qflag=_qflag;
		delete[] stream;
		return p;
	}


private:
	uint16_t getU16 ( int wordId0 ){return IrfPac::SaraPac::getWordAt(*this, wordId0);};
	uint32_t getU32 ( int wordId1 ){return IrfPac::SaraPac::getWordAt(*this, wordId1);};
	uint32_t getU32 ( int wordId0, int wordId1 ){return (getU32(wordId0) << 16)+getU32(wordId1);};
	uint64_t getU64 ( int wordId3 ){return IrfPac::SaraPac::getWordAt(*this, wordId3);};
	uint64_t getU64 ( int wordId2, int wordId3  ){return (getU64(wordId2) << 16)+getU64(wordId3);};
	uint64_t getU64 ( int wordId1, int wordId2 , int wordId3  ){
		return (getU64(wordId1) << 32)+(getU64(wordId2) << 16)+getU64(wordId3);};
	uint64_t getU64 ( int wordId0, int wordId1 , int wordId2 , int wordId3  ){
		return (getU64(wordId0) << 48) + (getU64(wordId1) << 32)+
			(getU64(wordId2) << 16)+getU64(wordId3);};

	/**@brief Returns the offset to the SwimMatrix.
	
		Note that the offset looks like different from the document.
		You need to add 1 (or 2) for padding if the ExDx5 is odd (or even).
	*/
	size_t offsetByteToSwimMatrix(){
		size_t offset=42;
		int n_e=getEnergyBins();
		int n_d=getDeflectionBins();
		offset += n_e*n_d*5;

		if(offset %2 == 0){
			offset+=2;    // THIS PADDING IS NOT WRITTEN IN THE DOCUMENT.
		}
		else{
			offset+=1;
		}

		return offset;
	
	}


};

/** @brief Array of SWIM packet.
*/
class SwimPacketArray {
private:
	std::vector<SwimPacket*> _arr;
public:
	~SwimPacketArray(){
		destroy();
	}
	void append(SwimPacket* pac){
		_arr.push_back(pac->newInstance());
	}
	/**@brief Get the newly generated SwimPacket* of i-th contents.

		You have to delete the returned pointer explicitly.
	*/
	SwimPacket* getPacket(int i){
		return (_arr.at(i))->newInstance();
	}
	/**@brief Get the i-th packet.
		
	@deprecated Use #getPacket(int) because this method can break the contents inside the array.
	*/
	DEPRECATED
	SwimPacket* get(int i){
		return _arr.at(i);
	}
	size_t size(){
		return _arr.size();
	}
	/**@brief Get the UTC timestamp of the i-th packet without copying the packet.
		
		This is useful when searching for a packet with a specific UTC without the need to copy each packet.
	 */
	double getPacketUtc(int i){
		return (_arr.at(i))->getUtc();
	}

	/**@brief Get the quality of the i-th packet without copying the packet.
		
		This is useful when searching for a packet with a specific quality without the need to copy each packet.
	 */
	int getPacketQualityFlag(int i){
		return (_arr.at(i))->getQualityFlag();
	}


	/**@brief Clear the array after destroy all the packet contained.  You don't need to call it directly (ie called from Destructor).
	 */
	void destroy(){
		std::vector<SwimPacket*>::iterator it=_arr.begin();
		while(it!=_arr.end()){
			SwimPacket* pac=(*it);
			delete pac;
			it++;
		}
		_arr.clear();

	}

};


// End of namespace
}}





#endif

