#ifndef __IRAPAC_SARAPAC_SWIMRAWMODEDATA_H
#define __IRAPAC_SARAPAC_SWIMRAWMODEDATA_H


/*
 *  SwimRawModeData.h
 *  libsadppac
 *
 *  Created by Martin Wieser on 09/12/02.
 *  Copyright 2007 __MyCompanyName__. All rights reserved.
 *
 */
#include<IrfPac/irfpac.h>
#include<linlog++.h>

namespace IrfPac{
namespace SaraPac{

class SwimRawModeDataDump;   // Forward declaration

/** @brief RAW mode data

	The size is described as (4+8+4080)*254 bytes.
*/
class SwimRawModeData : public PacketU8Arr {

public:

	static const int RAW_HEADER_BYTES=4;
	/**@brief Number of bytes used for tof events in RAW packet..
	*/
	static const int RAW_COUNTER_BYTES=8;
	/**@brief Number of bytes used for counters in RAW packet.
	*/
	static const int RAW_TOF_EVENT_BYTES=4080;
	/**@brief Number of bytes used for tof events in RAW packet.
	*/

	static const int RAW_PACKET_SIZE=RAW_HEADER_BYTES+RAW_COUNTER_BYTES+RAW_TOF_EVENT_BYTES;
	/**@brief Number of bytes used for a complete RAW packet.
	*/
	static const int RAW_PACKETS=0xfe;
	/**@brief Number of bytes used for a complete RAW packet.
	*/

	
	/** @brief A constructor of SWIM raw mode data array 
	
		size is (4+8+240)*254
	*/
	SwimRawModeData(uint8_t* stream);
	virtual ~SwimRawModeData();

	/**@brief Dump the contents using default formatter (SwimRawModeDataAllDump).
	 */
	virtual void dumpAll(FILE* fp);
	virtual void dump(FILE* fp, SwimRawModeDataDump* dump);


	/** @brief Returns the offset byte of the n'th RAW packet.
	 */
	virtual uint8_t getRawByte(int slot, int offset){
	    int index = slot*(RAW_PACKET_SIZE)+offset;	
		return getValue(index);
	}

	/** @brief Returns the slot ID (8bit) for the n'th RAW packet.
 
	 */
	virtual uint8_t getSlotId(int slot){
	    int index = slot*(RAW_PACKET_SIZE)+3;	
		return getValue(index);
	}

	/** @brief Returns the number of RAW packets available.
	  The highest slot id is number of RAW packets-1	
	  
	  @todo add some checks to validate the number?
	 */
	virtual int getMaxRawPacket(){
		return RAW_PACKETS;
	}

	/** @brief Returns the number of RAW packets available.
	  The number of TOF events in a RAW packet
	  
	  @todo add some checks to validate the number?
	 */
	virtual int getMaxTofEvents(){
		return (RAW_TOF_EVENT_BYTES);
	}


	/** @brief Returns Start counter value of the slot'th RAW packet.
	*/
	virtual uint16_t getStart(int slot){
		uint16_t cntr;
		cntr = ((uint16_t)getValue(slot*RAW_PACKET_SIZE + RAW_HEADER_BYTES+0) << 8) + 
		       (uint16_t)getValue(slot*RAW_PACKET_SIZE + RAW_HEADER_BYTES+1);	
		return cntr;
	}

	/** @brief Returns Stop counter value of the slot'th RAW packet.
	*/
	virtual uint16_t getStop(int slot){
		uint16_t cntr;
		cntr = ((uint16_t)getValue(slot*RAW_PACKET_SIZE + RAW_HEADER_BYTES+2) << 8) + 
		       (uint16_t)getValue(slot*RAW_PACKET_SIZE + RAW_HEADER_BYTES+3);	
		return cntr;
	}

	/** @brief Returns Over counter value of the slot'th RAW packet.
	*/
	virtual uint16_t getOver(int slot){
		uint16_t cntr;
		cntr = ((uint16_t)getValue(slot*RAW_PACKET_SIZE + RAW_HEADER_BYTES+4) << 8) + 
		       (uint16_t)getValue(slot*RAW_PACKET_SIZE + RAW_HEADER_BYTES+5);	
		return cntr;
	}

	/** @brief Returns Under counter value of the slot'th RAW packet.
	*/
	virtual uint16_t getUnder(int slot){
		uint16_t cntr;
		cntr = ((uint16_t)getValue(slot*RAW_PACKET_SIZE + RAW_HEADER_BYTES+6) << 8) + 
		       (uint16_t)getValue(slot*RAW_PACKET_SIZE + RAW_HEADER_BYTES+7);	
		return cntr;
	}

	/** @brief Returns the i_tof'th TOF value of the slot'th RAW packet.
	*/
	virtual uint8_t getTof(int slot, int i_tof){
		uint8_t tof = getValue(slot*RAW_PACKET_SIZE + RAW_HEADER_BYTES+RAW_COUNTER_BYTES+i_tof);
		return tof;
	}


private:
	SwimRawModeDataDump* _dump; //<! Default dump class

};

/** @brief Pure virtual class for formatting the contents of the SwimRawModeData.
    
    Dumping the contents is always the issue for this kind of library.
    As an trial implementation, I made a pure virtual class of SwimRawModeDataDump.
    This class is intened to be extended to a concrete class.
    The instance of one concrete class will be added to SwimRawModeData class,
    and the SwimRawModeData::dumpAll(FILE*) method will dump the contents.
 */
class SwimRawModeDataDump {
public:
	virtual ~SwimRawModeDataDump(){}
	virtual void dump(FILE* fp, SwimRawModeData* matx)=0;
};
/** @brief Implementation of HEX dumping class.
 */
class SwimRawModeDataHexDump : public SwimRawModeDataDump {
public:
	void dump(FILE* fp, SwimRawModeData* matx){
		matx->dumpHexStream(fp);
	}
};

/** @brief Implementation of All Dump.
 */
class SwimRawModeDataAllDump : public SwimRawModeDataDump {
public:
	void dump(FILE* fp, SwimRawModeData* matx);
};


}}  // namespace


#endif
