/*
 *  SwimSciAcc.h
 *  libsadppac
 *
 *  Created by futaana on 08/03/10.
 *  Copyright 2008 __MyCompanyName__. All rights reserved.
 *
 */

#ifndef __IRFPAC_SARAPAC_SWIMSCIACC_H
#define __IRFPAC_SARAPAC_SWIMSCIACC_H

#include<sstream>

#include"IrfPac/SaraPac/SwimScienceDataPacket.h"

namespace IrfPac{
namespace SaraPac{


/**@brief Science data of SWIM.

*/
class SwimSciAcc : public SwimScienceDataPacket {
public:
	SwimSciAcc(uint8_t *pac, int nEnergy, int nDefl, int nMass);
	virtual ~SwimSciAcc(){};

	/**@brief Create a 16x16 reformed science matrix for SWIM.
	   @return New packet that has E=16 D=16 matrix.  The returned object is a instance of SwimSciAcc16x16.
	   @throw std::runtime_error when this method is applied non-maximum mode (E=32,D=8) data.
	   @sa SwimSciAcc16x16
	*/
	SwimSciAcc* reform16x16();
	
	/**@brief Returns the raw count in the specified energy, deflection and mass.
	 */
	virtual uint8_t getRaw(int iE, int iD, int iM);
	/**@brief Returns the lin-log decompressed count in the specified energy, deflection and mass.
	 */
	virtual uint16_t get(int iE, int iD, int iM);

	/**@brief Returns the total count rate over all the energy, deflection and mass channel.
	 */
	virtual double collapseAll(){
		double cnt=0;
		for(int ie=0; ie<_ne; ie++){
			for(int id=0; id<_nd; id++){
				for(int im=0; im<_nm; im++){
					cnt+=get(ie,id,im);
				}
			}
		}
		return cnt;
	}

	/**@brief Returns the total count rate of specified mass bins summed over all the energy and deflection.
	 */
	virtual double collapseED(int im){
		double cnt=0;
		for(int ie=0; ie<_ne; ie++){
			for(int id=0; id<_nd; id++){
				cnt+=get(ie,id,im);
			}
		}
		return cnt;
	}

	/**@brief Returns the total count rate of specified energy bins summed over all the deflection and mass.
	 */
	virtual double collapseDM(int ie){
		double cnt=0;
		for(int id=0; id<_nd; id++){
			for(int im=0; im<_nm; im++){
				cnt+=get(ie,id,im);
			}
		}
		return cnt;
	}

	/**@brief Returns the total count rate of specified deflection bins summed over all the energy and mass.
	 */
	virtual double collapseEM(int id){
		double cnt=0;
		for(int ie=0; ie<_ne; ie++){
			for(int im=0; im<_nm; im++){
				cnt+=get(ie,id,im);
			}
		}
		return cnt;
	}
	
	virtual void dumpAll(){dumpAll(stdout);};
	virtual void dumpAll(FILE*fp);
	
	virtual int getNrEne(){return _ne;}
	virtual int getNrDefl(){return _nd;}
	virtual int getNrMass(){return _nm;}
private:
	int _ne, _nd, _nm;

};

class SwimSciAccDump{
public:
	virtual ~SwimSciAccDump(){}
	virtual void dump(FILE* fp, SwimSciAcc* matx)=0;
};

class SwimSciAccESpectraDump : public SwimSciAccDump {
public:
	void dump(FILE* fp, SwimSciAcc* matx){
		int ne=matx->getNrEne();
		int emax=32;
		fprintf(fp, "%3d ", ne);
		for(int i=0; i<ne; i++){
			fprintf(fp, "%6ld ", (long)matx->collapseDM(i));
		}
		for(int i=ne; i<emax; i++){
			fprintf(fp, "   -1 ");
		}
		fprintf(fp, "\n");
	}
};

class SwimSciAccDSpectraDump : public SwimSciAccDump {
public:
	void dump(FILE* fp, SwimSciAcc* matx){
		int nd=matx->getNrDefl();
		int dmax=32;
		fprintf(fp, "%3d ", nd);
		for(int i=0; i<nd; i++){
			fprintf(fp, "%6ld ", (long)matx->collapseEM(i));
		}
		for(int i=nd; i<dmax; i++){
			fprintf(fp, "   -1 ");
		}
		fprintf(fp, "\n");
	}
};

/**@brief Dump with 16 directions summed over all the energy and mass.
*/
class SwimSciAccD16SpectraDump : public SwimSciAccDump {
public:
	void dump(FILE* fp, SwimSciAcc* matx){
		if ((matx->getNrDefl() == 8) && (matx->getNrEne() == 32)) {
			SwimSciAcc* matx16x16=matx->reform16x16();   // Reform the science matrix
			int nd=matx16x16->getNrDefl();
			int dmax=32;
			fprintf(fp, "%3d ", nd);
			for(int i=0; i<nd; i++){
				fprintf(fp, "%6ld ", (long)matx16x16->collapseEM(i));
			}
			for(int i=nd; i<dmax; i++){
				fprintf(fp, "   -1 ");
			}
			fprintf(fp, "\n");
			delete matx16x16;
		} else {
			int nd=matx->getNrDefl();
			int dmax=32;
			fprintf(fp, "%3d ", nd);
			for(int i=0; i<nd; i++){
				fprintf(fp, "%6ld ", (long)matx->collapseEM(i));
			}
			for(int i=nd; i<dmax; i++){
				fprintf(fp, "   -1 ");
			}
			fprintf(fp, "\n");
		}
	}
};

class SwimSciAccMSpectraDump : public SwimSciAccDump {
public:
	void dump(FILE* fp, SwimSciAcc* matx){
		int nm=matx->getNrMass();
		int mmax=32;
		fprintf(fp, "%3d ", nm);
		for(int i=0; i<nm; i++){
			fprintf(fp, "%6ld ", (long)matx->collapseED(i));
		}
		for(int i=nm; i<mmax; i++){
			fprintf(fp, "   -1 ");
		}
		fprintf(fp, "\n");
	}
};

/**@brief An implementation of 16x16-reformed SWIM science matrix.

   SWIM has several binning parameters.
   Energy and deflection bin numbers (and mass as well) can be set by commanding.
   However, the case of E=32 D=8 is not straight forward.
   In this case, even DPU says E=32 D=8, the actual observations is done by E=16 D=16 mode.
   I don't know in which document this fact is described, but in fact this is what it is.

   This libsadppac supports this functionality by using a object-oriented advantage.
   SwimSciAcc instance has a method to convert the packet to 16x16-reformed packet.
   This method returns the pointer to SwimSciAcc, but it is in fact an instance of SwimSciAcc16x16.
   However, users can use the returned instance as it is SwimSciAcc instance, obviously, because
   SwimSciAcc16x16 is an extended class of SwimSciAcc.

   Anyway, you can look at the example below.

   <pre>
   SwimPacket *spac=...;   // Assume that we have SwimPacket with E=32,D=8.
   SwimSciAcc *sciMatx=spac->getSwimSciAcc();   // Obtaining science matrix SwimSciAcc.
   cout << sciMatx->getNrEne() << endl;    // Energy step is 32
   cout << sciMatx->getNrDefl() << endl;   // Deflector bins is 8

   SwimSciAcc *sciMatx16x16=sciMatx->reform16x16();   // Reform the science matrix
   cout << sciMatx16x16->getNrEne() << endl;  // Energy step is 16
   cout << sciMatx16x16->getNrDefl() << endl;   // Deflector bins is 16

   delete spac;
   delete sciMatx;
   delete sciMatx16x16;
   </pre>

   Similar conversion should be done in {@link SwimAccumTotMatx} class.
 */
class SwimSciAcc16x16 : public SwimSciAcc {
public:
	SwimSciAcc16x16(uint8_t *pac, int nMass);
	~SwimSciAcc16x16(){};
	/**@brief Obtain raw count for 16x16 matrix.(not yet implemented...sorry)
	   @sa get(int,int,int)
	 */
	uint8_t getRaw(int aE, int aD, int iM);
};


}}

#endif


