#ifndef _IRFPAC_SARAPAC_TIMECONVERSION_H
#define _IRFPAC_SARAPAC_TIMECONVERSION_H

namespace IrfPac{
namespace SaraPac{

/*!  @brief  Virtual class of time conversion between OBT and UTC.
 */
class TimeConversion {
public:
	virtual ~TimeConversion(){}
	/**@brief Convert double-precision OBT value to double-precision UTC time.

	   OBT time is obtained from CenaPacket or SwimPacket.  The unit should be in milisec.
	   UTC time is the Unix epoch time (identical to time_t for integer part) with 3-rd order after the decimal point.
	 */
	virtual double obt2utc(double obt)=0;
};

/**@brief An implementation of OBT-UTC conversion.

   One way to calculate UTC from OBT is using offset.
   Because the OBT is 21-day periodic with msec unit,
   if you know the origin of the OBT in UTC, you can convert
   OBT to UTC at any time.

   The offset looks like as follows.
   Note that this estimation is done very primitive way,
   it is not guaranteed the results.
   You may have to use TimeConversionTimeFile class for precise calculation.
   <pre>
   OFFSET(0)    2008-10-21T23:57:03    1224633423
   OFFSET(1)    2008-11-11T23:57:03    1226447823
   OFFSET(2)    2008-12-02T23:57:03    1228262223
   OFFSET(3)    2008-12-23T23:57:03    1230076623
   OFFSET(4)    2009-01-13T23:57:03    1231891023    
   OFFSET(5)    2009-02-03T23:57:03    1233705423
   OFFSET(6)    2009-02-24T23:57:03    1235519823
   ...
   </pre>
 */
class TimeConversionOffset : public TimeConversion {
private:
	double _offset;   // Offset in seconds.
public:
	static double OFFSET(int iteration_number){
		return 1224633423 + 1814400*iteration_number;
		// 1224633423 is 2008-10-23T23:57:03
		// 1814400 is 21 day in seconds.
	}

	TimeConversionOffset(double offset){
		_offset=offset;
	}

	double obt2utc(double obt){
		double obts = obt/1000.;  // Convert to msec.
		return obts + _offset;
	}

};

}}



#endif
