#ifndef __IRFPAC_SARAPAC_TIMEPACKET_H
#define __IRFPAC_SARAPAC_TIMEPACKET_H

#include<iostream>
#include<ctime>
#include"IrfPac/irfpac.h"
#include"IrfPac/SaraPac/sarapac.h"
#include <time.h>
#include<stdexcept>


typedef double double64;

namespace IrfPac{
namespace SaraPac{

/**@brief A utility of converting specified time from UTC to UNIX epoch time.

   The returned value is the same unit as time_t, but support millisec resolution as double.

	@todo Maybe move to another library in future
*/
	double convert_utc2epoch(int yr, int mo, int dy, int hr, int mn, int sec, int msec);
	
/**@brief Return the epoch time string.

   The returned value has the format "2009-06-23T23:25:31.947Z"
*/
	std::string epoch2str(double utctime);

/**@brief Return the epoch time string in xmgrace compatible format.

   The returned value has the format "2009-06-23T23:25:31.947"
*/
	std::string epoch2iso8601str(double utctime);

/**@brief Return the unix epoch time of the string isotime.

   The parameter isotime  has the format "2009-06-23T23:25:31.947. An invalid string creates an exception"
*/
	double iso8601str2epoch(std::string isotime);
	

/**@brief Contents of times file.

   Time file is provided from CY-1 project.
   The file contains the time conversion information, each of which corresponds
   to one packet.

   Packet structure is described by Structure(OBT-UTC conversion).doc, which is
   also provided from project.

   The packet constitute of a structure defined by
   <pre>
   typedef struct
   {
        int OrbitNo;   // 32bit
        long long DHOBT_l;   // 64bit -> 40 bit DH-OBT
        int DHOBT_i4;   // 32bit -> 32 bit DH-OBT
        double BMUOBT_d;  // 64bit -> BMU OBT
        int UTCTime_7i4[7];  // 32bit x 7  UTC time yr, mon, dy, hr, min, sec, .1 msec
        unsigned char BMUIndication_c;  // 8bit
   }TMRecord;
   </pre>

   One packet consistitue of 53 byte.  For SARA, we use BMUOBT_d and UTCTime_7i4 only.

 */
class TimePacket : PacketU8Arr {
private:
//  uint32_t _orbit_nr;
//  uint64_t _dhuobt_l;
//  uint32_t _dhobt_i;
//  double64 _bmuobt_d;
//  uint32_t _utctime_7i4[7]; 
//  uint8_t _bmuindication;

public:
  static const int PACKET_SIZE=53;

  TimePacket(uint8_t *pac) : PacketU8Arr(pac, PACKET_SIZE){
	if(sizeof(double64) != 8){
		fprintf(stderr, "### Fatal...\n");
		fprintf(stderr, "### Size of double is not 64 bit!!!\n");
		fprintf(stderr, "### In IrfPac/SaraPac/TimePacket.h, double64 type should be changed accordingly (e.g. long double).\n");
		fprintf(stderr, "### Aborting...\n");
	}
  }
  ~TimePacket(){}

  double getObtTimeDouble(){// from 16 to 23 byte.
    uint8_t* str=getStream();

    union {
      uint8_t _i[8];
      double64 _d;
    } _conv;

    for(int i=16; i<24; i++) _conv._i[i-16] = str[i];
    double64 d=_conv._d;

    delete[] str;
    return (double)d;
  }


  /**@brief Return the GMT in UNIX epoch time expression.

   The returned value is the same unit as time_t, but support millisec resolution as double.
   */
  double getGmt(){  // from 24 to 51 byte
    uint8_t* str=getStream();
    union {
      uint8_t _i[28];
      uint32_t _o[7];
    } _conv;

    for(int i=0; i<28; i++) _conv._i[i] = str[24+i];
    int yr=_conv._o[0];
    int mo=_conv._o[1];
    int dy=_conv._o[2];
    int hr=_conv._o[3];
    int mn=_conv._o[4];
    int se=_conv._o[5];
    int ms=_conv._o[6];

    return convert_utc2epoch(yr, mo, dy, hr, mn, se, ms/10);
    
  }


};



}}
#endif

