/**@mainpage
	@section Introduction
	@par
	This project is to implement a library packet analysis of SARA
	in C++ in library format.
	The library can be used as well from Python and Java through SWIG.
	@par
	The purposes of this library is @link AnalysisSara 'analyze the CENA/SWIM data' @endlink and
	'convert packet format in between'.  'Time conversion' is also supported in this library.

	@subsection Analysis Analysis of SARA
	@par
	Frequently data types are
	- IrfPac::SaraPac::CenaPacket for analysis of CENA data.
	- IrfPac::SaraPac::SwimPacket for analysis of SWIM data.
	@par
	Each has container
	- IrfPac::SaraPac::CenaPacketArray for CENA data array
	- IrfPac::SaraPac::SwimPacketArray for SWIM data array
	@par
	For users who want to learn how to analyse SARA data, see @link AnalysisSara @endlink for details.
	@subsection Format Format of SARA data
	@par
	For space mission, there are a bunch of data file format is available.
	In this library, the below files are supported.
	- BMU formatted SARA packet   is the ascii-bmu format of SaraPacket.  896 byte fixed.
	- BMU formatted CENA packet   is the ascii-bmu format of CenaPacket.  4-sec data is 1 packet.
	- BMU formatted SWIM packet   is the ascii-bmu format of SwimPacket.  8-sec data is 1 packet.
	- Space packet                is the format provided from ISSDC.
	@par
	For the conversion, you can see the document of @link ConversionPacketFormat @endlink.
	@section TCNV Time conversion
	@par
	Time conversion is one critical issue of this kind of packet converting tool.
	Here the description is written.  Note that this feature may be deprecated
	due to the change of the mission side, or upgrading SARA team's knowledgement.
	@subsection OBT2DPU Synchronization of time between s/c and SARA
	To synchronize sensor (SARA) and spacecraft, there are two time definitions.
	- OBT time is generated by spacecraft
	- DPU time is generated by SARA/DPU
	Time tag for each packet is done by DPU when DPU received the packet.
	This is obtained by CenaPacket::getObsTimeDpu().
	To obtain OBT-DPU synchronization, it is needed to compare between getMarkerObt() and getMarkerDpu().
	Considering all above, getObsTimeObt() will return the observation (end) time in the reference of OBT in msec.
	@subsection OBT2UTC Conversion of time from OBT to UTC.
	Conversion from OBT to UTC is another difficulty.  Two ways of implementations are (will be) provided.
	- Using time file.  (Official way)
	- Using offset of the OBT epoch.
	Sample code is found in @link UniversalTimeSample @endlink
	@section MEM Pointer and memory
	@par
	Memory management is always the problem for C/C++ library.
	In this library, the pointer is used everywhere.
	This is because the support of the reference will increase the work of the library developer
	considering memory leak and memory allocation or 'Bus error' or copy constructor etc...
	I do not want to support it: The neverending story.
	Therefore, the user is recommended NOT TO USE the reference.
	There are many methods which return the pointer to object.
	Normally, the pointer points the object which is 'new'-ed in the methods.
	This means that the users must destroy the pointer explicitly by
	delete or delete[].

**/



/*
	@par
	The IRF packet is intended to be used as a project-wide purpose.
	All the classes are in the namespace of IrfPac.
	General packets are implemented as a GeneralPacket class
	with minimul methods.  Normally you can use UInt8Arr and its public
	extended classes.
	@par
	Classes and methods can be used after IrfPac/irfpac.h is included.
	You need to link libirfpac.a.
	@par
	SARA DPU output dedicated packets are implemented under the IrfPac::SaraPac namespace.
	Classes and their methods can be used after IrfPac/irfpac.h and
	IrfPac/SaraPac/sarapac.h is included.
	You also need to link libirfpac.a and libsarapac.a.
	@par
	You can download the IRF-SARA packet library from SARA subversion repository.
	- https://username@butler.irf.se/sara/DPU/libsadppac
	@par
	The IRF-SARA libaray depends on two other project.
	- liblinlog : Log compression library
	- librice : Rice compression library
	@par
	Both project relates to compression scheme.
	You can download these library from SARA subversion repository.
	- https://username@butler.irf.se/sara/COMMON/liblinlog/trunk
	- https://username@butler.irf.se/sara/COMMON/librice/trunk
	@section CLT Command line tools 
	@par
	Command line tools are installed into ${HOME}/SARA/bin as a default.
	@subsection SaraDemultiplex
	@par
	When you want to demultiplex the ascii format (BMU format) data,
	you can use SaraDemultiplex command.
	@par
	<pre>
	SaraDemultiplex   ascii_format_filename
	</pre>
	@par
	<dfn>ascii_format_filename</dfn>.sarapkg is created.
	@section SampleCode Sample code
	@par
	Sample code of demultiplexing is found in
	@link DemultiplexorSample @endlink.
	
*/


/** @page UniversalTimeSample

	@section UTCsample How to get UT

	@subsection UTCsampleUSER For user
	@par
	Normally analysis-ers start from 'BMU formatted CENA/SWIM data'.
	In this case, it is very simple to get UTC with getUtc() method.
	You may also need to check the quality flag to know the quality of time conversion.
	@code
	CenaBmuReader reader;
	reader.readFile("XXXXXX_cena.bmu");    //  Read the data from BMU format file
	CenaPacketArray* cenaArray = reader.getArray();  // Return the pointer to the array of CenaPacket.

	CenaPacket* cenaPacket = cenaArray.getPacket(30);  // You can get 30th packet in the array.

	printf("%f  %d\n", cenaPacket->getUtc(), cenaPacket->getQualityFlag());

	delete cenaPacket;
	delete cenaArray;
	@endcode

	@subsection UTCproblem Problems
	@par
	It is important to understand what causes the UTC conversion error.
	There are two main problems which have been understood.
	@par
	UTC is calculated from @link TIMEfile time file @endlink.
	The file is provided from ISSDC corresponding to the time range of the science data, however, sometimes the coverage of the time file does not coincide with that of the science data.
	This happenes during the conversion from BMU formatted SARA packet to BMU formatted SWIM/CENA packet.
	In this case, the time of the packet is calculated from the extrapolation, but set the QFLAG_TIMESTAMP_NG flag to the product.
	@par
	Another problem is the jump of the spacecraft clock.
	Spacecraft and SARA/DPU has different internal clock, and both are synchronized every 32 sec (TBC).
	When the synchronization occurred, both time stamp are recorded in the packet contents (getObtTime() for spacecraft and getDpuTime() in the CenaPacket and SwimPacket).
	However, sometimes synchronization timestamp of the onboard clock exhibits quite a jump.
	The OBT jumps make jumps in the observation time.
	The corrections are not yet supported, we have ideas.  A quality flag QFLAG_OBTTIMEJUMP_NG will be set to the product.
	@subsection UTCsampleDEV For developer
	@par
	Because the science data does not contain UTC but internal clock, one need to convert the internal clock to UTC.
	There are two ways of obtaining universal time.
	The official way is using the time file.
	For very conventional use, you may also use offset for OBT time (note: this is NOT recommended).
	Here, these two ways are realized using 'algorithm' design-pattern.
	You may instance either TimeConversionTimeFile class or TimeConversionOffset class up to your favorite.
	Because both of them has the same pure virtual parent class called TimeConversion,
	you can always use CenaPacket::getUniversalTime() method in your application whichever the method you want to use (even at runtime).
	@subsection TIMEfile Using time file
	@par
	Time file is provided by the project.  It contains the conversion from OBT time to UTC.
	@par
	You may instance TimeFile class to use the time file.
	@code
	#include <IrfPac/SaraPac/AscFmt/ascfmt.h>
	using namespace IrfPac::SaraPac::AscFmt;
	FILE* fp=fopen(filename.time, "rb");
      	TimeFile* tf=new TimeFile();
	tf->readFile(tf);
	@endcode
	Then, conversion will be done by TimeConversionTimeFile class.
	@code
	TimeConversionTimeFile* tconv=new TimeConversionTimeFile(tf);
	@endcode
	You may call getUniversalTime() method with TimeConversionTimeFile instance.
	@code
	double t=cena_packet -> getUniversalTime(tconv);
	cout << t << endl;
	cout << epoch2str(t) << endl;
	@endcode
	Here t will be the time of the (end of) the observation expressed in the UNIX epoch time (c.f. same as the result of time() function in time.h)
	The resolution is msec.
	You can use IrfPac::SaraPac::epoch2str() function to convert the time to human readable format.
	@subsection OFFset Using offset
	@par
	If you know the offset (origin) of the OBT in UTC, you can know the UTC at any time.
	Actually because OBT is cyclic in 21 days, the offset is also be set every 21 days.
	This is not enough time scale for long-term analysis, but, nothing to do...
	In addition, because the offset is not clearly documented anywhere and unofficial, this will not be used for long-term perspective.
	@par
	Ok, first, you may need to instance TimeConversionOffset class.
	@code
	import <IrfPac/SaraPac/sarapac.h>
	using namespace IrfPac::SaraPac;
	TimeConversionOffset* tconv = new TimeConversionOffset(TimeConversioOffset::OFFSET(0))
	@endcode
	The constructor of TimeConversionOffset need one argument, i.e., the unix epoch time of the origin.
	The origin can be found OFFSET static function, which also need the argument, i.e., number of cyclic iteration in 21 day.
	In the sample's case, the offset is 2008-10-23T23:57:03.  If you set OFFSET(1), the offset will be 2008-11-11T23:57:03.
	@par
	Then, you can call getUniversalTime() method.
	@code
	double t=cena_packet -> getUniversalTime(tconv);
	cout << t << endl;
	cout << epoch2str(t) << endl;
	@endcode
	This last part is completely transparent if you use either time file or offset.
*/

/** @page AnalysisSara
    @section Archive
    @par
    It is not yet clear how to archive the dataset, but I have a recommendation to use CENA-BMU or SWIM-BMU format.
    The data is currently located at butler:/irf/sara/saradds/archive/bmu and the file name is *.cena and *.swim.

    @section Sample code in C++
    @par
    The sample code is sample/CenaBmuReader_sample.cpp and compilation is as follows:
    @code
    % g++ -o CenaBmuReader sample/CenaBmuReader_sample.cpp -I${HOME}/SARA/include -L${HOME}/SARA/lib -lirfpac -lsarapac -lascfmt -llinlog -lm
    % ./CenaBmuReader  /irf/sara/saradds/archive/bmu/SRA01D18CH101116003709039223301229.cena
    @endcode
    
 */




/*
	Other Important classes of this project are
	- IrfPac::SaraPac::SaraPacket for analysis of SARA-DPU packet. (The packet coming from DPU.  448 words fixed)
	- IrfPac::SaraPac::SaraBufferedDemultiplexor for demultiplexing DPU output stream to generate CENA and SWIM packet.
*/

/** @page ConversionPacketFormat
    @section PF Packet Format
    @par
    @TODO Write the document.
    @subsection BMUFORMAT BMU format
    @subsection PPL Pipeline

 */
